Package: pkgdiff
Type: Package
Title: Identifies Package Differences
Version: 1.0.3
Maintainer: David Bosak <dbosak01@gmail.com>
Authors@R: c(
    person(
      "David", "Bosak",
      email = "dbosak01@gmail.com",
      role = c("aut", "cre")
    ),
    person(
      "Brian", "Varney",
      email = "brian.varney@experis.com",
      role = c("ctb")
    ),
    person(
      "Kevin", "Putschko",
      email = "kevin.putschko@experis.com",
      role = c("ctb")
    )
  )
Description: Identifies differences between
    versions of a package.  Specifically, the functions help
    determine if there are breaking changes from one package version
    to the next.  The package also includes a stability assessment,
    to help you determine the overall stability of a package, or even
    an entire repository.
License: CC0
Encoding: UTF-8
URL: https://pkgdiff.r-sassy.org, https://github.com/dbosak01/pkgdiff
BugReports: https://github.com/dbosak01/pkgdiff/issues
RoxygenNote: 7.3.3
Depends: R (>= 4.0.0)
Imports: rvest, common, crayon, utils, cranlogs, grDevices, graphics
Suggests: knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-10 15:09:55 UTC; dbosa
Author: David Bosak [aut, cre],
  Brian Varney [ctb],
  Kevin Putschko [ctb]
Repository: CRAN
Date/Publication: 2026-05-10 15:40:02 UTC
Built: R 4.6.0; ; 2026-05-10 17:26:25 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(print,pcache)
S3method(print,pdiff)
S3method(print,pdiff_score)
S3method(print,pinfo)
S3method(print,prepo)
S3method(print,rbreak)
S3method(print,rstability)
export(pkg_cache)
export(pkg_diff)
export(pkg_info)
export(pkg_repo)
export(pkg_stability)
export(pkg_summary)
export(pkg_versions)
export(repo_breakages)
export(repo_stability)
import(common)
import(cranlogs)
import(crayon)
import(grDevices)
import(graphics)
import(rvest)
import(tools)
import(utils)

PkgDiff 1.8
===========

Package Changes Analyzer (pkgdiff) — a tool for visualizing changes in Linux software packages (RPM, DEB, TAR.GZ, etc).

Contents
--------

1. [ About   ](#about)
2. [ Install ](#install)
3. [ Usage   ](#usage)

About
-----

The tool is intended for Linux maintainers who are interested in ensuring compatibility of old and new versions of packages. The tool can compare directories as well (with the help of the -d option).

Sample report: https://abi-laboratory.pro/tracker/package_diff/libssh/0.6.5/0.7.0/report.html

The tool is developed by Andrey Ponomarenko.

Install
-------

    sudo make install prefix=/usr

###### Requires

* Perl 5
* GNU Diff
* GNU Wdiff
* GNU Awk
* GNU Binutils
* Perl-File-LibMagic

###### Suggests

* ABI Compliance Checker 1.99.1 or newer: https://github.com/lvc/abi-compliance-checker/
* ABI Dumper 0.97 or newer: https://github.com/lvc/abi-dumper

Usage
-----

    pkgdiff PKG1 PKG2 [options]

###### Example

    pkgdiff libssh-0.6.5.tar.xz libssh-0.7.0.tar.xz

###### Compare directories

    pkgdiff -d DIR1/ DIR2/ [options]

###### Useful options

| Option              | Meaning                                      |
|---------------------|----------------------------------------------|
| -c/-hide-unchanged  | Don't show unchanged files in the report     |
| -list-added-removed | Show content of added and removed text files |
| -skip-pattern REGEX | Don't check files matching REGEX             |
| -tmp-dir DIR        | Use custom temp directory                    |
| -d/-directories     | Compare directories instead of packages      |

###### Adv. usage

For advanced usage, see output of -help option.

NAME:
  Package Changes Analyzer (pkgdiff) - a tool for visualizing changes in Linux software packages (RPM, DEB, TAR.GZ, etc). The tool is intended for Linux maintainers who are interested in ensuring compatibility of old and new versions of packages.
  
  Sample report: http://abi-laboratory.pro/tracker/package_diff/libssh/0.6.5/0.7.0/report.html
  
  The tool is developed by Andrey Ponomarenko: http://abi-laboratory.pro/

INSTALL:
  sudo make install prefix=/usr

REQUIRES:
  Perl 5
  GNU Diff
  GNU Wdiff
  GNU Awk
  GNU Binutils

USAGE:
  pkgdiff PKG1 PKG2

EXAMPLE:
  pkgdiff libssh-0.6.5.tar.xz libssh-0.7.0.tar.xz

ADV. USAGE:
  For advanced usage, see output of --help option

SUGGESTS:
  ABI Compliance Checker 1.99.1 or newer: https://github.com/lvc/abi-compliance-checker/
  ABI Dumper 0.97 or newer: https://github.com/lvc/abi-dumper

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(pkgdiff)

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # Examine tidymodels package
#  pkg_info("tidymodels")
#  # # A package info object: tidymodels package
#  # - Version: v1.2.0
#  # - Release Date: 2024-03-25
#  # - Title: Easily Install and Load the 'Tidymodels' Packages
#  # - Maintainer: Max Kuhn <max@posit.co>
#  # - License: MIT + file LICENSE
#  # - Description: The tidy modeling "verse" is a collection of packages for
#  # modeling and statistical analysis that share the underlying design
#  # philosophy, grammar, and data structures of the tidyverse.
#  # - Depends: R (>= 3.6)
#  # - Imports: broom (>= 1.0.5), cli (>= 3.6.2), conflicted (>= 1.2.0),
#  # dials (>= 1.2.1), dplyr (>= 1.1.4), ggplot2 (>= 3.5.0), hardhat (>= 1.3.1),
#  # infer (>= 1.0.6), modeldata (>= 1.3.0), parsnip (>= 1.2.1), purrr (>= 1.0.2),
#  # recipes (>= 1.0.10), rlang (>= 1.1.3), rsample (>= 1.2.1), rstudioapi (>= 0.16.0),
#  # tibble (>= 3.2.1), tidyr (>= 1.3.1), tune (>= 1.2.0), workflows (>= 1.1.4),
#  # workflowsets (>= 1.1.0), yardstick (>= 1.3.1)
#  # - Suggests: covr, glue, knitr, rmarkdown, testthat (>= 3.0.0), xml2
#  # - Repository: CRAN
#  # - Cached: TRUE
#  # - Functions: 9
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # Capture package info object
#  res <- pkg_info("tidymodels")
#  
#  # View version
#  res$Version
#  # [1] "1.3.0"
#  
#  # View title
#  res$Title
#  # [1] "Easily Install and Load the 'Tidymodels' Packages"
#  
#  # View release date
#  res$ReleaseDate
#  # [1] "2025-02-21"
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # Get package info
#  res <- pkg_info("tidymodels")
#  
#  # View functions
#  print(res, verbose = TRUE)
#  # # A package info object: tidymodels package
#  # - Version: v1.3.0
#  # - Release Date: 2025-02-21
#  # - Title: Easily Install and Load the 'Tidymodels' Packages
#  # - Maintainer: Max Kuhn <max@posit.co>
#  # - License: MIT + file LICENSE
#  # - Description: The tidy modeling "verse" is a collection of packages for
#  # modeling and statistical analysis that share the underlying design
#  # philosophy, grammar, and data structures of the tidyverse.
#  # - Depends: R (>= 3.6)
#  # - Imports: broom (>= 1.0.7), cli (>= 3.6.4), conflicted (>= 1.2.0),
#  # dials (>= 1.4.0), dplyr (>= 1.1.4), ggplot2 (>= 3.5.1), hardhat (>= 1.4.1),
#  # infer (>= 1.0.7), modeldata (>= 1.4.0), parsnip (>= 1.3.0), purrr (>= 1.0.4),
#  # recipes (>= 1.1.1), rlang (>= 1.1.5), rsample (>= 1.2.1), rstudioapi (>= 0.17.1),
#  # tibble (>= 3.2.1), tidyr (>= 1.3.1), tune (>= 1.3.0), workflows (>= 1.2.0),
#  # workflowsets (>= 1.1.0), yardstick (>= 1.3.2)
#  # - Suggests: covr, glue, knitr, rmarkdown, testthat (>= 3.0.0), xml2
#  # - Downloads/Month: 31789
#  # - Repository: CRAN
#  # - Cached: TRUE
#  # - Functions:
#  #   - pkg_deps(): x, recursive
#  #   - print.tidymodels_conflicts(): x, ..., startup
#  #   - tag_attach(): tag
#  #   - tag_show()
#  #   - tag_update(): tag
#  #   - tidymodels_conflicts()
#  #   - tidymodels_packages(): include_self
#  #   - tidymodels_prefer(): quiet
#  #   - tidymodels_update(): pkg, recursive, ...

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # View package stability
#  pkg_stability("tidymodels")
#  # # A stability score: tidymodels package
#  # - Age: 6.58 years
#  # - Score: 100.0
#  # - Assessment: Perfect
#  # - Version Range: 0.0.1/1.3.0
#  # - Release Range: 2018-07-27/2025-02-21
#  # - Release Count: 14
#  # - Breaking Releases: 0
#  # - Data:
#  #       Package Version                FileName    Release  Size AF AP RF RP BC TF
#  # 1  tidymodels   1.3.0 tidymodels_1.3.0.tar.gz 2025-02-21 46.4K  0  0  0  0  0  9
#  # 2  tidymodels   1.2.0 tidymodels_1.2.0.tar.gz 2024-03-25 49.6K  0  0  0  0  0  9
#  # 3  tidymodels   1.1.1 tidymodels_1.1.1.tar.gz 2023-08-24   50K  0  0  0  0  0  9
#  # 4  tidymodels   1.1.0 tidymodels_1.1.0.tar.gz 2023-05-01   49K  0  0  0  0  0  9
#  # 5  tidymodels   1.0.0 tidymodels_1.0.0.tar.gz 2022-07-13   75K  0  0  0  0  0  9
#  # 6  tidymodels   0.2.0 tidymodels_0.2.0.tar.gz 2022-03-19   75K  0  0  0  0  0  9
#  # 7  tidymodels   0.1.4 tidymodels_0.1.4.tar.gz 2021-10-01   17K  0  0  0  0  0  9
#  # 8  tidymodels   0.1.3 tidymodels_0.1.3.tar.gz 2021-04-19   15K  1  0  0  0  0  9
#  # 9  tidymodels   0.1.2 tidymodels_0.1.2.tar.gz 2020-11-22   26K  0  0  0  0  0  8
#  # 10 tidymodels   0.1.1 tidymodels_0.1.1.tar.gz 2020-07-14   25K  0  0  0  0  0  8
#  # 11 tidymodels   0.1.0 tidymodels_0.1.0.tar.gz 2020-02-16   57K  0  0  0  0  0  8
#  # 12 tidymodels   0.0.3 tidymodels_0.0.3.tar.gz 2019-10-05   19K  0  0  0  0  0  8
#  # 13 tidymodels   0.0.2 tidymodels_0.0.2.tar.gz 2018-11-27   18K  0  0  0  0  0  8
#  # 14 tidymodels   0.0.1 tidymodels_0.0.1.tar.gz 2018-07-27   18K  8  8  0  0  0  8
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # View package stability
#  pkg_stability("dbplyr")
#  # # A stability score: dbplyr package
#  # - Age: 7.67 years
#  # - Score: 73.7
#  # - Assessment: Unstable
#  # - Version Range: 1.0.0/2.5.0
#  # - Release Range: 2017-06-09/2024-03-19
#  # - Release Count: 23
#  # - Breaking Releases: 10
#  # - Data:
#  #    Package Version            FileName    Release   Size  AF  AP RF RP BC  TF
#  # 1   dbplyr   2.5.0 dbplyr_2.5.0.tar.gz 2024-03-19 752.6K  32   6 12  0  1 611
#  # 2   dbplyr   2.4.0 dbplyr_2.4.0.tar.gz 2023-10-26   743K  74  80 45 33  1 591
#  # 3   dbplyr   2.3.4 dbplyr_2.3.4.tar.gz 2023-09-27   706K   0   0  0  0  0 562
#  # 4   dbplyr   2.3.3 dbplyr_2.3.3.tar.gz 2023-07-07   706K   0   0  0  0  0 562
#  # 5   dbplyr   2.3.2 dbplyr_2.3.2.tar.gz 2023-03-21   707K   0   1  0  1  1 562
#  # 6   dbplyr   2.3.1 dbplyr_2.3.1.tar.gz 2023-02-24   707K   5   6  1  0  1 562
#  # 7   dbplyr   2.3.0 dbplyr_2.3.0.tar.gz 2023-01-16   688K  33  16  5  1  1 558
#  # 8   dbplyr   2.2.1 dbplyr_2.2.1.tar.gz 2022-06-27   655K   4   0  0  0  0 530
#  # 9   dbplyr   2.2.0 dbplyr_2.2.0.tar.gz 2022-06-05   655K 136  47 53  1  1 526
#  # 10  dbplyr   2.1.1 dbplyr_2.1.1.tar.gz 2021-04-06   568K   5   0  0  0  0 443
#  # 11  dbplyr   2.1.0 dbplyr_2.1.0.tar.gz 2021-02-03   566K  15   2  0  0  0 438
#  # 12  dbplyr   2.0.0 dbplyr_2.0.0.tar.gz 2020-11-03   522K 139  33 94 11  1 423
#  # 13  dbplyr   1.4.4 dbplyr_1.4.4.tar.gz 2020-05-27   417K   5   0  0  0  0 378
#  # 14  dbplyr   1.4.3 dbplyr_1.4.3.tar.gz 2020-04-19   413K   9   3  2  2  1 373
#  # 15  dbplyr   1.4.2 dbplyr_1.4.2.tar.gz 2019-06-17   296K   1   0  0  0  0 366
#  # 16  dbplyr   1.4.1 dbplyr_1.4.1.tar.gz 2019-06-06   296K   1   0  0  0  0 365
#  # 17  dbplyr   1.4.0 dbplyr_1.4.0.tar.gz 2019-04-24   295K  24  30 17 13  1 364
#  # 18  dbplyr   1.3.0 dbplyr_1.3.0.tar.gz 2019-01-09   263K   2   0  0  0  0 357
#  # 19  dbplyr   1.2.2 dbplyr_1.2.2.tar.gz 2018-07-25   258K   0   0  0  0  0 355
#  # 20  dbplyr   1.2.1 dbplyr_1.2.1.tar.gz 2018-02-19   257K   0   0  0  0  0 355
#  # 21  dbplyr   1.2.0 dbplyr_1.2.0.tar.gz 2018-01-03   257K  43   6  2  2  1 355
#  # 22  dbplyr   1.1.0 dbplyr_1.1.0.tar.gz 2017-06-27   241K  23   2  0  0  0 314
#  # 23  dbplyr   1.0.0 dbplyr_1.0.0.tar.gz 2017-06-09   235K 291 291  0  0  0 291
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # Get package summary report
#  pkg_summary("dplyr")

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # Compare package versions
#  pkg_diff("dbplyr", "2.4.0", "2.5.0")
#  # # A difference object: dbplyr package
#  # - Comparing: v2.4.0/v2.5.0
#  # - Breaking Changes: TRUE
#  # - Added Functions:
#  #   - [.dbplyr_table_path()
#  #   - [[.dbplyr_table_path()
#  #   - as_table_path()
#  #   - c.dbplyr_table_path()
#  #   - check_table_path()
#  #   - complete.tbl_lazy()
#  #   - db_explain.Oracle()
#  #   - db_explain.OraConnection()
#  #   - db_sql_render.Microsoft SQL Server()
#  #   - escape.dbplyr_table_path()
#  #   - escape.default()
#  #   - expand.tbl_lazy()
#  #   - fill.tbl_lazy()
#  #   - group_by_drop_default.tbl_lazy()
#  #   - intersect.tbl_lazy()
#  #   - is_table_path()
#  #   - pivot_longer.tbl_lazy()
#  #   - pivot_wider.tbl_lazy()
#  #   - print.dbplyr_table_path()
#  #   - reframe.tbl_lazy()
#  #   - replace_na.tbl_lazy()
#  #   - setdiff.OraConnection()
#  #   - setdiff.tbl_lazy()
#  #   - setdiff.tbl_Oracle()
#  #   - simulate_mariadb()
#  #   - sql_escape_ident.default()
#  #   - sql_escape_ident.MySQLConnection()
#  #   - sql_escape_string.default()
#  #   - table_path_components()
#  #   - table_path_components.default()
#  #   - table_path_name()
#  #   - union.tbl_lazy()
#  # - Added Parameters:
#  #   - flatten_query.base_query(): con
#  #   - flatten_query.multi_join_query(): con
#  #   - flatten_query.select_query(): con
#  #   - flatten_query.union_query(): con
#  #   - flatten_query.values_query(): con
#  #   - sql_translation.MySQL(): con
#  # - Removed Functions:
#  #   - as_table_ident.character()
#  #   - as_table_ident.dbplyr_catalog()
#  #   - as_table_ident.dbplyr_schema()
#  #   - as_table_ident.dbplyr_table_ident()
#  #   - as_table_ident.Id()
#  #   - as_table_ident.ident()
#  #   - as_table_ident.ident_q()
#  #   - as_table_ident.sql()
#  #   - escape.dbplyr_table_ident()
#  #   - format.dbplyr_table_ident()
#  #   - sql_escape_ident.DBIConnection()
#  #   - sql_escape_string.DBIConnection()
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # Get package versions
#  pkg_versions("stringr")
#  #    Package Version              FileName    Release   Size
#  # 1  stringr   1.5.1  stringr_1.5.1.tar.gz 2023-11-14   172K
#  # 2  stringr   1.5.0  stringr_1.5.0.tar.gz 2022-12-02   172K
#  # 3  stringr   1.4.1  stringr_1.4.1.tar.gz 2022-08-21   133K
#  # 4  stringr   1.4.0  stringr_1.4.0.tar.gz 2019-02-10   133K
#  # 5  stringr   1.3.1  stringr_1.3.1.tar.gz 2018-05-10   116K
#  # 6  stringr   1.3.0  stringr_1.3.0.tar.gz 2018-02-19   116K
#  # 7  stringr   1.2.0  stringr_1.2.0.tar.gz 2017-02-18    92K
#  # 8  stringr   1.1.0  stringr_1.1.0.tar.gz 2016-08-19    62K
#  # 9  stringr   1.0.0  stringr_1.0.0.tar.gz 2015-04-30    34K
#  # 10 stringr   0.6.2  stringr_0.6.2.tar.gz 2012-12-06    20K
#  # 11 stringr   0.6.1  stringr_0.6.1.tar.gz 2012-07-25    20K
#  # 12 stringr     0.6    stringr_0.6.tar.gz 2011-12-08    20K
#  # 13 stringr     0.5    stringr_0.5.tar.gz 2011-06-30    18K
#  # 14 stringr     0.4    stringr_0.4.tar.gz 2010-08-24    16K
#  # 15 stringr     0.3    stringr_0.3.tar.gz 2010-02-15    11K
#  # 16 stringr     0.2    stringr_0.2.tar.gz 2009-11-16    10K
#  # 17 stringr  0.1.10 stringr_0.1.10.tar.gz 2009-11-09   6.8K
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # Check version in repo
#  pkg_repo("stringr")
#  # # A package repo object
#  # - Repo Version: 4.4
#  # - Packages:
#  #   Package Version
#  # 1 stringr   1.5.1
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # Create vector of packages
#  v1 <- c("curl", "dplyr", "fmtr", "purrr")
#  
#  # Get current versions
#  res1 <- pkg_repo(v1, ver = "current")
#  
#  # View results
#  res1
#  # # A package repo object
#  # - Repo Version: 4.4
#  # - Packages:
#  #   Package Version
#  # 1    curl   5.2.1
#  # 2   dplyr   1.1.4
#  # 3    fmtr   1.6.5
#  # 4   purrr   1.0.2
#  
#  # Get latest versions
#  res2 <- pkg_repo(v1, ver = "latest")
#  
#  # View results
#  res2
#  # # A package repo object
#  # - Repo Version: latest
#  # - Packages:
#  #   Package Version
#  # 1    curl   6.2.1
#  # 2   dplyr   1.1.4
#  # 3    fmtr   1.6.8
#  # 4   purrr   1.0.4
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # View upgrade breakages
#  repo_breakages(res1, res2)
#  # Comparing curl v5.2.1/v6.2.0
#  # Comparing fmtr v1.6.5/v1.6.8
#  # Comparing purrr v1.0.2/v1.0.4
#  # # A repo breakages object
#  # - Run Datetime: 2025-03-01 17:58 UTC
#  # - Summary:
#  #   Package Version1 Version2 Breakages
#  # 1    curl    5.2.1    6.2.1      TRUE
#  # 2   dplyr    1.1.4    1.1.4     FALSE
#  # 3    fmtr    1.6.5    1.6.8     FALSE
#  # 4   purrr    1.0.2    1.0.4     FALSE
#  # - Details:
#  # # A difference object: curl package
#  # - Comparing: v5.2.1/v6.2.1
#  # - Breaking Changes: TRUE
#  # - Added Functions: 1
#  # - Added Parameters: 3
#  # - Removed Parameters:
#  #   - multi_download(): timeout

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  
#  # Get current versions from local repo
#  r1 <- pkg_repo(ver = "current")
#  
#  # View Results
#  r1
#  # # A package repo object
#  # - Repo Version: 4.2
#  # - Packages:
#  #            Package              Version
#  # 1            abind                1.4-5
#  # 2          admiral               0.10.1
#  # 3       admiraldev                0.3.0
#  # 4      admiralonco                0.3.0
#  # 5             ards                0.1.0
#  # 6            arrow             11.0.0.3
#  # 7          askpass                  1.1
#  # 8       assertthat                0.2.1
#  # 9           aws.s3               0.3.21
#  # 10   aws.signature                0.6.0
#  # 11       backports                1.4.1
#  # 12       base64enc                0.1-3
#  # 13             bit                4.0.4
#  # 14           bit64                4.0.5
#  # 15          bitops                1.0-7
#  # 16            blob                1.2.3
#  # ...
#  
#  # Get latest versions of current repo on CRAN
#  r2 <- pkg_repo(r1$Package, ver = "latest")
#  
#  # View Results
#  r2
#  # # A package repo object
#  # - Repo Version: latest
#  # - Packages:
#  #            Package    Version
#  # 1            abind      1.4-8
#  # 2          admiral      1.2.0
#  # 3       admiraldev      1.2.0
#  # 4      admiralonco      1.2.0
#  # 5             ards      0.1.1
#  # 6            arrow   18.1.0.1
#  # 7          askpass      1.2.1
#  # 8       assertthat      0.2.1
#  # 9           aws.s3     0.3.21
#  # 10   aws.signature      0.6.0
#  # 11       backports      1.5.0
#  # 12       base64enc      0.1-3
#  # 13             bit    4.5.0.1
#  # 14           bit64    4.6.0-1
#  # 15          bitops      1.0-9
#  # 16            blob      1.2.4
#  # ...
#  
#  
#  # Check for breakages on whole repository
#  res <- repo_breakages(r1, r2)
#  # Comparing abind v1.4-5/v1.4-8
#  # Comparing admiral v0.10.1/v1.2.0
#  # Comparing admiraldev v0.3.0/v1.2.0
#  # Comparing admiralonco v0.3.0/v1.2.0
#  # Comparing ards v0.1.0/v0.1.1
#  # Archive versions of 'ards' not available.
#  # Comparing arrow v11.0.0.3/v18.1.0.1
#  # Comparing askpass v1.1/v1.2.1
#  # Comparing backports v1.4.1/v1.5.0
#  # Comparing bit v4.0.4/v4.5.0.1
#  # Comparing bit64 v4.0.5/v4.6.0-1
#  # Comparing bitops v1.0-7/v1.0-9
#  # Comparing blob v1.2.3/v1.2.4
#  # Comparing box v1.1.3/v1.2.0
#  # Comparing brew v1.0-8/v1.0-10
#  # Comparing brio v1.1.3/v1.1.5
#  # Comparing broom v1.0.1/v1.0.7
#  # Comparing bslib v0.4.0/v0.9.0
#  # Comparing cachem v1.0.6/v1.1.0
#  # Comparing callr v3.7.3/v3.7.6
#  # ...

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # View Summary Table
#  res$Summary
#  #            Package             Version1   Version2 Breakages
#  # 1            abind                1.4-5      1.4-8     FALSE
#  # 2          admiral               0.10.1      1.2.0      TRUE
#  # 3       admiraldev                0.3.0      1.2.0      TRUE
#  # 4      admiralonco                0.3.0      1.2.0      TRUE
#  # 5             ards                0.1.0      0.1.1        NA
#  # 6            arrow             11.0.0.3   18.1.0.1     FALSE
#  # 7          askpass                  1.1      1.2.1     FALSE
#  # 8       assertthat                0.2.1      0.2.1     FALSE
#  # 9           aws.s3               0.3.21     0.3.21     FALSE
#  # 10   aws.signature                0.6.0      0.6.0     FALSE
#  # 11       backports                1.4.1      1.5.0     FALSE
#  # 12       base64enc                0.1-3      0.1-3     FALSE
#  # 13             bit                4.0.4    4.5.0.1     FALSE
#  # 14           bit64                4.0.5    4.6.0-1      TRUE
#  # 15          bitops                1.0-7      1.0-9     FALSE
#  # 16            blob                1.2.3      1.2.4     FALSE
#  # ...
#  
#  # View Difference Details
#  res$Details$admiralonco
#  # # A difference object: admiralonco package
#  # - Comparing: v0.3.0/v1.2.0
#  # - Breaking Changes: TRUE
#  # - Added Functions:
#  #   - bor_cr()
#  #   - bor_ne()
#  #   - bor_non_crpd()
#  #   - bor_pd()
#  #   - bor_pr()
#  #   - bor_sd()
#  #   - cb_y()
#  #   - cbor_cr()
#  #   - cbor_pr()
#  #   - crsp_y_cr()
#  #   - crsp_y_pr()
#  #   - no_data_missing()
#  #   - no_data_n()
#  #   - rsp_y()
#  # - Removed Functions:
#  #   - call_aval_fun()
#  #   - derive_param_lasta()
#  

## ----eval=FALSE, echo=TRUE----------------------------------------------------
#  # Create vector of packages
#  v1 <- c("curl", "dplyr", "dbplyr", "purrr")
#  
#  # Assess stability
#  res <- repo_stability(v1)
#  # Getting stability score for package 'curl'...
#  # Getting stability score for package 'dplyr'...
#  # Getting stability score for package 'dbplyr'...
#  # Getting stability score for package 'purrr'...
#  
#  # View results
#  res
#  # # A repo stability object
#  # - Run Datetime: 2025-03-01 17:46 UTC
#  # - Summary:
#  #   Package    FV    LV         FR         LR TR BR Score        Assessment
#  # 1    curl   0.2 6.2.1 2014-11-20 2025-02-19 51  1  98.0       Very Stable
#  # 2   dplyr   0.1 1.1.4 2014-01-16 2023-11-17 45 20  87.5 Somewhat Unstable
#  # 3  dbplyr 1.0.0 2.5.0 2017-06-09 2024-03-19 23 10  74.1          Unstable
#  # 4   purrr 0.1.0 1.0.4 2015-09-28 2025-02-05 18  6  92.1            Stable
#  

# Introductory Documentation ----------------------------------------------

#' @title ards: A package for creating Analysis Results Datasets
#'
#' @description An Analysis Results Dataset (ARDS) is commonly used in the pharma-biotech industry
#' to capture the results of an analysis in a tabular data structure.  The \strong{ards}
#' package helps create the ARDS. \strong{ards} functions can be called from inside 
#' a report program or a data preparation program.  The functions use a 
#' "bucketing" approach, whereby data can be added to the ARDS in multiple 
#' calls and from multiple intermediate data sets.    
#'     
#' @details 
#' The \strong{ards} package allows you to easily create an ARDS dataset
#' in a standard clinical reporting or data preparation program.  The 
#' \strong{ards} package contains only three functions, and each will
#' be used to create the ARDS dataset.  The functions will be called in the 
#' following order, and for the described purpose:
#' \itemize{
#'   \item \code{\link{init_ards}}: Initialize an ARDS dataset
#'   \item \code{\link{add_ards}}: Add data to an ARDS dataset
#'   \item \code{\link{get_ards}}: Extract the completed ARDS
#' }
#' Click on the links above for more information about each function.
#' 
#' The following figure describes the structure of the ARDS dataset.  This
#' structure is recommended by CDISC.
#' \figure{structure.png}
#' 
#' @docType package
#' @keywords internal
#' @name ards
NULL

# Globals -----------------------------------------------------------------



ardsenv <- new.env()

ardsenv$ards <- NULL
ardsenv$template <- NULL



# Functions ---------------------------------------------------------------



#' @title Initialize the Analysis Results Dataset
#' @description A function to initialize the Analysis Results Dataset (ARDS).
#' This function will
#' first create a data template in the desired structure, and then
#' populate common values across the dataset from that template.  
#' These common values will be
#' repeated on each row of the analysis data frame for subsequent inserts
#' from the \code{\link{add_ards}} function.
#' @param studyid The study for which the analysis was performed. 
#' This parameter is optional.
#' @param tableid A table identifier to use for the results. This value 
#' identifies the table within the study.  Optional string
#' value.
#' @param adsns A vector of source dataset names.  This parameter is used to 
#' identify the input data for the analysis.  This parameter is optional.
#' @param population A description of the analysis population.  This parameter
#' is used to identify the population for analysis.  This parameter is optional.
#' @param time A description of the time frame used in the analysis.  For example, 
#' in a clinical study, the "time" value may identify the visit on which the
#' analysis is based.
#' @param where An optional description of the criteria used to subset the 
#' data for analysis.
#' @param reset If true, clears out the existing ARDS dataset and replaces with
#' an empty template.  Otherwise, just assign new parameter values to the 
#' existing template.  The default
#' value is TRUE, meaning the ARDS in memory will be cleared every time 
#' \code{init_ards}
#' is called.  If you wish to assign new initialization values, but 
#' keep appending to the existing ARDS dataset, set this parameter to FALSE.  
#' This feature is used when you are creating two different tables in the 
#' same program.
#' @return The initialized analysis dataset.
#' @family ards
#' @examples 
#' library(ards)
#' library(dplyr)
#' 
#' # Initialize the ARDS
#' # - These values will be common through the dataset
#' init_ards(studyid = "MTCARS",
#'           tableid = "01", adsns = "mtcars",
#'           population = "all cars",
#'           time = "1973")
#' 
#' # Perform analysis on MPG
#' # - Using cylinders as a by-group
#' analdf <- mtcars |> 
#'   select(cyl, mpg) |> 
#'   group_by(cyl) |> 
#'   summarize(n = n(),
#'             mean = mean(mpg),
#'             std = sd(mpg),
#'             min = min(mpg),
#'             max = max(mpg))
#' 
#' # View analysis data
#' analdf
#' #     cyl     n  mean   std   min   max
#' #   <dbl> <int> <dbl> <dbl> <dbl> <dbl>
#' # 1     4    11  26.7  4.51  21.4  33.9
#' # 2     6     7  19.7  1.45  17.8  21.4
#' # 3     8    14  15.1  2.56  10.4  19.2
#' 
#' # Add analysis data to ARDS
#' # - These values will be unique per row
#' add_ards(analdf, 
#'          statvars = c("n", "mean", "std", "min", "max"),
#'          anal_var = "mpg", trtvar = "cyl")
#' 
#' 
#' # Get the ARDS
#' ards <- get_ards() 
#' 
#' # Uncomment to view ards
#' # View(ards)
#' @export
init_ards <- function(studyid = NA,
                      tableid = NA, adsns = NA,
                      population = NA, time = NA, where = NA, reset = TRUE) {


  if (reset) {
    if (!is.null(ardsenv$ards)) {
      ardsenv$ards <- NULL
    }
  }
  
  # Collapse analysis data
  if (all(is.na(adsns) == FALSE))
    adsns <- paste(adsns, sep ="", collapse = "|")

  # Create template record
  ardsenv$template <- data.frame(studyid = studyid, resultid = 0,
                                 tableid = tableid,
                                 adsns = adsns,
                                 population = population, time = time,
                                 where = where, byvar1 = NA, byvar2 = NA,
                                 byvar3 = NA, byvar4 = NA, byvar5 = NA,
                                 byvar6 = NA, byvar7 = NA, byvar8 = NA,
                                 byvar9 = NA, byval1 = NA, byval2 = NA,
                                 byval3 = NA, byval4 = NA, byval5 = NA,
                                 byval6 = NA, byval7 = NA, byval8 = NA,
                                 byval9 = NA, trtvar = NA, trtval = NA,
                                 paramcd = NA, anal_var = NA, anal_val = NA,
                                 statname = NA, statval = NA, statdesc = NA)

  if (reset) {
    # Assign empty template to ards
    ardsenv$ards <- ardsenv$template[0, ]
  }

  return(ardsenv$template)

}



#' @title Adds data to an Analysis Results Dataset
#' @description The \code{add_ards} function dumps data from an input analysis 
#' dataset
#' to the ARDS dataset.  The function is designed to be pipe-friendly, and will
#' return the input dataset unaltered.  The parameters on the function
#' define how to extract the desired data from the analysis dataset.
#' The "statvars" parameter defines which columns contain desired
#' analysis results.  The values in these columns will be used to populate the
#' "statval" variable in the output dataset.  Other parameters are used to
#' define identifying information for the statistics values, and are optional.
#' 
#' The \code{add_ards} function should be called immediately after any
#' calculations, while the analysis results are still in numeric form.  This 
#' recommendation is to ensure that the ARDS will contain full precision of the
#' analysis values.  Once the analysis values are dumped into the ARDS, you
#' may proceed to transform and format your analysis data, without affecting
#' the values captured in the ARDS. 
#' @param data The input data to create analysis results for.  This parameter
#' is required.
#' @param statvars  A vector of column names that identify the desired results.
#' Statvar columns must be numeric.  This parameter is required.
#' @param statdesc A vector of values or a column name that identifies a description 
#' for each statvar. If passed as a vector of values, the number of values
#' should correspond to the number of 'statvar' variables.  
#' @param byvars A vector of column names to use for by variables.
#' @param trtvar A column name to use for the treatment variable.
#' @param paramcd A character string that describes the analysis parameter 
#' code or column name that contains the parameter code. If supplied as a
#' column name, the function will populate the 'paramcd' column in the ARDS
#' with the value of the 'paramcd' column.
#' @param anal_var A column name for the analysis variable or a string
#' that identifies the analysis variable.
#' @param anal_val The analysis variable value.  Can be identified by a column
#' name or a vector of values. By default, the analysis values will be taken
#' from the values of the variable passed in 'anal_var'.  This parameter 
#' exists so that you may pass in the values from a different variable, if desired.
#' @family ards
#' @return The input data frame, unaltered.
#' @examples 
#' library(ards)
#' library(dplyr)
#' 
#' # Initialize the ARDS
#' init_ards(studyid = "MTCARS",
#'           tableid = "01", adsns = "mtcars",
#'           population = "all cars",
#'           time = "1973")
#' 
#' # Perform analysis on MPG
#' # - Add to ARDS from within continuous variable pipeline
#' mpgdf <- mtcars |> 
#'   select(cyl, mpg) |> 
#'   group_by(cyl) |> 
#'   summarize(n = n(),
#'             mean = mean(mpg),
#'             std = sd(mpg),
#'             min = min(mpg),
#'             max = max(mpg)) |> 
#'   add_ards(statvars = c("n", "mean", "std", "min", "max"),
#'          anal_var = "mpg", trtvar = "cyl")
#'             
#' # Perform analysis on GEAR
#' # - Add to ARDS from within categorical variable pipeline
#' geardf <- mtcars |> 
#'   mutate(denom = n()) |> 
#'   select(cyl, gear, denom) |> 
#'   group_by(cyl, gear) |> 
#'   summarize(cnt = n(), 
#'             denom = max(denom)) |>
#'   mutate(pct = cnt / denom * 100) |> 
#'   add_ards(statvars = c("cnt", "pct", "denom"),
#'          anal_var = "gear", trtvar = "cyl")
#' 
#' # Get the ARDS
#' ards <- get_ards() 
#' 
#' # Uncomment to view ards
#' # View(ards)
#' @export
add_ards <- function(data, statvars, statdesc = NULL,
                     byvars = NULL, trtvar = NULL, paramcd = NULL,
                     anal_var = NULL, anal_val = NULL) {

  #browser()
  
  if (is.null(ardsenv$ards)) {
    
    stop("ARDS dataset is not initialized.") 
  }

  nms <- names(data)

  for (i in seq_len(length(statvars))) {

    # Start with template and turn into list
    ret <- unclass(ardsenv$template)

    # Populate statistics value
    ret$statval <- data[[statvars[[i]]]]

    # Populate statistic name from variable name
    ret$statname <- statvars[[i]]

    # Determine starting point of ResultID
    if (nrow(ardsenv$ards) == 0) {
      strt <- 1
    } else {
      strt <- max(ardsenv$ards$resultid) + 1
    }

    # Populate ResultID
    ret$resultid <- seq(strt, strt + nrow(data) - 1)

    # Populate Stat Description
    if (!is.null(statdesc)) {
      if (all(statdesc %in% nms)) {
        if (length(statdesc) == 1) {

          ret$statdesc <- data[[statdesc]]
        } else if (length(statdesc) >= i) {
          ret$statdesc <- data[[statdesc[[i]]]]
        }
      } else {
        if (length(statdesc) == 1) {
          ret$statdesc <- statdesc
        } else if (length(statdesc) >= i) {
          ret$statdesc <- statdesc[[i]]
        }
      }
    }

    # Populate By variables and values
    if (!is.null(byvars)) {
      for (j in seq_along(length(byvars))) {
        ret[[paste0("byvar", j)]] <- byvars[[j]]

        if (!is.null(data[[byvars[[j]]]]))
          ret[[paste0("byval", j)]] <- data[[byvars[[j]]]]

      }
    }

    # Populate Analysis Variable and Value
    if (!is.null(anal_var)) {
      ret[["anal_var"]] <- anal_var
      if (is.null(anal_val)) {
        if (all(anal_var %in% nms)) {

          ret[["anal_val"]] <- data[[anal_var]]
        }
      } else if (all(anal_val %in% nms)) {
        ret[["anal_val"]] <- data[[anal_val]]
      } else {

        ret[["anal_val"]] <- anal_var
      }
    }


    # Populate Treatment Variable
    if (!is.null(trtvar)) {
      if (trtvar %in% nms) {

        ret[["trtvar"]] <- trtvar
        ret[["trtval"]] <- data[[trtvar]]

      } else {

        ret[["trtval"]] <- trtvar
      }
    }
    
    # Populate PARAM Variable
    if (!is.null(paramcd)) {
      if (paramcd %in% nms) {
        
        ret[["paramcd"]] <- paramcd
        ret[["paramcd"]] <- data[[paramcd]]
        
      } else {
        
        ret[["paramcd"]] <- paramcd
      }
    }



    # Convert to data frame and recycle template values.
    ret <- data.frame(ret)

    # Append to ards
    ardsenv$ards <- rbind(ardsenv$ards, ret)

  }

  return(data)

}

#' @title Returns the current Analysis Results Dataset
#' @description The \code{get_ards} function returns the current state
#' of the Analysis Results Dataset (ARDS) as an R data frame. 
#' This data frame may be saved to disk, saved in
#' a database, or examined from code.  The function takes no parameters.
#' @return A data frame of the current analysis results.
#' @family ards
#' @examples 
#' library(ards)
#' library(dplyr)
#' 
#' # Initialize the ARDS
#' # - These values will be common through the dataset
#' init_ards(studyid = "IRIS",
#'           tableid = "01", adsns = "iris",
#'           population = "all flowers",
#'           time = "1973")
#' 
#' # Perform analysis on Petal.Length
#' # - Using Species as a by-group
#' analdf1 <- iris |> 
#'   select(Petal.Length, Species) |> 
#'   group_by(Species) |> 
#'   summarize(n = n(),
#'             mean = mean(Petal.Length),
#'             std = sd(Petal.Length),
#'             min = min(Petal.Length),
#'             max = max(Petal.Length)) |> 
#'   add_ards(statvars = c("n", "mean", "std", "min", "max"),
#'            statdesc = c("Count", "Mean", "STD", "Minimum", "Maximum"),
#'            anal_var = "Petal.Length", trtvar = "Species")
#'            
#' # Perform analysis on Petal.Width
#' # - Using Species as a by-group
#' analdf2 <- iris |> 
#'   select(Petal.Width, Species) |> 
#'   group_by(Species) |> 
#'   summarize(n = n(),
#'             mean = mean(Petal.Width),
#'             std = sd(Petal.Width),
#'             min = min(Petal.Width),
#'             max = max(Petal.Width)) |> 
#'   add_ards(statvars = c("n", "mean", "std", "min", "max"),
#'            statdesc = c("Count", "Mean", "STD", "Minimum", "Maximum"),
#'            anal_var = "Petal.Width", trtvar = "Species")
#' 
#' # Get the ARDS
#' ards <- get_ards() 
#' 
#' # Uncomment to view ards
#' # View(ards)
#' @export
get_ards <- function() {

 return(ardsenv$ards)

}

# Package Repo ------------------------------------------------------------



#' @title Retrieve Package Versions from a Repository
#' @description
#' The \code{pkg_repo} function combines information related to a package repository.
#' The function retrieves all the packages in a repository for a specified version
#' of R.  Results will list the package name and version.
#' @param pkgs A vector of package names used to subset the repository list.
#' Default is NULL, which means all packages in the repository will be returned.
#' @param ver The R version of the repository. Pass the R version
#' as a quoted string. Default is "current",
#' meaning the current version of the repository. The value "latest"
#' will return the latest versions on CRAN.
#' @param libpaths A vector of paths specifying the locations of the repositories
#' to query. Default
#' is NULL, meaning the function will use the default R locations on the machine.
#' These default paths are identified by the Base R function \code{.libPaths()}.
#' @returns An object of class "prepo".  The object will contain a data
#' frame of R packages with corresponding version numbers.  This table
#' may then be passed to \code{\link{repo_breakages}} or
#' \code{\link{repo_stability}}.
#'
#' If a package is not found in the repo,
#' a row will still be returned for the package.  However, the version
#' will appear as NA.
#' @examples
#' # Create vector of packages
#' pkgs <- c("common", "dplyr", "rvest", "stringr")
#'
#' # Retrieve latest versions
#' pkg_repo(pkgs, ver = "latest")
#' # # A package repo object
#' # - Repo Version: latest
#' # - Packages:
#' #   Package Version
#' # 1  common   1.1.3
#' # 2   dplyr   1.1.4
#' # 3   rvest   1.0.4
#' # 4 stringr   1.5.1
#'
#' @family prepo
#' @import common
#' @export
pkg_repo <- function(pkgs = NULL, ver = "current", libpaths = NULL) {

  mver <- NULL

  refresh_package_lists()

  if (ver == "current") {
    lst <- installed_packages(pkgs, repos = libpaths)

    if (is.null(libpaths))
      mver <- get_rversion()
    else
      mver <- "current"

  } else if (ver == "latest") {

    if (is.null(pkgs))
      lst <- e$AvailablePackages[ , c("Package", "Version")]
    else
      lst <- e$AvailablePackages[e$AvailablePackages$Package %in% pkgs,
                                 c("Package", "Version")]

    mver <- ver

  } else {

    if (is.null(libpaths)) {

      libpaths <- .libPaths()
    }

    if (!is.null(ver)) {

      mver <- ver

      # Default repo
      bp <- get_base_paths(libpaths)

      vpths <- c()

      sver <- split_version(ver)

      idx <- 1
      for (pbp in bp) {
        tmp <- common::dir.find(pbp, sver$lver, up = 0, down = 2)
        if (is.null(tmp))
          tmp <- common::dir.find(pbp, sver$ver, up = 0, down = 2)
        else {

          if (length(tmp) > 1) {
            stop("Multiple releases found for R version ", ver, ".\n",
                 "Please use a three-level version specification:\n",
                 paste(tmp, collapse = "\n"))
          }

          if (basename(tmp) != "library")
            tmp <- file.path(tmp, "library")

        }

        if (!is.null(tmp)) {
          vpths[idx] <- tmp
          idx <- idx + 1
        }
      }

      libpaths <- vpths

      lst <- installed_packages(pkgs, repos = vpths)


    } else {

      lst <- installed_packages(pkgs, repos = libpaths)

    }

  }

  if (nrow(lst) < length(pkgs)) {

    mpkgs <- pkgs[!pkgs %in% lst$Package]
    mdf <- data.frame(Package = mpkgs, Version = NA)
    lst <- rbind(lst, mdf)
  }

  ret <- lst
  class(ret) <- c("prepo", class(ret))
  rownames(ret) <- NULL

  if (!is.null(mver))
    attr(ret, "Version") <- mver
  if (!is.null(libpaths))
    attr(ret, "LibPaths") <- libpaths

  return(ret)
}


#' @title Print a Package Repo Object
#' @description
#' Print routine for a package repo object of class "prepo".
#' @param x The package repo to print.
#' @param ... Follow-on parameters to the print function.
#' @family prepo
#' @import crayon
#' @returns The package repo object, invisibly.
#' @examples
#' # Create vector of packages
#' pkgs <- c("tidymodels", "rsample", "parsnip", "recipes", "workflows")
#'
#' # Retrieve latest versions
#' res <- pkg_repo(pkgs, ver = "latest")
#'
#' # Print results
#' print(res)
#' # # A package repo object
#' # - Repo Version: latest
#' # - Packages:
#' #      Package Version
#' # 1    parsnip   1.3.0
#' # 2    recipes   1.1.1
#' # 3    rsample   1.2.1
#' # 4 tidymodels   1.3.0
#' # 5  workflows   1.2.0
#' @export
print.prepo <- function(x, ...) {

  grey60 <- crayon::make_style(grey60 = "#999999")
  cat(grey60(paste0("# A package repo object\n")))


  ver <- attr(x, "Version")

  if (!is.null(ver)) {
    cat(paste0("- Repo Version: ", ver , "\n"))
  }

  pths <- attr(x, "LibPaths")

  if (!is.null(pths)) {
    cat(paste0("- Lib Paths:\n"))
    cat(paste0("  ", pths, "\n"))
  }

  if (!is.null(x)) {
    cat(paste0("- Packages:\n"))
    print(as.data.frame(x))
  }

  invisible(x)
}


# Utilities ---------------------------------------------------------------



get_base_paths <- function(pths) {

  ver <- get_rversion()

  lst <- strsplit(pths, "/", fixed = TRUE)

  ret <- c()
  idx <- 1
  for (li in lst) {
    gpos <- grepl(ver, li, fixed = TRUE)
    tmp <- ""
    for (pt in seq_len(length(li))) {
      if (gpos[pt] == FALSE) {
        tmp <- paste0(tmp, li[pt], sep = "/")
      } else
        break
    }
    ret[length(ret) + 1] <- tmp
  }

  return(ret)
}

get_rversion <- function() {

  mj <- R.version$major
  pos <- as.numeric(regexpr(".", R.version$minor, fixed = TRUE))
  mn <- substring(R.version$minor, 1, pos - 1)
  ver <- paste0(mj, ".", mn)

  return(ver)
}

split_version <- function(rversion) {

  tpos <- gregexpr(".", rversion, fixed = TRUE)

  ret <- list()
  if (length(tpos[[1]]) == 1) {
    ret$lver <- paste0("R-", rversion, ".*")
    ret$ver <- rversion
  } else if (length(tpos[[1]]) > 1) {
    ret$lver <- paste0("R-", rversion)
    ret$ver <- substring(rversion, 1, tpos[[1]][2] - 1)
  } else if (length(tpos[[1]]) == 0) {
    ret$lver <- paste0("R-", rversion)
    ret$ver <- rversion
  }


  return(ret)
}

dev <- FALSE


test_that("repo1: get_base_paths() works as expected.", {

  if (dev & FALSE) {

    res <- get_base_paths(.libPaths())

    res

    expect_equal(is.null(res), FALSE)
    expect_equal(length(res), 2)


  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("repo2: get_rversion() works as expected.", {


  res <- get_rversion()

  res

  expect_equal(is.null(res), FALSE)
  expect_equal(nchar(res) > 0, TRUE)


})


test_that("repo3: pkg_repo() basic functionality no params.", {

  res <- pkg_repo()

  expect_equal("data.frame" %in% class(res), TRUE)
  expect_equal("prepo" %in% class(res), TRUE)
  expect_equal(is.null(attr(res, "Version")), FALSE)
  expect_equal(is.null(attr(res, "LibPaths")), TRUE)
})


test_that("repo4: pkg_repo() basic functionality no params.", {

  res <- pkg_repo(c("common", "rvest", "crayon"))

  res

  expect_equal("data.frame" %in% class(res), TRUE)
  expect_equal("prepo" %in% class(res), TRUE)
  expect_equal(is.null(attr(res, "Version")), FALSE)
  expect_equal(is.null(attr(res, "LibPaths")), TRUE)
  expect_equal(nrow(res) == 3, TRUE)

  res <- pkg_repo(c("common", "rvest", "crayon"), ver = "latest")

  res

  expect_equal("data.frame" %in% class(res), TRUE)
  expect_equal("prepo" %in% class(res), TRUE)
  expect_equal(is.null(attr(res, "Version")), FALSE)
  expect_equal(attr(res, "Version"), "latest")
  expect_equal(is.null(attr(res, "LibPaths")), TRUE)
  expect_equal(nrow(res) == 3, TRUE)

})



test_that("repo5: split_version() works as expected.", {

   res <- split_version("4.1")

   expect_equal(res$lver, "R-4.1.*")
   expect_equal(res$ver, "4.1")

   res <- split_version("4.1.2")
   expect_equal(res$lver, "R-4.1.2")
   expect_equal(res$ver, "4.1")

   res <- split_version("4")
   expect_equal(res$lver, "R-4.*")
   expect_equal(res$ver, "4")
})

test_that("repo6: pkg_repo() ver parameter.", {

  if (dev) {

    expect_error(pkg_repo(c("common", "rvest", "crayon"), "4.2"))

    res <- pkg_repo(c("common", "rvest", "crayon"), "4.2.1")

    res

    expect_equal("data.frame" %in% class(res), TRUE)
    expect_equal("prepo" %in% class(res), TRUE)
    expect_equal(attr(res, "Version"), "4.2.1")
    expect_equal(length(attr(res, "LibPaths")), 2)
    expect_equal(nrow(res) == 3, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})


test_that("repo7: pkg_repo() libpaths parameter.", {

  if (dev) {

    pths <- c("C:/Users/dbosa/AppData/Local/R/win-library/4.2",
             "C:/Program Files/R/R-4.2.1/library")

    res <- pkg_repo(c("common", "rvest", "crayon"), libpaths = pths)

    res

    expect_equal("data.frame" %in% class(res), TRUE)
    expect_equal("prepo" %in% class(res), TRUE)
    expect_equal(attr(res, "Version"), "current")
    expect_equal(length(attr(res, "LibPaths")), 2)
    expect_equal(nrow(res) == 3, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})


test_that("repo8: pkg_repo() version and libpaths parameter.", {

  if (dev) {

    pths <- c("C:/Users/dbosa/AppData/Local/R/win-library/",
              "C:/Program Files/R/")

    res <- pkg_repo(c("common", "rvest", "crayon"), ver = "4.2.1",
                    libpaths = pths)

    res

    expect_equal("data.frame" %in% class(res), TRUE)
    expect_equal("prepo" %in% class(res), TRUE)
    expect_equal(attr(res, "Version"), "4.2.1")
    expect_equal(length(attr(res, "LibPaths")), 2)
    expect_equal(nrow(res) == 3, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})


test_that("repo9: pkg_repo() non-CRAN package.", {


  p <- c("common", "rvest", "forker22")

  res <- pkg_repo(p)

  res
  expect_equal(nrow(res) == 3, TRUE)
  expect_equal("forker22" %in% res$Package, TRUE)

  p <- c("forker22")

  res <-  pkg_repo(p)

  expect_equal(nrow(res) == 1, TRUE)
  expect_equal("forker22" %in% res$Package, TRUE)


})


test_that("repo10: pkg_repo() troublesome packages.", {

    pkgs <- c("tcltk", "Matrix")

    res <- repo_stability(pkgs)

    expect_equal(nrow(res), 2)
    expect_equal(all(is.na(res$Score)), TRUE)

})



# ip <- as.data.frame(installed.packages()[,c(1,3:4)])
# rownames(ip) <- NULL
# ip <- ip[is.na(ip$Priority),1:2,drop=FALSE]

#' @noRd
descriptions <- function(x) {

  ret <- list()

  for (nm in names(x)) {

    if (!is.null(attr(x[[nm]], "description", exact = TRUE))) {
      ret[[nm]] <- attr(x[[nm]], "description", exact = TRUE)
    }

  }

  ret <- forker()

  return(ret)

}

#' @noRd
`descriptions<-` <- function(x, verbose = FALSE, value) {


  if (verbose) {
    x <- descriptions_verbose(x, value)

  } else {

    x <- assign_descriptions(x, value)

  }

  return(x)

}

assign_descriptions <- function(x, value) {

  if (all(is.null(value))) {

    for (nm in names(x)) {

      attr(x[[nm]], "description") <- NULL
    }


  } else {

    for (nm in names(value)) {

      if (!is.null(x[[nm]]))
        attr(x[[nm]], "description") <- value[[nm]]

    }
  }

  return(x)

}

#' @noRd
descriptions_verbose <- function(x, value){

  if(any(duplicated(names(value)))){
    stop("List `value` names must be unique.")
  }
  vars.overdescribed <- setdiff(names(value), names(x))

  if(length(vars.overdescribed) > 0){
    message("The following variables are defined in descriptions list and not in dataframe: ")
    cat("  ", paste0(vars.overdescribed, collapse = ", "), "\n")
  }


  cur.descriptions = descriptions(x)
  description.collisions = intersect(names(cur.descriptions), names(value))
  if(length(description.collisions) > 0){
    description.overwrites = data.frame(variable = description.collisions,
                                        original = do.call(c, cur.descriptions[description.collisions]),
                                        new = do.call(c, value[description.collisions]))
    description.overwrites = description.overwrites[description.overwrites$original != description.overwrites$new,]
    #print(description.overwrites)
    if(nrow(description.overwrites) > 0){
      #browser()
      updates.formatted = paste0("- ", description.overwrites$variable, ":  ",
                                 description.overwrites$original, " -> ",
                                 description.overwrites$new, "\n")

      message("The following descriptions are being updated:")

      for (uf in updates.formatted) {
        cat(paste0(uf))
      }
    }
  }
  x <- assign_descriptions(x, value)

  vars.undefined = setdiff(names(x), names(descriptions(x)))
  if(length(vars.undefined)>0){
    message("The following variables are still undescribed:")
    cat("  ", paste0(vars.undefined, collapse = ", "), "\n")

  } else {
    message("All variables described")
  }
  return(x)
}

#' @title Identify Differences Between Package Versions
#'
#' @description The \strong{pkgdiff} package contains functions to
#' identify differences between versions of R packages. The \strong{pkgdiff}
#' package was written specifically to detect breaking changes and
#' help with package upgrades.
#'
#' @section Functions:
#' The main functions included in the \strong{pkgdiff} package are
#' as follows:
#' \itemize{
#'   \item {\code{\link{pkg_info}}: Retrieves general information about a package,
#'   mostly from the package description file.}
#'   \item {\code{\link{pkg_versions}}: Gets a data frame containing all
#'   versions of a package and the release date for each.}
#'   \item {\code{\link{pkg_diff}}: Compares two package versions, and
#'   returns an object that contains the differences in terms of exported
#'   function signatures.}
#'   \item {\code{\link{pkg_stability}}: Compiles information on package
#'   stability over time, and calculates a score based on this information.
#'   This function is useful when making package selection choices.}
#'   \item {\code{\link{pkg_repo}}: Fetches package names and versions for
#'   packages in a specified repository.}
#'   \item {\code{\link{repo_breakages}}: Identifies breakages when
#'   upgrading from one version of a repository to another.}
#'   \item {\code{\link{repo_stability}}: Performs a stability assessment
#'   on multiple packages, even an entire repository.}
#'   \item {\code{\link{pkg_summary}}: Generates a summary report in HTML
#'   that can be viewed interactively.}
#' }
#' For additional information on \strong{pkgdiff}, see the package vignette
#' \code{vignette('pkgdiff')}, or visit the package web site
#' [here](https://pkgdiff.r-sassy.org/index.html).
#'
#' @keywords internal
#' @aliases pkgdiff-package
#' @name pkgdiff
"_PACKAGE"


# Globals -----------------------------------------------------------------


e <- new.env(parent = emptyenv())
e$LoadDate <- Sys.Date()
e$GithubPath <- "https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/data"

e$RStudioPath <- "https://cran.rstudio.com/src/contrib"
e$Mirror <- "https://cran.rstudio.com/"
e$CranArchivePath <- "https://cran.r-project.org/src/contrib/Archive"
e$CranCurrentPath <- "https://cran.r-project.org/src/contrib/"
e$MirrorArchivePath <- "https://cran.rstudio.com/src/contrib/Archive"
e$MirrorCurrentPath <- "https://cran.rstudio.com/src/contrib/"
e$CranPackagePath <- "https://cran.r-project.org/web/packages"
e$AvailablePackages <- NULL
e$SavedPackages <- NULL
e$BasePackages <- NULL


.onAttach <- function(...) {


  refresh_package_lists()

}


# Utility -----------------------------------------------------------------


#' @import utils
#' @noRd
refresh_package_lists <- function(force = FALSE) {

  ret <- FALSE

  ts <- Sys.Date() - e$LoadDate
  if (is.null(e$AvailablePackages) ||
      is.null(e$SavedPackages) || ts > 1 || force == TRUE) {

    e$AvailablePackages <- tryCatch(available_packages(),
                                    error = function(cond) {
                                      NULL})

    e$SavedPackages <- tryCatch(github_packages(),
                                error = function(cond) {
                                  NULL})

    e$BasePackages <- tryCatch(base_packages(),
                               error = function(cond) {
                                 NULL})

    ret <- TRUE
  }

  return(ret)
}

# Experiments -------------------------------------------------------------

# compareVersion("1.0", "1.0-1")

# getNamespaceExports("fmtr")
# pdb <- tools::CRAN_package_db()

# names(pdb)
#
# getOption("repos")
#
# pdb[1, ]
#
# utils::findCRANmirror("web")
#
# utils::available.packages()
#
# get("R_CRAN_WEB")
#
# packageVersion("fmtr")
# packageDate("fmtr")
# packageDescription("fmtr")
#
# contrib.url("fmtr")
# .packages(all.available = TRUE)
#
# installed.packages()
#
# cranlogs::cran_top_downloads(count = 100)
# cranlogs::cran_downloads()
# cranlogs::cran_downloads(c("accrual", "accessr", "fmtr", "procs"))

# res1 <- as.data.frame(installed.packages())
#
# res1[res1$Package == "tools", ]
#
# res2 <- as.data.frame(available.packages())
#
# res2[res2$Package %in% c("tools", "utils"), ]


# Package Info ------------------------------------------------------------


#' @title A Package Information Class
#' @description
#' Combines information related to a package. Objects of this class are
#' used in several other functions in the \strong{pkgdiff} system.  This
#' class also makes a compact storage format for pre-processed package
#' information.
#' @param pkg The package name.
#' @param ver The package version.
#' @param release The package release date.
#' @param title The package title.
#' @param description The package description.
#' @param maintainer The package maintainer.
#' @param depends A set of package dependencies.
#' @param linkingto The packages this package links to.
#' @param suggests The packages this package suggests.
#' @param enhances The package this package enhances.
#' @param license The package license.
#' @param repository The package repository.
#' @param functions The functions associated with this package.
#' @returns An object of class "pinfo".
#' @family pdiff
#' @noRd
pinfo <- function(pkg, ver = "current", release = NULL, title = NULL, description = NULL,
                     maintainer = NULL,
                     depends = NULL, linkingto = NULL, imports = NULL,
                     suggests = NULL, enhances = NULL, license = NULL,
                     repository = NULL, functions = NULL) {

  inf <- structure(list(), class = c("pinfo", "list"))

  inf$PackageName <- pkg
  inf$Version <- ver
  inf$Title <- title
  inf$Description <- description
  inf$Maintainer <- maintainer
  inf$Depends <- depends
  inf$LinkingTo <- linkingto
  inf$Imports <- imports
  inf$Suggests <- suggests
  inf$Enhances <- enhances
  inf$License <- license
  inf$Repository <- repository
  inf$ReleaseDate <- release
  inf$Functions <- functions

  return(inf)
}



#' @title Print a Package Info Object
#' @description
#' Print routine for a package info object of class "pinfo".
#' @param x The package info to print.
#' @param ... Follow-on parameters to the print function.
#' @param verbose Whether to print all function names or not.
#' Default is FALSE.  When FALSE, only a count of the functions
#' will be printed.
#' @family pdiff
#' @import crayon
#' @returns The package info object, invisibly.
#' @examples
#' # Capture package info
#' res <- pkg_info("patchwork")
#'
#' # View package info
#' print(res)
#' # # A package info object: patchwork package
#' # - Version: v1.3.0
#' # - Release Date: 2024-09-16
#' # - Title: The Composer of Plots
#' # - Maintainer: Thomas Lin Pedersen <thomasp85@gmail.com>
#' #   - License: MIT + file LICENSE
#' # - Description: The 'ggplot2' package provides a strong API for sequentially
#' # building up a plot, but does not concern itself with composition of multiple
#' # plots. 'patchwork' is a package that expands the API to allow for
#' # arbitrarily complex composition of plots by, among others, providing
#' # mathematical operators for combining multiple plots. Other packages that try
#' # to address this need (but with a different approach) are 'gridExtra' and
#' # 'cowplot'.
#' # - Imports: ggplot2 (>= 3.0.0), gtable, grid, stats, grDevices, utils,
#' # graphics, rlang (>= 1.0.0), cli, farver
#' # - Suggests: knitr, rmarkdown, gridGraphics, gridExtra, ragg,
#' # testthat (>= 2.1.0), vdiffr, covr, png, gt (>= 0.11.0)
#' # - Downloads/Month: 239555
#' # - Repository: CRAN
#' # - Cached: TRUE
#' # - Functions: 90
#' @export
print.pinfo <- function(x, ..., verbose = FALSE) {




    # pkg, version, release, title = NULL, description = NULL,
    # maintainer = NULL,
    # depends = NULL, linkingto = NULL,
    # suggests = NULL, enhances = NULL, license = NULL,
    # repository = NULL, functions = NULL

    grey60 <- crayon::make_style(grey60 = "#999999")
    cat(grey60(paste0("# A package info object: ",
                      as.character(x$PackageName), " package\n")))

    if (!is.null(x$Version))
      cat(paste0("- Version: ", "v", x$Version, "\n"))

    if (!is.null(x$Release))
      cat(paste0("- Release Date: ", as.character(x$Release), "\n"))

    if (!is.null(x$Title))
      cat(paste0("- Title: ", as.character(x$Title), "\n"))

    if (!is.null(x$Maintainer))
      cat(paste0("- Maintainer: ", as.character(x$Maintainer), "\n"))

    if (!is.null(x$License))
      cat(paste0("- License: ", as.character(x$License), "\n"))

    if (!is.null(x$Description))
      cat(paste0("- Description: ", as.character(x$Description), "\n"))

    if (!is.null(x$Depends))
      cat(paste0("- Depends: ", as.character(x$Depends), "\n"))

    if (!is.null(x$LinkingTo))
      cat(paste0("- Linking To: ", as.character(x$LinkingTo), "\n"))

    if (!is.null(x$Suggests))
      cat(paste0("- Imports: ", as.character(x$Imports), "\n"))

    if (!is.null(x$Suggests))
      cat(paste0("- Suggests: ", as.character(x$Suggests), "\n"))

    if (!is.null(x$Enhances))
      cat(paste0("- Enhances: ", as.character(x$Enhances), "\n"))

    if (!is.null(x$LastMonthDownloads)) {

      if (x$LastMonthDownloads > 0)
        cat(paste0("- Downloads/Month: ", as.character(x$LastMonthDownloads), "\n"))

    }

    if (!is.null(x$Suggests))
      cat(paste0("- Repository: ", as.character(x$Repository), "\n"))

    if (!is.null(x$Cached))
        cat(paste0("- Cached: ", as.character(x$Cached), "\n"))

    if (!is.null(x$Archived))
      cat(paste0("- Archived: ", as.character(x$Archived), "\n"))

    if (!is.null(x$BasePackage)) {
      if (x$BasePackage == TRUE)
        cat(paste0("- Base Package: TRUE\n"))
    }

    if (verbose == FALSE) {

      if (!is.null(x$Functions))
        cat(paste0("- Functions: ", length(x$Functions), "\n"))

    } else {

      if (!is.null(x$Functions)) {
        cat(paste0("- Functions: \n"))

        for (nm in names(x$Functions)) {
          prms <- paste0(x$Functions[[nm]], collapse = ", ")
          if (nchar(prms) == 0)
            cat(paste0("  - ", nm, "()\n"))
          else
            cat(paste0("  - ", nm, "(): ", prms, "\n"))

        }
      }
    }


  invisible(x)
}


# Package Cache --------------------------------------------------------



#' @title Queries the Package Cache
#' @description
#' The \code{pkg_cache} function queries the package cache, and
#' returns information on which packages are included in the cache.
#' The function also returns the last version of the package cached.
#' @details
#' The \strong{pkgdiff} cache is used to speed up \strong{pkgdiff} functions.
#' Information about each package is pre-processed and stored in the cache
#' on Github.  The functions then retrieve this pre-processed information
#' instead of pulling packages down from CRAN.  To learn more about the cache,
#' refer to \code{vignette('pkgdiff-cache')}.
#' @param pkgs A vector of package names to retrieve cache information about.
#' Default is NULL, which means to return all packages in the cache.
#' @returns An data frame showing the package name and latest package
#' version stored in the cache.  If the package is not stored in the cache,
#' the package version will be NA.  The "Last Update" time stamp is the
#' last day and time the cache was updated.
#' @examples
#' # View single package
#' pkg_cache("dplyr")
#' # # A package cache object
#' # - Last Update: 2025-02-25 14:58 UTC
#' # - Packages:
#' #   Package Version
#' # 1   dplyr   1.1.4
#'
#' # View multiple packages
#' pkgs <- c("dplyr", "tidyr", "stringr")
#' pkg_cache(pkgs)
#' # # A package cache object
#' # - Last Update: 2025-02-25 14:58 UTC
#' # - Packages:
#' #   Package Version
#' # 1   dplyr   1.1.4
#' # 2   tidyr   1.3.1
#' # 3 stringr   1.5.1
#'
#' # Get all packages
#' res <- pkg_cache()
#'
#' # View first 10
#' res[1:10, ]
#' # A package cache object
#' # - Last Update: 2025-02-25 14:58 UTC
#' # - Packages:
#' #        Package Version
#' # 1           A3   1.0.0
#' # 2   abbreviate     0.1
#' # 3     abc.data     1.1
#' # 4          abc   2.2.2
#' # 5  ABCanalysis   1.2.1
#' # 6          abd   0.2-8
#' # 7        abind   1.4-8
#' # 8      acepack   1.6.1
#' # 9       actuar   3.3-5
#' # 10         ada   2.0-5
#' @family pdiff
#' @export
pkg_cache <- function(pkgs = NULL) {

  ov <- github_packages(pkgs)

  nms <- names(ov)
  names(ov) <- NULL

  if (is.null(pkgs)) {
    ret <- data.frame(Package = nms, Version = ov)
  } else {
    ret <- data.frame(Package = pkgs, Version = ov)
  }

  attr(ret, "LastUpdated") <- github_update()

  class(ret) <- c("pcache", "data.frame", "list")

  return(ret)
}



#' @title Print a Package Cache Object
#' @description
#' Print routine for a package cache object of class "pcache".
#' @param x The package cache to print.
#' @param ... Follow-on parameters to the print function.
#' @family pdiff
#' @import crayon
#' @returns The package cache object, invisibly.
#' @examples
#' # Create vector of packages
#' pkgs <- c("ggplot2", "patchwork", "gt")
#'
#' # Capture cache versions
#' res <- pkg_cache(pkgs)
#'
#' # Print cache versions
#' print(res)
#' # # A package cache object
#' # - Last Update: 2025-03-01 20:51 UTC
#' # - Packages:
#' #     Package Version
#' # 1   ggplot2   3.5.1
#' # 2 patchwork   1.3.0
#' # 3        gt  0.11.1
#' @export
print.pcache <- function(x, ...) {

  grey60 <- crayon::make_style(grey60 = "#999999")
  cat(grey60(paste0("# A package cache object\n")))


  lu <- attr(x, "LastUpdate")

  if (!is.null(lu)) {
    tmstmp <- format(lu, format = "%Y-%m-%d %H:%M UTC")
    cat(paste0("- Last Update: ", tmstmp , "\n"))

  }

  if (!is.null(x)) {
    cat(paste0("- Packages:\n"))
    print(as.data.frame(x))
  }

  invisible(x)
}


# Get Info ----------------------------------------------------------------



#' @title Get Information for a Package
#' @description
#' Combines information related to a package. Objects of this class are
#' used by several other functions in the \strong{pkgdiff} system.  This
#' class also makes a compact storage format for pre-processed package
#' information.
#' @details
#' Package information is unique for each version of a package.  Since
#' a package can have multiple versions, you must select which version
#' you wish to return information for.  By default, the version
#' returned is the current version of the package on the current machine.
#' You may also specify a previous version number from the CRAN archive.
#' Another option is to specify the latest version on CRAN.  See the
#' "ver" parameter for additional details on how to select these
#' different package versions.
#'
#' The package info object contains a list of functions and function
#' parameters associated with the specified version of the package.
#' You may access this list using the \code{$Functions} item name.
#'
#' Most other information contained in the info object is retrieved
#' from the package description file. One exception is the downloads
#' per month.  This information is retrieved from CRAN logs.
#'
#' In addition, the package cache status is appended to the info object.
#' The package cache status indicates whether the package info
#' has been stored in the \strong{pkgdiff} Github cache.  Packages that
#' have been stored in the cache enjoy faster results from
#' \strong{pkgdiff} functions.
#'
#' Note that \code{pkg_info} and other \strong{pkgdiff} functions
#' only work with contributed CRAN packages.  They do not
#' work with Base R packages.
#' @param pkg The package name as a quoted string. This parameter is
#' required.
#' @param ver The version of the package to retrieve information for.
#' Pass the version as a quoted string.
#' Special values are "current" and
#' "latest".  The value "current" is the current version of the package
#' running on the machine.  The value "latest" is the latest version
#' of the package from CRAN.
#' @param cache Whether to retrieve the info from the Github cache, or
#' from CRAN.  If TRUE, the function will first search the cache, and
#' return the info if available.  If the info is not available in the
#' Github cache, or the cache parameter is set to FALSE, the info will be
#' retrieved from CRAN.
#' @returns A package information object of class "pinfo".  This object contains
#' a set of general information about the package, such as the version,
#' release date, maintainer, title, etc.  Most of this information comes
#' from the package description file.  In addition, the info object also
#' contains a list of functions in the package and their parameters.
#'
#' If the package has been archived on CRAN, info will be returned,
#' but the "Archived" flag will be set to TRUE. If the
#' package is not found on CRAN, the function will emit a message and return
#' NULL.
#'
#' If the package is a Base R package, a reduced number of fields will be returned,
#' and the package will be marked with \code{Base Package: TRUE}.  Function
#' lists are not available for Base packages.
#' @examples
#' # View package info
#' pkg_info("glue")
#' # A package info object: glue package
#' # - Version: v1.7.0
#' # - Release Date: 2024-01-09
#' # - Title: Interpreted String Literals
#' # - Maintainer: Jennifer Bryan <jenny@posit.co>
#' # - License: MIT + file LICENSE
#' # - Description: An implementation of interpreted string literals, inspired by
#' # Python's Literal String Interpolation
#' # <https://www.python.org/dev/peps/pep-0498/> and Docstrings
#' # <https://www.python.org/dev/peps/pep-0257/> and Julia's Triple-Quoted
#' # String Literals
#' # <https://docs.julialang.org/en/v1.3/manual/strings/#Triple-Quoted-String-Literals-1>.
#' #   - Depends: R (>= 3.6)
#' # - Imports: methods
#' # - Suggests: crayon, DBI (>= 1.2.0), dplyr, knitr, magrittr, rlang,
#' # rmarkdown, RSQLite, testthat (>= 3.2.0), vctrs (>= 0.3.0),
#' # waldo (>= 0.3.0), withr
#' # - Downloads/Month: 1463244
#' # - Repository: CRAN
#' # - Cached: TRUE
#' # - Functions: 24
#'
#' # Get info object
#' res <- pkg_info("glue")
#'
#' # Extract package version
#' res$Version
#' # [1] "1.7.0"
#'
#' # Extract maintainer
#' res$Maintainer
#' # [1] "Jennifer Bryan <jenny@posit.co>"
#'
#' # Extract function list
#' res$Functions
#' # $`[.glue`
#' # [1] "x"   "i"   "..."
#' #
#' # $`[[.glue`
#' # [1] "x"   "i"   "..."
#' #
#' # $`+.glue`
#' # [1] "e1" "e2"
#' #
#' # $as.character.glue
#' # [1] "x"   "..."
#' #
#' # $as_glue
#' # [1] "x"   "..."
#' #
#' # $as_glue.character
#' # [1] "x"   "..."
#' #
#' # $as_glue.default
#' # [1] "x"   "..."
#' #
#' # $as_glue.glue
#' # [1] "x"   "..."
#' #
#' # $backtick
#' # [1] "x"
#' #
#' # $double_quote
#' # [1] "x"
#' #
#' # $glue
#' # [1] "..."          ".sep"         ".envir"
#' # [4] ".open"        ".close"       ".na"
#' # [7] ".null"        ".comment"     ".literal"
#' # [10] ".transformer" ".trim"
#' # ...
#' @family pdiff
#' @export
pkg_info <- function(pkg, ver = "current", cache = TRUE) {

  # browser()

  lv <- get_latest_version(pkg, msg = FALSE)
  archived <- FALSE
  if (!is.null(lv)) {
    if (lv == "archived")
      archived <- TRUE
  }

  based <- is_base(pkg)

  tst <- github_packages(pkg)

  if (is.null(ver)) {

    ver <- get_current_version(pkg)

  } else if (ver == "current") {

    ver <- get_current_version(pkg)

  } else if (ver == "latest") {

    ver <- get_latest_version(pkg)
  }

  if (length(ver) == 0) {
    ver <- lv
  } else if (is.na(ver)) {
    ver <- lv
  }

  if (!is.null(ver) && based == FALSE) {
    tmp <- get_archive_versions(pkg)
    if (archived) {

      if (nrow(tmp) > 0) {

        ver <- tmp[1, "Version"]
        archived <- TRUE

      }
    }

    # No archive versions
    if (is.null(tmp)) {
      if (ver != lv) {
        ver <- lv
      }
    } else if (!ver %in% tmp$Version) {  # Current version doesn't match anything

      ver <- lv
    }
  }

  if (is.null(ver)) {

    message(paste0("No version info available for '", pkg, "'."))
    ret <- NULL

  } else {

    if (!is.null(tst) && !is.na(tst) && cache == TRUE) {
      ret <- get_info_github(pkg, ver)

      if (is.null(ret)) {

        message(paste0("Version '", ver, "' of '", pkg, "' not found.\n",
                       "Getting latest version."))
        ver <- get_latest_version(pkg)

        ret <- get_info_cran(pkg, ver)
      }
    } else if (based == TRUE) {

      ret <- get_info_description(pkg)

      if (archived)
        ret$Archived <- TRUE

      ret$Cached <- FALSE

    } else {

        ret <- get_info_cran(pkg, ver)
    }

    if (archived == FALSE) {
      dm  <- cranlogs::cran_downloads(pkg, when = "last-month")
      ret$LastMonthDownloads <- sum(dm$count)

      cs <- github_packages(pkg)

      if (!is.na(cs))
        ret$Cached <- TRUE
      else
        ret$Cached <- FALSE

    } else {

      ret$Archived <- as.logical(archived)
    }

    if (based) {

      ret$BasePackage <- TRUE
    } else {
      ret$BasePackge <- FALSE
    }

  }

  return(ret)
}

#' @noRd
get_info_github <- function(pkg, ver) {

  gpkg <- github_package(pkg)

  ret <- NULL

  if (ver %in% names(gpkg$infos)) {
    ret <- gpkg$infos[[ver]]
  }

  return(ret)
}

#' @import utils
#' @import tools
#' @noRd
get_info_cran <- function(pkg, ver) {

  # Construct path
  if (get_latest_version(pkg) == ver ) {

    pth <- file.path(e$MirrorCurrentPath, get_file_name(pkg, ver))

  } else {

    pth <- file.path(e$MirrorArchivePath, pkg, get_file_name(pkg, ver))
  }

  # Unzip package
  package <- unzip_package(pth)

  # Deconstruct path
  bp <- basename(package)
  dn <- dirname(package)

  # Automatic clean up
  on.exit(unlink(dn, recursive = TRUE))

  # Get description
  pd <- utils::packageDescription(bp, dn)

  if (!is.null(pd$Depends))
    pd$Depends <- gsub('\n', ' ', pd$Depends)

  if (!is.null(pd$LinkingTo))
    pd$LinkingTo <- gsub('\n', ' ', pd$LinkingTo)

  if (!is.null(pd$Imports))
    pd$Imports <- gsub('\n', ' ', pd$Imports)

  if (!is.null(pd$Suggests))
    pd$Suggests <- gsub('\n', ' ', pd$Suggests)

  if (!is.null(pd$Enhances))
    pd$Enhances <- gsub('\n', ' ', pd$Enhances)

  # Create package info
  pi <- pinfo(pkg, ver,
              release = as.Date(pd$`Date/Publication`),
              title = pd$Title, description = pd$Description,
              maintainer = pd$Maintainer,
              depends = pd$Depends,
              linkingto = pd$LinkingTo,
              imports = pd$Imports,
              suggests = pd$Suggests,
              enhances = pd$Enhances,
              license = pd$License,
              repository = pd$Repository)

  ## All functions
  code_files <- common::file.find(file.path(package, 'R'), "*.R", up = 0,
                                  down = 2)
  # Collated functions
  coll <- pd$Collate
  if(!is.null(coll)) {
    coll_order <- strsplit(coll, "[[:space:]]")[[1]]
    coll_order <- file.path(package, 'R', gsub("'", '', coll_order))
    code_files <- c(coll_order, setdiff(code_files, coll_order))
  }

  # Read namespace file
  nsf <- tryCatch({parseNamespaceFile(bp, dn)},
                  error = function(er) {
                    message(paste0("Package ", pkg, " version ", ver,
                                   " has no namespace file."))
                   NULL
                  })

  if (is.null(nsf)) {
     nsf <- list()
     nsf$exportPatterns <- "."
  }

  exp <- nsf$exports

  ## Exports
  # exp_list <- nsf[grep('export', names(nsf))]
  # patterns <- grep('Patterns', names(exp_list))
  # pat_list <- unname(unlist(exp_list[patterns]))
  # exp <- unname(unlist(exp_list[-patterns]))
  if(!is.null(nsf$S3methods)) {
    if(nrow(nsf$S3methods)) {
      s3_fun <- paste(nsf$S3methods[,1], nsf$S3methods[,2], sep = '.')
      exp <- c(exp, s3_fun)
    }
  }
  # if(length(pat_list)) {
  #   pat_fun <- fun[unlist(lapply(pat_list, grep, fun))]
  #   exp <- c(exp, pat_fun)
  # }
  # exp <- sort(unique(exp))

  mthds <- NULL
  if (length(nsf$exportMethods) > 0) {
      exp <- c(exp, nsf$exportMethods)
      mthds <- nsf$exportMethods
  }

  exp <- sort(unique(exp))
  # browser()

  # Prepare function list
  funcs <- list()
  for (nm in exp) {
    funcs[[nm]] <- ""  # Just export everything
  }

  exppat <- NULL
  if (length(nsf$exportPatterns) > 0)
    exppat <- nsf$exportPatterns

  # Extract function parameters
  for (fl in code_files) {
    # print(fl)
    tmp <- get_functions(fl, exp, exppat, mthds)
    for (nm in names(tmp)) {
      funcs[[nm]] <- tmp[[nm]]
    }
  }


  pi$Functions <- funcs


  # Need to deal with S4 methods
  # Somehow get method parameters
  # nsf$exportClasses
  # nsf$exportMethods

  return(pi)
}


#' @noRd
get_info_description <- function(pkg) {

  ret <- NULL

  ds <- packageDescription(pkg)

  if (length(ds) > 1) {

    blt <- strsplit(ds$Built, ";", fixed = TRUE)[[1]][3]

    blts <- parse_date(blt)

    ret <- pinfo(pkg, ds$Version,
                release = blts,
                title = ds$Title,
                description = ds$Description,
                maintainer = ds$Maintainer,
                imports = ds$Imports,
                license = ds$License)

    ret$BasePackage <- is_base(pkg)
  }

  return(ret)
}

# Export Patterns:
# ^[^\\.]  (starts with a dot)
# ^[[:alpha:]]+ (starts with an alphabetic character)
# . (everything except newline)

#' @noRd
get_functions <- function(filepath, funcs, exppat, mthds = NULL) {
  # print("Debug 1")
  code <- tryCatch({parse(filepath)},
                   error = function(er) {
                     NULL
                   })
  segs <- as.list(code)
  ret <- list()
  # print("Debug 2")
  for (ii in seq_along(segs)) {
    part <- segs[[ii]]
    if (!is.null(part)) {
      strf <- deparse1(part)
      # print("Debug 3")
      mflag <- FALSE
      if (grepl('setMethod', strf, fixed = TRUE)) {
        # Get method name
        myreg <- regexec('setMethod\\(\\s*"([^"]+)"', strf)

        rm <- regmatches(strf, myreg)

        if (length(rm[[1]]) > 1) {

          nm <- rm[[1]][2]
        } else {
          nm <- "-not found-"
        }
        nm <- sub("\"", "", nm)
        mflag <- TRUE
      } else {
        # Get function name
        tokens <- strsplit(strf, " ", fixed = TRUE)[[1]]
        nm <- gsub("`", "", tokens[[1]], fixed = TRUE)
      }
      # print("Debug 4")
      if (!is.null(exppat))
        pat <- any(suppressWarnings(grepl(exppat, nm)))
      else
        pat <- FALSE

      # print("Debug 5")
      if (nm %in% funcs || pat) {
        ne <- new.env()
        # Not working on methods
        # Functions are fine
        tryCatch({suppressWarnings(eval(part, ne))},
                 error = function(er){ne[[nm]] <- NULL})
        # print("Debug 6")
        if (is.function(ne[[nm]])) {
          prms <- tryCatch({names(formals(ne[[nm]]))},
                           error = function(er) {""})

          ret[[nm]] <- prms
          # print("Debug 7")
        }
      }
    }
  }
  return(ret)
}



# Package Version ---------------------------------------------------------

#' @title Retrieves All Versions of a Package
#' @description
#' The \code{pkg_versions} function queries the CRAN archive,
#' and returns a data frame of package version information for
#' all releases of the package. This function may be used to understand
#' what versions of a package are available. It is often used in
#' conjunction with \code{\link{pkg_info}} to get a general overview
#' of a package.
#' @param pkg A package name to return versions for.  Pass the package
#' name as a quoted string.  This parameter is required.
#' @returns A data frame with one row per package release, showing the
#' version and date of each release. If the package is not found on CRAN,
#' a message will be generated and the function will return NULL.
#' @examples
#' # View package versions
#' pkg_versions("rsample")
#' #    Package Version             FileName    Release   Size
#' # 1  rsample   1.2.1 rsample_1.2.1.tar.gz 2024-03-25 320.9K
#' # 2  rsample   1.2.0 rsample_1.2.0.tar.gz 2023-08-23   321K
#' # 3  rsample   1.1.1 rsample_1.1.1.tar.gz 2022-12-07   318K
#' # 4  rsample   1.1.0 rsample_1.1.0.tar.gz 2022-08-08   306K
#' # 5  rsample   1.0.0 rsample_1.0.0.tar.gz 2022-06-24   268K
#' # 6  rsample   0.1.1 rsample_0.1.1.tar.gz 2021-11-08   274K
#' # 7  rsample   0.1.0 rsample_0.1.0.tar.gz 2021-05-08   274K
#' # 8  rsample   0.0.9 rsample_0.0.9.tar.gz 2021-02-17   269K
#' # 9  rsample   0.0.8 rsample_0.0.8.tar.gz 2020-09-23   261K
#' # 10 rsample   0.0.7 rsample_0.0.7.tar.gz 2020-06-04   248K
#' # 11 rsample   0.0.6 rsample_0.0.6.tar.gz 2020-03-31   299K
#' # 12 rsample   0.0.5 rsample_0.0.5.tar.gz 2019-07-13   297K
#' # 13 rsample   0.0.4 rsample_0.0.4.tar.gz 2019-01-07   254K
#' # 14 rsample   0.0.3 rsample_0.0.3.tar.gz 2018-11-20   254K
#' # 15 rsample   0.0.2 rsample_0.0.2.tar.gz 2017-11-12   339K
#' # 16 rsample   0.0.1 rsample_0.0.1.tar.gz 2017-07-08   180K
#' @family pdiff
#' @export
pkg_versions <- function(pkg) {

  ret <- get_all_versions(pkg)

  return(ret)
}

base_path <- "c:/packages/pkgdiff/tests/testthat"
base_path <- "."

dev <- FALSE  # Set to TRUE for testing


test_that("info1: pinfo() basic functionality.", {

  if (dev) {
    flst <- list()

    flst[["myfunc1"]] <- c("param1", "param2")


    flst[["myfunc2"]] <- c("param1", "param2")


    res <- pinfo("logr", "1.3.7", release = as.Date("2022-01-01"),
                    title = "My title", description = "My description",
                    maintainer = "David Bosak",
                    depends = "R (>= 4.0)", linkingto = "something",
                    suggests = "common", enhances = NULL, license = "CC0",
                    repository = "https://cran.com",
                    functions = flst)

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$PackageName, "logr")
    expect_equal(res$Version, "1.3.7")
    expect_equal(res$Title, "My title")
    expect_equal(res$Maintainer, "David Bosak")
    expect_equal(res$Suggests, "common")
    expect_equal(res$License, "CC0")
    expect_equal(res$Release, as.Date("2022-01-01"))
    expect_equal(length(res$Functions), 2)

  } else {
    expect_equal(TRUE, TRUE)
  }

})



test_that("info2: pkg_info() default params github.", {

  if (dev) {
    res <- pkg_info("logr")

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "logr")
    expect_equal(is.null(res$Version), FALSE)

  } else {
    expect_equal(TRUE, TRUE)
  }


})

test_that("info3: pkg_info() version params github.", {

  if (dev) {
    res <- pkg_info("logr", "1.3.5")

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "logr")
    expect_equal(res$Version, "1.3.5")
  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("info4: get_functions() works as expected.", {

  if (dev) {
    fs <- c("descriptions", "descriptions<-")

    pth <- file.path(base_path, "data/descriptions.R")

    res <- get_functions(pth, fs, NULL)

    expect_equal(is.null(res), FALSE)
    expect_equal(length(res), 2)
    expect_equal(is.null(res[["descriptions"]]), FALSE)
    expect_equal(res[["descriptions"]], "x")

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("info5: pkg_info() default params cran.", {


  if (dev) {
    res <- pkg_info("ards")

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "ards")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)

    res <- pkg_info("Matrix")

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "Matrix")
    expect_equal(is.null(res$Version), FALSE)
    #expect_equal(length(res$Functions) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("info6: pkg_info() cache parameter, exportPatterns, and special cases.", {

  if (dev) {

    # Ignore cache and take directly from CRAN
    # Old versions had no functions.  Was using exportPattern ^[^\\.]
    res <- pkg_info("ggplot2", "0.8.9", cache = FALSE)

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "ggplot2")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)

    # Had no functions.  Was using export pattern ^[[:alpha:]]+
    res <- pkg_info("ccRemover", "latest", cache = FALSE)

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "ccRemover")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)

    # Had no functions.  Not sure what problem was.  Is working now.
    res <- pkg_info("digest", "latest", cache = FALSE)

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "digest")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)

    # R6 package.  Appears to be working good.
    # All exported functions captured.
    res <- pkg_info("tinkr", "latest", cache = FALSE)

    res

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "tinkr")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)
    expect_equal(is.null(res$Functions[["yarn"]]), FALSE)

    # S4 package.
    res <- pkg_info("Matrix", "latest", cache = FALSE)

    print(res, verbose = TRUE)

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "Matrix")
    expect_equal(is.null(res$Version), FALSE)
    # expect_equal(length(res$Functions) > 0, TRUE)
    # expect_equal(is.null(res$Functions[["Schur"]]), FALSE)
    # expect_equal(length(res$Functions[["Schur"]]) > 0, TRUE)

    # Special characters causing errors
    res <- pkg_info("R.oo", "1.4.2", cache = FALSE)

    print(res, verbose = TRUE)

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "R.oo")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)

    # Errors creating package cache
    res <- pkg_info("xtable", "1.0-1", cache = FALSE)

    print(res, verbose = TRUE)

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "xtable")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)

    # Errors creating package cache
    res <- pkg_info("gtools", "2.0.7", cache = FALSE)

    print(res, verbose = TRUE)

    expect_equal("pinfo" %in% class(res), TRUE)
    expect_equal(res$Package, "gtools")
    expect_equal(is.null(res$Version), FALSE)
    expect_equal(length(res$Functions) > 0, TRUE)


  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("info7: print.pinfo() works as expected.", {

  if (dev) {

    res <- pkg_info("logr")

    print(res)


    print(res, verbose = TRUE)

  }

  expect_equal(TRUE, TRUE)

})


test_that("info8: pkg_info() handles archived packages.", {

  if (dev) {
    res <- pkg_info("grid")

    res$Archived

    expect_equal(res$Archived, TRUE)

    res <- pkg_info("maptools")

    res

    expect_equal(res$Archived, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("info9: pkg_cache() works as expected.", {

  if (dev) {

    # All exist
    res <- pkg_cache(c("procs", "logr", "libr"))

    res

    expect_equal(nrow(res), 3)
    expect_equal(any(is.na(res$Version)), FALSE)
    expect_equal("POSIXct" %in% class(attr(res, "LastUpdated")), TRUE)

    # One doesn't exist
    res <- pkg_cache(c("procs", "forker", "libr"))

    res

    expect_equal(nrow(res), 3)
    expect_equal(any(is.na(res$Version)), TRUE)

    # Only doesn't exist
    res <- pkg_cache("forker")

    res

    expect_equal(nrow(res), 1)
    expect_equal(any(is.na(res$Version)), TRUE)


    # All
    res <- pkg_cache()

    res
    expect_equal(nrow(res) > 10, TRUE)
    expect_equal(any(is.na(res$Version)), FALSE)

  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("info10: pkg_versions() works as expected.", {


  if (dev) {

    res <- pkg_versions("common")

    res
    expect_equal(is.null(res), FALSE)
    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("info11: pkg_info() edge cases.", {

  if (dev) {

    # Not on local machine, not archived
    # Therefore no current version
    res <- pkg_info("pharmaRTF")

    res

    expect_equal(is.null(res), FALSE)

    # Local version ahead of CRAN
    res <- pkg_info("pkgdiff")

    expect_equal(is.null(res), FALSE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

# Not doing good with S4 classes
# pth <- "https://cran.r-project.org/src/contrib/Matrix_1.7-2.tar.gz"
#
# library(packageDiff)
# res2 <- packageDiff::pkgInfo(pth)
#
# length(res$Functions)
# length(res2$ExportedFunctions)
#
# length(res2$FormalArgs)

# mym <- ' setMethod("all.equal", c(target = "numLike", current = "abIndex"),
#           function(target, current, ...) all.equal.abI(as(target, "abIndex"),
#                                                        current, ...))'
#
# grepl("setMethod", mym, fixed = TRUE)
#
# myreg <- regexec('setMethod\\(\\s*"([^"]+)"', mym)
#
# rm <- regmatches(mym, myreg)
#
# if (length(rm[[1]]) > 1) {
#
#   nm <- rm[[1]][2]
# }
#
# formals("all.equal")

#' @title Compare Package Versions
#' @description
#' The \code{pkg_diff} function is used to compare two different versions of
#' a package.  The function returns an object that contains the differences.
#' Differences include any functions added or removed, and any function parameters
#' added or removed. The object may
#' be printed directly, or stored in a variable and examined programmatically.
#' @details
#' The difference operation only compares functions that are exported
#' from the package. It does not compare internal functions or function documentation.
#' Exported functions are identified in the package namespace.
#'
#' The \code{pkg_diff} function can compare any two versions of a package.
#' They do not need to be consecutive. They must, however, exist in the CRAN
#' archive.
#'
#' If the package versions are included in the \strong{pkgdiff} cache, the
#' function will pull them from there.  Otherwise, they will be pulled from the
#' CRAN mirror.  The function does not access any packages on the local system.
#' @param pkg The package name.
#' @param v1 The earlier package version. Pass the version value as a quoted string.
#' Default is "current", which means the function will look up the currently
#' installed version.
#' @param v2 The later package version. Pass the version value as a quoted string.
#' The default is "latest", which is the
#' latest version of the package available on CRAN.
#' @returns An object of class "pdiff", which contains information regarding
#' the differences between two versions of a package.  The object includes
#' a list of functions that were added, a list of function parameters that were added,
#' a list of functions removed, and a list of function parameters that
#' were removed.  The object also contains some general information like
#' the package versions examined, and a TRUE or FALSE flag indicating whether
#' there were any breaking changes.
#' @examples
#' # View package stability
#' pkg_stability("rsample")
#' # # A stability score: rsample package
#' # - Age: 7.58 years
#' # - Score: 80.3
#' # - Assessment: Somewhat Unstable
#' # - Version Range: 0.0.1/1.2.1
#' # - Release Range: 2017-07-08/2024-03-25
#' # - Release Count: 16
#' # - Breaking Releases: 6
#' # - Data:
#' #   Package Version             FileName    Release   Size  AF AP RF RP BC  TF
#' # 1  rsample   1.2.1 rsample_1.2.1.tar.gz 2024-03-25 320.9K   1  0  0  0  0 401
#' # 2  rsample   1.2.0 rsample_1.2.0.tar.gz 2023-08-23   321K  43  6  1  3  1 400
#' # 3  rsample   1.1.1 rsample_1.1.1.tar.gz 2022-12-07   318K  37  5  0  0  0 358
#' # 4  rsample   1.1.0 rsample_1.1.0.tar.gz 2022-08-08   306K  72  1  0  0  0 321
#' # 5  rsample   1.0.0 rsample_1.0.0.tar.gz 2022-06-24   268K   1  0  1  0  1 249
#' # 6  rsample   0.1.1 rsample_0.1.1.tar.gz 2021-11-08   274K   4  1  0  1  1 249
#' # 7  rsample   0.1.0 rsample_0.1.0.tar.gz 2021-05-08   274K   0  5  0  0  0 245
#' # 8  rsample   0.0.9 rsample_0.0.9.tar.gz 2021-02-17   269K  19  0  6  0  1 245
#' # 9  rsample   0.0.8 rsample_0.0.8.tar.gz 2020-09-23   261K  59  0  0  0  0 232
#' # 10 rsample   0.0.7 rsample_0.0.7.tar.gz 2020-06-04   248K 101  1  0  0  0 173
#' # 11 rsample   0.0.6 rsample_0.0.6.tar.gz 2020-03-31   299K   7  2  0  0  0  72
#' # 12 rsample   0.0.5 rsample_0.0.5.tar.gz 2019-07-13   297K   4  5  0  0  0  65
#' # 13 rsample   0.0.4 rsample_0.0.4.tar.gz 2019-01-07   254K   0  0  4  0  1  61
#' # 14 rsample   0.0.3 rsample_0.0.3.tar.gz 2018-11-20   254K   7  0  1  0  1  65
#' # 15 rsample   0.0.2 rsample_0.0.2.tar.gz 2017-11-12   339K  24  0  0  0  0  59
#' # 16 rsample   0.0.1 rsample_0.0.1.tar.gz 2017-07-08   180K  35 35  0  0  0  35
#'
#' # Examine differences between versions
#' pkg_diff("rsample", "1.1.1", "1.2.0")
#' # # A difference object: rsample package
#' # - Comparing: v1.1.1/v1.2.0
#' # - Breaking Changes: TRUE
#' # - Added Functions:
#' #   - analysis.default()
#' #   - analysis.initial_validation_split()
#' #   - analysis.rsplit()
#' #   - analysis.val_split()
#' #   - assessment.default()
#' #   - assessment.initial_validation_split()
#' #   - assessment.rsplit()
#' #   - assessment.val_split()
#' #   - dim.initial_validation_split()
#' #   - group_initial_validation_split()
#' #   - initial_validation_split()
#' #   - initial_validation_time_split()
#' #   - int_bca.bootstraps()
#' #   - int_pctl.bootstraps()
#' #   - int_t.bootstraps()
#' #   - print.initial_validation_split()
#' #   - testing.default()
#' #   - testing.initial_validation_split()
#' #   - testing.rsplit()
#' #   - testing.val_split()
#' #   - training.default()
#' #   - training.initial_validation_split()
#' #   - training.rsplit()
#' #   - training.val_split()
#' #   - validation()
#' #   - validation.default()
#' #   - validation.initial_validation_split()
#' #   - validation.val_split()
#' #   - validation_set()
#' #   - vec_cast.data.frame.validation_set()
#' #   - vec_cast.tbl_df.validation_set()
#' #   - vec_cast.validation_set.data.frame()
#' #   - vec_cast.validation_set.tbl_df()
#' #   - vec_cast.validation_set.validation_set()
#' #   - vec_ptype_abbr.group_initial_validation_split()
#' #   - vec_ptype_abbr.initial_validation_split()
#' #   - vec_ptype_abbr.validation_set()
#' #   - vec_ptype2.data.frame.validation_set()
#' #   - vec_ptype2.tbl_df.validation_set()
#' #   - vec_ptype2.validation_set.data.frame()
#' #   - vec_ptype2.validation_set.tbl_df()
#' #   - vec_ptype2.validation_set.validation_set()
#' #   - vec_restore.validation_set()
#' # - Added Parameters:
#' #   - int_pctl(): ...
#' #   - int_t(): ...
#' #   - testing(): ...
#' #   - tidy.nested_cv(): unique_ind
#' #   - tidy.rset(): unique_ind
#' #   - training(): ...
#' # - Removed Functions:
#' #   - gather()
#' # - Removed Parameters:
#' #   - int_bca(): statistics, alpha, .fn
#' #   - int_pctl(): statistics, alpha
#' #   - int_t(): statistics, alpha
#' @family pdiff
#' @export
pkg_diff <- function(pkg, v1 = "current",
                     v2 = "latest") {

  #browser()

  d <- structure(list(), class = c("pdiff", "list"))

  v1_diff_info <- NULL
  v2_diff_info <- NULL

  # print("Debug 1")

  if ("pinfo" %in% class(v1)) {
    v1_diff_info <- v1
    v1 <- v1$Version
  } else if (v1 == "current") {
    v1 <- get_current_version(pkg)

    if (length(v1) == 0) {
      stop(paste0("There is no current version of the package '", pkg, "'."))
    }
  }

  # print("Debug 2")

  # Collect data
  if ("pinfo" %in% class(v2)) {
    v2_diff_info <- v2
    v2 <- v2$Version
  }

  # print("Debug 3")

  v2data <- get_latest_data(pkg)
  vLatest <- v2data$Version[[1]]

  if (v2 == "latest") {
    v2 <- vLatest
  }
  # print("Debug 4")

  # Check that versions exist
  avers <- get_all_versions(pkg)

  if (!v1 %in% avers$Version) {
    stop("Version '", v1, "' of package '", pkg, "' does not exist on CRAN.")
  }

  if (!v2 %in% avers$Version) {
    stop("Version '", v2, "' of package '", pkg, "' does not exist on CRAN.")
  }

  if (is.null(v1_diff_info) && is.null(v2_diff_info)) {
    infos <- get_fastest_infos(pkg, v1, v2)
    v1_diff_info <- infos[[v1]]
    v2_diff_info <- infos[[v2]]
  }
  if (is.null(v1_diff_info)) {
    v1_diff_info <- pkg_info(pkg, v1, cache = FALSE)
  }
  if (is.null(v2_diff_info)) {
    v2_diff_info <- pkg_info(pkg, v2, cache = FALSE)
  }

  # print("Debug 5")

  if (is.null(v1_diff_info)) {

    stop(paste0("Could not retrieve package ", pkg, " v",  v1))
  }

  if (is.null(v2_diff_info)) {

    stop(paste0("Could not retrieve package ", pkg, " v",  v2))
  }

  if (!is.null(v1_diff_info) && !is.null(v2_diff_info)) {


    # Get deprecated functions
    depf <- get_removed_functions(v1_diff_info, v2_diff_info)

    # print("Debug 6")

    # Get deprecated parameters
    depp <- get_removed_parameters(v1_diff_info, v2_diff_info)

    # print("Debug 7")

    # Get breaking changes
    if (length(depf) > 0 || length(depp) > 0)
      bc <- TRUE
    else
      bc <- FALSE

    addf <- get_added_functions(v1_diff_info, v2_diff_info)
    addp <- get_added_parameters(v1_diff_info, v2_diff_info)
    af <- get_all_functions(v1_diff_info, v2_diff_info)

    # print("Debug 8")

    # Populate pdiff object
    d$PackageName <- pkg
    d$Version1 <- v1
    d$Version2 <- v2
    d$Version1DiffInfo <- v1_diff_info
    d$Version2DiffInfo <- v2_diff_info
    d$AddedFunctions <- addf
    d$AddedParameters <- addp
    d$RemovedFunctions <- depf
    d$RemovedParameters <- depp
    d$BreakingChanges <- bc
    d$AllFunctions <- af

  }

  return(d)

}



#' @title Print a Package Difference Object
#' @description
#' Print routine for a package difference object of class "pdiff".
#' @param x The package difference to print
#' @param ... Follow-on parameters to the print function
#' @param verbose Whether to print all added and removed functions
#' and function parameters.  Default is TRUE.  If FALSE, the function
#' will print removed functions and parameters, but only print
#' counts of added functions and parameters.
#' @returns The package difference object, invisibly.
#' @examples
#' # View versions
#' pkg_versions("rvest")
#' #    Package Version           FileName    Release Size
#' # 1    rvest   1.0.4 rvest_1.0.4.tar.gz 2024-02-12 113K
#' # 2    rvest   1.0.3 rvest_1.0.3.tar.gz 2022-08-20  92K
#' # 3    rvest   1.0.2 rvest_1.0.2.tar.gz 2021-10-17  92K
#' # 4    rvest   1.0.1 rvest_1.0.1.tar.gz 2021-07-26  92K
#' # 5    rvest   1.0.0 rvest_1.0.0.tar.gz 2021-03-09  92K
#' # 6    rvest   0.3.6 rvest_0.3.6.tar.gz 2020-07-25 2.7M
#' # 7    rvest   0.3.5 rvest_0.3.5.tar.gz 2019-11-09 1.1M
#' # 8    rvest   0.3.4 rvest_0.3.4.tar.gz 2019-05-15 1.6M
#' # 9    rvest   0.3.3 rvest_0.3.3.tar.gz 2019-04-11 1.6M
#' # 10   rvest   0.3.2 rvest_0.3.2.tar.gz 2016-06-17 1.5M
#' # 11   rvest   0.3.1 rvest_0.3.1.tar.gz 2015-11-11 1.5M
#' # 12   rvest   0.3.0 rvest_0.3.0.tar.gz 2015-09-23 1.5M
#' # 13   rvest   0.2.0 rvest_0.2.0.tar.gz 2015-01-01 5.2M
#' # 14   rvest   0.1.0 rvest_0.1.0.tar.gz 2014-11-22 5.1M
#'
#' # Find differences in latest version
#' res <- pkg_diff("rvest", "1.0.3", "1.0.4")
#'
#' # View differences
#' print(res)
#' # # A difference object: rvest package
#' # - Comparing: v1.0.3/v1.0.4
#' # - Breaking Changes: FALSE
#' # - Added Functions:
#' #   - html_element.LiveHTML()
#' #   - html_elements.LiveHTML()
#' #   - html_table.LiveHTML()
#' #   - LiveHTML()
#' #   - read_html_live()
#' @family pdiff
#' @import crayon
#' @export
print.pdiff <- function(x, ..., verbose = TRUE) {


    grey60 <- crayon::make_style(grey60 = "#999999")
    cat(grey60(paste0("# A difference object: ",
                 as.character(x$PackageName), " package\n")))

    if (!is.null(x$Version1))
      cat(paste0("- Comparing: ", "v", x$Version1, "/v", x$Version2, "\n"))

    if (!is.null(x$BreakingChanges))
      cat(paste0("- Breaking Changes: ", as.character(x$BreakingChanges), "\n"))


    if (verbose == FALSE) {

      if (!is.null(x$AddedFunctions))
        cat(paste0("- Added Functions: ", length(x$AddedFunctions), "\n"))

      if (!is.null(x$AddedParameters))
        if (length(x$AddedParameters) > 0)
          cat(paste0("- Added Parameters: ", length(x$AddedParameters), "\n"))

    } else {


      if (!is.null(x$AddedFunctions)) {
        if (length(x$AddedFunctions > 0)) {
          cat(paste0("- Added Functions: \n"))

          for (nm in x$AddedFunctions) {
            cat(paste0("  - ",  nm, "()\n"))
          }
        }
      }

      if (!is.null(x$AddedParameters)) {
        if (length(x$AddedParameters) > 0) {
          cat(paste0("- Added Parameters: \n"))

          nms <- names(x$AddedParameters)
          for (nm in nms) {
            parms <- paste(x$AddedParameters[[nm]], collapse = ", ")
            cat(paste0("  - ",  nm, "(): ", parms, "\n"))
          }
        }
      }
    }

    if (!is.null(x$RemovedFunctions)) {
      cat(paste0("- Removed Functions: \n"))

      for (nm in x$RemovedFunctions) {
        cat(paste0("  - ",  nm, "()\n"))
      }
    }

    if (!is.null(x$RemovedParameters)) {
      if (length(x$RemovedParameters) > 0) {
        cat(paste0("- Removed Parameters: \n"))

        nms <- names(x$RemovedParameters)
        for (nm in nms) {
          parms <- paste(x$RemovedParameters[[nm]], collapse = ", ")
          cat(paste0("  - ",  nm, "(): ", parms, "\n"))
        }
      }
    }

  invisible(x)
}

# @title View Package Difference Details
# @description The \strong{view_details} function displays package differences
# in the RStudio viewer, so they can be examined in detail. Package information
# is taken from CRAN.
# @param diff A package difference object of class "pdiff".  This object is
# returned from \code{\link{pkg_diff}}.
# @param docs Whether to include the function documentation
# in the viewer comparison.  Default is TRUE.
# @return A package version comparison is displyed in the RStudio viewer.
# @family pdiff
# @export
# view_details <- function(diff, docs = TRUE) {
#
#   if (!"pdiff" %in% class(diff)) {
#
#     stop("Parameter 'diff' must be an object of class 'pdiff'.")
#   }
#
#   d1 <- diff$Version1DiffInfo
#   d2 <- diff$Version2DiffInfo
#
#
#   pkgDiff(d1, d2, doc = docs, )
#
# }

dev <- FALSE

test_that("diff1: pkg_diff() basic functionality.", {

  if (dev) {

    res <- pkg_diff("logr", "1.3.7", "1.3.8")


    expect_equal("pdiff" %in% class(res), TRUE)
    expect_equal(res$PackageName, "logr")
    expect_equal(res$Version1, "1.3.7")
    expect_equal(res$Version2, "1.3.8")

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("diff2: pkg_diff() many changes.", {

  if (dev) {

    res <- pkg_diff("logr", "1.3.4", "1.3.7")


    expect_equal("pdiff" %in% class(res), TRUE)
    expect_equal(res$PackageName, "logr")
    expect_equal(res$Version1, "1.3.4")
    expect_equal(res$Version2, "1.3.7")
    # expect_equal(nchar(res$Version1Path) > 0, TRUE)
    # expect_equal(nchar(res$Version2Path) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("diff3: pkg_diff() breaking changes.", {

  if (dev) {

    # Lot of changes
    res <- pkg_diff("admiral", "1.1.0", "1.2.0")

    expect_equal(res$BreakingChanges, TRUE)
    expect_equal(length(res$RemovedFunctions) > 0, TRUE)
    expect_equal(length(res$RemovedParameters) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})



# test_that("diff4: view_details() works.", {
#
#   if (dev) {
#
#     # Not many changes
#     res <- pkg_diff("logr", "1.3.7", "1.3.8")
#
#
#     view_details(res)
#
#     # Lot of changes
#     res <- pkg_diff("admiral", "1.1.0", "1.2.0")
#
#
#     view_details(res)
#
#     # No error is good
#     expect_equal(TRUE, TRUE)
#
#   } else {
#
#     expect_equal(TRUE, TRUE)
#   }
# })



test_that("diff5: pkg_diff() check printing.", {


  if (dev) {

    res <- pkg_diff("logr", "1.3.6", "1.3.7")

    print(res)

    res <- pkg_diff("admiral", "0.12.3", "1.0.0")

    print(res)

  }

  # If no errors, it is good
  expect_equal(TRUE, TRUE)


})

test_that("diff7: pkg_diff_by_version() old version.", {

  if (dev) {

    res <- pkg_diff("logr", "1.2.9", "1.3.0")

    expect_equal("pdiff" %in% class(res), TRUE)
    expect_equal(res$PackageName, "logr")
    expect_equal(res$Version1, "1.2.9")
    expect_equal(res$Version2, "1.3.0")

  } else {
    expect_equal(TRUE, TRUE)
  }


})



test_that("diff8: pkg_diff() parameter checks.", {

  expect_error(pkg_diff("logxx", "1.3.7", "1.3.8"))

  expect_error(pkg_diff("logr", "0.0.0", "1.3.8"))

  expect_error(pkg_diff("logr", "1.3.7", "0.0.0"))

  # No error OK
  # expect_error(pkg_diff("logr", "1.3.7", "1.3.6"))
  #
  # pkg_diff("logr", "1.3.6", "1.3.7")

})


test_that("diff9: pkg_diff() with pkgInfo works.", {

  if (dev) {

     inf <- pkg_info("logr", "1.0.4")

     df <- pkg_diff("logr", "1.0.3", inf)

     expect_equal(is.null(df), FALSE)
     expect_equal("pdiff" %in% class(df), TRUE)
     expect_equal(df$Version1, "1.0.3")
     expect_equal(df$Version2, "1.0.4")



     inf <- pkg_info("logr", "1.0.3")

     df <- pkg_diff("logr", inf, "1.0.4")

     df
     expect_equal(is.null(df), FALSE)
     expect_equal("pdiff" %in% class(df), TRUE)
     expect_equal(df$Version1, "1.0.3")
     expect_equal(df$Version2, "1.0.4")

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("diff10: print.pdiff() works as expected.", {

  if (dev) {

    res <- pkg_diff("fmtr", "1.5.7", "1.5.8")

    print(res)

    print(res, verbose = FALSE)

    expect_equal(TRUE, TRUE)

    res <- pkg_diff("fmtr", "1.5.5", "1.5.7")

    print(res)

    print(res, verbose = FALSE)

    expect_equal(TRUE, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

#' @title Collect Stability Data for a Package
#' @param pkgs A character vector of one or more package names.
#' @param releases An integer indicating the number of releases to collect
#' stability data for. For example, \code{releases = 10} will return stability
#' data for the last 10 releases of the package. Default is NULL, which means
#' the function will return data for all releases.
#' @param months An integer indicating the number of months back to collect
#' stability data for.  For example, \code{months = 24} will collect
#' stability data for the previous 2 years.  Default is NULL, meaning there is
#' no limitation on the number of release months, and the function will collect
#' data from all releases.
#' @family stability
#' @import common
#' @noRd
get_stability_data <- function(pkgs, releases = NULL, months = NULL) {


  ret <- NULL
  for (pkg in pkgs) {

    obj <- pkg_stability(pkg, releases = releases, months = months)

    dat <- obj$StabilityData

    if (!is.null(dat)) {
      if (is.null(ret))
        ret <- dat
      else
        ret <- rbind(ret, dat)
    }

  }


  return(ret)

}


#' @title Calculate Stability Score for a Package
#' @description
#' The \code{pkg_stablity} function calculates a stability score for
#' a specified package.  The score is calculated as the percentage of
#' releases with no breaking changes.  A breaking change is defined as
#' either a removed function or removed function parameter. The object
#' also returns the raw data from which the stability score is calculated.
#' This data can be useful for review or even custom analytics.
#' @details
#' The \code{pkg_stability} function is used to get an overall feel for the
#' stability of a package.  The stability score and other information returned
#' by the function are based on data retrieved from CRAN.  The function
#' compares each version of a package and determines if any functions or
#' function parameters have been removed from the previous release.
#' If a release removes functions or parameters contained in the previous release,
#' it is flagged as a "breaking release" or "breaking change".
#'
#' The stability score is calculated as the percentage of non-breaking releases.
#' For example, if a package has 10 releases, and one breaking release,
#' the stability score will be .9, or 90%.
#'
#' The stability assessment is a categorization of the stability score. The
#' aim of the assessment is to provide a general evaluation of the package,
#' whether it is stable or unstable.  The
#' assessment has 5 categories:
#' \itemize{
#'   \item {\strong{Perfect}: Stability score of 100.  Package has never
#'   had a breaking release in its entire history.
#'   }
#'   \item {\strong{Very Stable}: Stability score between 95 and 100. Package
#'   has had a breaking release, but very rarely: less than 1 in 20 releases.
#'   }
#'   \item {\strong{Stable}: Stability score between 90 and 95.  The package
#'   has had some breaking releases, but it is still rather rare: less than
#'   1 in 10 releases.
#'   }
#'   \item {\strong{Somewhat unstable}: Stability score between 80 and 90. The
#'   package sometimes has a breaking release, at a rate of about 1 in 5.
#'   }
#'   \item {\strong{Unstable}: Stability score below 80.  The package breaks
#'   more frequently than every 5th release.  This frequency of breaking releases
#'   is considered unstable.
#'   }
#' }
#' As can be seen above, the stability assessment is weighted heavily toward
#' the high end.  This categorization is intentionally designed to encourage
#' package stability, and discourage the number of breaking changes.
#'
#' @section Stability Data: Stability calculations are based on a
#' data frame of stability information gathered from each package release.
#' The data frame has one row per release.  Each row
#' contains comparison information against the prior release.  The
#' columns are organized as follows:
#' \itemize{
#'   \item {\strong{Package}: The package name.
#'   }
#'   \item {\strong{Version}: The version number of the release.
#'   }
#'   \item {\strong{FileName}: The file name of the package.
#'   }
#'   \item {\strong{Release}: The release date of the package.
#'   }
#'   \item {\strong{Size}: The size of the package file on disk.
#'   }
#'   \item {\strong{AF}: The number of functions added from the previous release.
#'   }
#'   \item {\strong{AP}: The number of functions that had parameters added from the
#'   previous release.
#'   }
#'   \item {\strong{RF}: The number of functions removed from the previous release.
#'   }
#'   \item {\strong{RP}: The number of functions that had parameters removed from
#'   the previous release.
#'   }
#'   \item {\strong{BC}: Whether the release had any breaking changes.
#'   Breaking changes are removed functions or removed parameters.  Values
#'   are zero or one.  The value one (1) means the release had breaking changes.
#'   The value zero (0) means the release had no breaking changes.
#'   }
#'   \item {\strong{TF}: The total number of functions in the package.
#'   }
#' }
#' The above data will be included in the function print out.  It can also be
#' accessed on the returned object from the "data" list item, i.e.
#' \code{obj$data}.
#' @section Disclaimers: Note that a "breaking release" does not factor
#' in the number of
#' functions deprecated. One deprecated function counts the same as 10 deprecated
#' functions within a release.  The methodology also does not differentiate
#' between removed
#' functions and removed parameters. They are both considered "breaking", and
#' counted equally.
#'
#' The methodology also does not factor in
#' changed functions.  The reason is simply that it is much more difficult to
#' identify whether or not a changed function is breaking.  That is to say,
#' you may still have breaking changes that are not identified by the
#' \strong{pkgdiff} package.
#' @section Performance:  Performance of the \code{pkg_stability} function
#' can vary greatly depending on the package selected.  The information
#' for some packages has been cached on Github, and can be retrieved very
#' quickly.  For packages that have not been cached, the function must
#' download each version of the package from the CRAN archive, and compare
#' consecutive versions.  This process can take considerable time.
#'
#' The
#' most popular packages on CRAN have been cached.  If there is a package
#' you query frequently that has not been cached, please submit
#' an issue to the \strong{pkgdiff} issue log on Github. In the issue description
#' simply request that the package be added to the cache.
#'
#' For more information on the package cache, see \code{vignette("pkgdiff-cache")}.
#'
#' For more information about how the stability score and assessment are
#' calculated, see \code{vignette("pkgdiff-stability")}.
#' @param pkg The name of the package.
#' @param releases An integer indicating the number of releases to collect
#' stability data for. For example, \code{releases = 10} will return stability
#' data for the last 10 releases of the package. Default is NULL, which means
#' the function will return data for all releases.
#' @param months An integer indicating the number of months back to collect
#' stability data for.  For example, \code{months = 24} will collect
#' stability data for the previous 2 years.  Default is NULL, meaning there is
#' no limitation on the number of release months, and the function will collect
#' data from all releases.
#' @returns An stability score object of class "pdiff_score".  The object
#' contains the stability score, plus other useful information such as
#' the release and version ranges, the number of releases, and number of
#' breaking releases. All of these items can be accessed using dollar sign ($)
#' syntax.
#' @examples
#' # View package stability
#' pkg_stability("curl")
#' # # A stability score: curl package
#' # - Age: 10.33 years
#' # - Score: 98.0
#' # - Assessment: Very Stable
#' # - Version Range: 0.2/6.2.1
#' # - Release Range: 2014-11-20/2025-02-19
#' # - Release Count: 51
#' # - Breaking Releases: 1
#' # - Data:
#' #   Package Version          FileName    Release   Size AF AP RF RP BC TF
#' # 1     curl   6.2.1 curl_6.2.1.tar.gz 2025-02-19 911.7K  0  0  0  0  0 46
#' # 2     curl   6.2.0 curl_6.2.0.tar.gz 2025-01-23 911.3K  0  1  0  0  0 46
#' # 3     curl   6.1.0 curl_6.1.0.tar.gz 2025-01-06   911K  0  1  0  0  0 46
#' # 4     curl   6.0.1 curl_6.0.1.tar.gz 2024-11-14   911K  0  0  0  0  0 46
#' # 5     curl   6.0.0 curl_6.0.0.tar.gz 2024-11-05   911K  1  1  0  1  1 46
#' # 6     curl   5.2.3 curl_5.2.3.tar.gz 2024-09-20   700K  0  0  0  0  0 45
#' # 7     curl   5.2.2 curl_5.2.2.tar.gz 2024-08-26   701K  0  0  0  0  0 45
#' # 8     curl   5.2.1 curl_5.2.1.tar.gz 2024-03-02   700K  0  0  0  0  0 45
#' # 9     curl   5.2.0 curl_5.2.0.tar.gz 2023-12-08   699K  1  0  0  0  0 45
#' # 10    curl   5.1.0 curl_5.1.0.tar.gz 2023-10-02   697K  0  1  0  0  0 44
#' # 11    curl   5.0.2 curl_5.0.2.tar.gz 2023-08-14   697K  0  0  0  0  0 44
#' # 12    curl   5.0.1 curl_5.0.1.tar.gz 2023-06-08   697K  0  1  0  0  0 44
#' # 13    curl   5.0.0 curl_5.0.0.tar.gz 2023-01-12   666K  1  1  0  0  0 44
#' # 14    curl   4.3.3 curl_4.3.3.tar.gz 2022-10-06   655K  0  0  0  0  0 43
#' # 15    curl   4.3.2 curl_4.3.2.tar.gz 2021-06-23   775K  0  0  0  0  0 43
#' # 16    curl   4.3.1 curl_4.3.1.tar.gz 2021-04-30   775K  0  0  0  0  0 43
#' # 17    curl     4.3   curl_4.3.tar.gz 2019-12-02   658K  0  0  0  0  0 43
#' # 18    curl     4.2   curl_4.2.tar.gz 2019-09-24   652K  0  0  0  0  0 43
#' # 19    curl     4.1   curl_4.1.tar.gz 2019-09-16   652K  0  1  0  0  0 43
#' # 20    curl     4.0   curl_4.0.tar.gz 2019-07-22   370K  3  0  0  0  0 43
#' # 21    curl     3.3   curl_3.3.tar.gz 2019-01-10   363K  2  0  0  0  0 40
#' # 22    curl     3.2   curl_3.2.tar.gz 2018-03-28   358K  0  0  0  0  0 38
#' # 23    curl     3.1   curl_3.1.tar.gz 2017-12-12   358K  0  0  0  0  0 38
#' # 24    curl     3.0   curl_3.0.tar.gz 2017-10-06   403K  1  2  0  0  0 38
#' # 25    curl   2.8.1 curl_2.8.1.tar.gz 2017-07-22   400K  0  1  0  0  0 37
#' # 26    curl     2.7   curl_2.7.tar.gz 2017-06-26   397K  1  1  0  0  0 37
#' # 27    curl     2.6   curl_2.6.tar.gz 2017-04-27   396K  0  1  0  0  0 36
#' # 28    curl     2.5   curl_2.5.tar.gz 2017-04-14   396K  1  0  0  0  0 36
#' # 29    curl     2.4   curl_2.4.tar.gz 2017-03-24   394K  2  0  0  0  0 35
#' # 30    curl     2.3   curl_2.3.tar.gz 2016-11-24   391K  0  0  0  0  0 33
#' # 31    curl     2.2   curl_2.2.tar.gz 2016-10-21   389K  0  0  0  0  0 33
#' # 32    curl     2.1   curl_2.1.tar.gz 2016-09-22   388K  0  0  0  0  0 33
#' # 33    curl     2.0   curl_2.0.tar.gz 2016-09-17   389K 10  0  0  0  0 33
#' # 34    curl     1.2   curl_1.2.tar.gz 2016-08-13   287K  0  0  0  0  0 23
#' # 35    curl     1.1   curl_1.1.tar.gz 2016-07-26   277K  0  0  0  0  0 23
#' # 36    curl     1.0   curl_1.0.tar.gz 2016-07-24   277K  0  0  0  0  0 23
#' # 37    curl   0.9.7 curl_0.9.7.tar.gz 2016-04-10   258K  0  0  0  0  0 23
#' # 38    curl   0.9.6 curl_0.9.6.tar.gz 2016-02-17   256K  1  0  0  0  0 23
#' # 39    curl   0.9.5 curl_0.9.5.tar.gz 2016-01-23   255K  1  0  0  0  0 22
#' # 40    curl   0.9.4 curl_0.9.4.tar.gz 2015-11-20   240K  0  1  0  0  0 21
#' # 41    curl   0.9.3 curl_0.9.3.tar.gz 2015-08-25   240K  0  0  0  0  0 21
#' # 42    curl   0.9.2 curl_0.9.2.tar.gz 2015-08-08   239K  1  0  0  0  0 21
#' # 43    curl   0.9.1 curl_0.9.1.tar.gz 2015-07-04   237K  0  0  0  0  0 20
#' # 44    curl     0.9   curl_0.9.tar.gz 2015-06-19   236K  0  0  0  0  0 20
#' # 45    curl     0.8   curl_0.8.tar.gz 2015-06-06   235K  0  0  0  0  0 20
#' # 46    curl     0.7   curl_0.7.tar.gz 2015-05-22   236K  0  0  0  0  0 20
#' # 47    curl     0.6   curl_0.6.tar.gz 2015-05-19   235K 14  2  0  0  0 20
#' # 48    curl     0.5   curl_0.5.tar.gz 2015-02-01    16K  0  0  0  0  0  6
#' # 49    curl     0.4   curl_0.4.tar.gz 2015-01-08    16K  4  0  0  0  0  6
#' # 50    curl     0.3   curl_0.3.tar.gz 2014-12-01   7.5K  1  0  0  0  0  2
#' # 51    curl     0.2   curl_0.2.tar.gz 2014-11-20   5.5K  1  1  0  0  0  1
#'
#' # View package stability
#' pkg_stability("ggplot2")
#' # # A stability score: ggplot2 package
#' # - Age: 17.75 years
#' # - Score: 82.5
#' # - Assessment: Somewhat Unstable
#' # - Version Range: 0.5/3.5.1
#' # - Release Range: 2007-06-01/2024-04-23
#' # - Release Count: 50
#' # - Breaking Releases: 29
#' # - Data:
#' #   Package Version               FileName    Release Size  AF  AP  RF RP BC  TF
#' # 1  ggplot2   3.5.1   ggplot2_3.5.1.tar.gz 2024-04-23 3.4M   0   4   0  0  0 711
#' # 2  ggplot2   3.5.0   ggplot2_3.5.0.tar.gz 2024-02-23 3.4M  37  92  26  6  1 711
#' # 3  ggplot2   3.4.4   ggplot2_3.4.4.tar.gz 2023-10-12 3.0M   0   0   0  0  0 700
#' # 4  ggplot2   3.4.3   ggplot2_3.4.3.tar.gz 2023-08-14 2.1M   0   0   0  0  0 700
#' # 5  ggplot2   3.4.2   ggplot2_3.4.2.tar.gz 2023-04-03 3.0M   3   0   0  1  1 700
#' # 6  ggplot2   3.4.1   ggplot2_3.4.1.tar.gz 2023-02-10 3.0M   2   1   0  0  0 697
#' # 7  ggplot2   3.4.0   ggplot2_3.4.0.tar.gz 2022-11-04 3.0M  28  15   0  0  0 695
#' # 8  ggplot2   3.3.6   ggplot2_3.3.6.tar.gz 2022-05-03 2.9M   0   0   0  1  1 667
#' # 9  ggplot2   3.3.5   ggplot2_3.3.5.tar.gz 2021-06-25 2.9M   0   0   0  0  0 667
#' # 10 ggplot2   3.3.4   ggplot2_3.3.4.tar.gz 2021-06-16 2.9M   8  16   0  1  1 667
#' # 11 ggplot2   3.3.3   ggplot2_3.3.3.tar.gz 2020-12-30 2.9M   0   0   0  0  0 659
#' # 12 ggplot2   3.3.2   ggplot2_3.3.2.tar.gz 2020-06-19 2.9M  17   5   0  1  1 659
#' # 13 ggplot2   3.3.1   ggplot2_3.3.1.tar.gz 2020-05-29 2.9M   0   0   0  0  0 642
#' # 14 ggplot2   3.3.0   ggplot2_3.3.0.tar.gz 2020-03-05 2.9M  68  47   0  1  1 642
#' # 15 ggplot2   3.2.1   ggplot2_3.2.1.tar.gz 2019-08-11 3.1M   0   0   0  0  0 574
#' # 16 ggplot2   3.2.0   ggplot2_3.2.0.tar.gz 2019-06-16 3.0M   3   7   3  0  1 574
#' # 17 ggplot2   3.1.1   ggplot2_3.1.1.tar.gz 2019-04-07 2.7M   0   2   0  0  0 574
#' # 18 ggplot2   3.1.0   ggplot2_3.1.0.tar.gz 2018-10-25 2.7M   5   8   0  0  0 574
#' # 19 ggplot2   3.0.0   ggplot2_3.0.0.tar.gz 2018-07-03 2.7M  89  53   6  4  1 569
#' # 20 ggplot2   2.2.1   ggplot2_2.2.1.tar.gz 2016-12-30 2.1M   0   0   0  0  0 486
#' # 21 ggplot2   2.2.0   ggplot2_2.2.0.tar.gz 2016-11-11 2.1M  22  27  21  2  1 486
#' # 22 ggplot2   2.1.0   ggplot2_2.1.0.tar.gz 2016-03-01 1.5M  11   6   0  2  1 485
#' # 23 ggplot2   2.0.0   ggplot2_2.0.0.tar.gz 2015-12-18 1.5M 159  94 103 33  1 474
#' # 24 ggplot2   1.0.1   ggplot2_1.0.1.tar.gz 2015-03-17 2.2M   0   0   0  0  0 418
#' # 25 ggplot2   1.0.0   ggplot2_1.0.0.tar.gz 2014-05-21 2.2M  12   5   0  0  0 418
#' # 26 ggplot2 0.9.3.1 ggplot2_0.9.3.1.tar.gz 2013-03-02 2.2M   0   1   0  0  0 406
#' # 27 ggplot2   0.9.3   ggplot2_0.9.3.tar.gz 2012-12-05 2.2M   4   4   0  0  0 406
#' # 28 ggplot2 0.9.2.1 ggplot2_0.9.2.1.tar.gz 2012-09-11 2.2M   0   0   0  0  0 402
#' # 29 ggplot2   0.9.2   ggplot2_0.9.2.tar.gz 2012-09-04 2.2M  26  16   8  6  1 402
#' # 30 ggplot2   0.9.1   ggplot2_0.9.1.tar.gz 2012-05-08 9.1M  13  11   0  3  1 384
#' # 31 ggplot2   0.9.0   ggplot2_0.9.0.tar.gz 2012-03-01 1.6M 282   4 149  5  1 371
#' # 32 ggplot2   0.8.9   ggplot2_0.8.9.tar.gz 2010-12-23 2.0M   4   1   2  0  1 238
#' # 33 ggplot2   0.8.8   ggplot2_0.8.8.tar.gz 2010-07-05 2.0M   6   0   2  0  1 236
#' # 34 ggplot2   0.8.7   ggplot2_0.8.7.tar.gz 2010-03-02 2.0M   2   0   0  0  0 232
#' # 35 ggplot2   0.8.6   ggplot2_0.8.6.tar.gz 2010-02-18 2.0M   0   0   0  0  0 230
#' # 36 ggplot2   0.8.5   ggplot2_0.8.5.tar.gz 2009-12-16 2.0M   9   2   4  0  1 230
#' # 37 ggplot2   0.8.4   ggplot2_0.8.4.tar.gz 2009-12-09 2.0M   3   0   2  0  1 225
#' # 38 ggplot2   0.8.3   ggplot2_0.8.3.tar.gz 2009-04-20 2.0M  51   2   6  2  1 224
#' # 39 ggplot2   0.8.2   ggplot2_0.8.2.tar.gz 2009-02-25 2.0M  11   1  10  0  1 179
#' # 40 ggplot2   0.8.1   ggplot2_0.8.1.tar.gz 2008-12-14 1.9M  19   3   3  3  1 178
#' # 41 ggplot2     0.8     ggplot2_0.8.tar.gz 2008-11-21 1.9M  25   6   2  2  1 162
#' # 42 ggplot2     0.7     ggplot2_0.7.tar.gz 2008-10-05 1.9M  45   7  26  3  1 139
#' # 43 ggplot2     0.6     ggplot2_0.6.tar.gz 2008-04-03 1.9M  30   5 125  2  1 120
#' # 44 ggplot2   0.5.7   ggplot2_0.5.7.tar.gz 2008-01-11 1.8M  16   3   3  0  1 215
#' # 45 ggplot2   0.5.6   ggplot2_0.5.6.tar.gz 2007-10-20 1.8M  36   0   5  0  1 202
#' # 46 ggplot2   0.5.5   ggplot2_0.5.5.tar.gz 2007-09-01 1.8M   1   0   0  2  1 171
#' # 47 ggplot2   0.5.4   ggplot2_0.5.4.tar.gz 2007-07-08 1.8M   3   0   1  0  1 170
#' # 48 ggplot2   0.5.2   ggplot2_0.5.2.tar.gz 2007-06-18 1.8M   2   0   0  0  0 168
#' # 49 ggplot2   0.5.1   ggplot2_0.5.1.tar.gz 2007-06-10 1.8M  11   1   5  0  1 166
#' # 50 ggplot2     0.5     ggplot2_0.5.tar.gz 2007-06-01 7.1M 160 160   0  0  0 160
#' @family pdiff
#' @export
pkg_stability <- function(pkg, releases = NULL, months = NULL) {


  d <- structure(list(), class = c("pdiff_score", "list"))

  pk <- github_packages(pkg)
  if (is.na(pk))
    dat <- get_cran_data(pkg, releases, months)
  else
    dat <- get_github_data(pkg, releases, months)

  if (!is.null(dat)) {
    if (nrow(dat) == 1) {

      spn <- elapsed_months( dat$Release[[nrow(dat)]])

      # Populate pdiff_score object
      d$PackageName <- pkg
      d$PackageAge <- sprintf("%0.2f years", spn / 12)
      d$FirstRelease <- dat$Release[[1]]
      d$LastRelease <- dat$Release[[1]]
      d$StabilityScore <- 1
      d$Assessment <- get_stability_assessment(1)
      d$NumReleases <- 1
      d$BreakingReleases <- 0
      d$FirstVersion <- dat$Version[[1]]
      d$LastVersion <- dat$Version[[1]]
      d$AddedFunctions <- 0
      d$AddedParameters <- 0
      d$RemovedFunctions <- 0
      d$RemovedParameters <- 0
      d$StabilityData <- dat

    } else {

      # Calculate stability score
      scr <- stability_score(dat)

      spn <- elapsed_months( dat$Release[[nrow(dat)]])

      # Populate pdiff_score object
      d$PackageName <- pkg
      d$PackageAge <- sprintf("%0.2f years", spn / 12)
      d$FirstRelease <- dat$Release[[nrow(dat)]]
      d$LastRelease <- dat$Release[[1]]
      d$StabilityScore <- scr
      d$Assessment <- get_stability_assessment(scr)
      d$NumReleases <- nrow(dat)
      d$BreakingReleases <- sum(dat$BC, na.rm = TRUE)
      d$FirstVersion <- dat$Version[[nrow(dat)]]
      d$LastVersion <- dat$Version[[1]]
      d$AddedFunctions <- sum(dat$AF, na.rm = TRUE)
      d$AddedParameters <- sum(dat$AP, na.rm = TRUE)
      d$RemovedFunctions <- sum(dat$RF, na.rm = TRUE)
      d$RemovedParameters <- sum(dat$RP, na.rm = TRUE)
      d$StabilityData <- dat

    }
  } else {

    d$PackageName <- pkg
    d$Repository <- "?"
  }

  return(d)
}


#' @title Print a Stability Score Object
#' @description
#' Print routine for a package stability score object of class "pdiff_score".
#' @param x The package stability score to print
#' @param ... Follow-on parameters to the print function
#' @family pdiff
#' @import crayon
#' @returns The stability score object, invisibly.
#' @examples
#' # View package stability
#' res <- pkg_stability("rsample")
#'
#' # Print stability scores
#' print(res)
#' # # A stability score: rsample package
#' # - Age: 7.67 years
#' # - Score: 80.7
#' # - Assessment: Somewhat Unstable
#' # - Version Range: 0.0.1/1.2.1
#' # - Release Range: 2017-07-08/2024-03-25
#' # - Release Count: 16
#' # - Breaking Releases: 6
#' # - Data:
#' #    Package Version             FileName    Release   Size  AF AP RF RP BC  TF
#' # 1  rsample   1.2.1 rsample_1.2.1.tar.gz 2024-03-25   321K   1  0  0  0  0 401
#' # 2  rsample   1.2.0 rsample_1.2.0.tar.gz 2023-08-23   321K  43  6  1  3  1 400
#' # 3  rsample   1.1.1 rsample_1.1.1.tar.gz 2022-12-07   318K  37  5  0  0  0 358
#' # 4  rsample   1.1.0 rsample_1.1.0.tar.gz 2022-08-08   306K  72  1  0  0  0 321
#' # 5  rsample   1.0.0 rsample_1.0.0.tar.gz 2022-06-24   268K   1  0  1  0  1 249
#' # 6  rsample   0.1.1 rsample_0.1.1.tar.gz 2021-11-08   274K   4  1  0  1  1 249
#' # 7  rsample   0.1.0 rsample_0.1.0.tar.gz 2021-05-08   274K   0  5  0  0  0 245
#' # 8  rsample   0.0.9 rsample_0.0.9.tar.gz 2021-02-17   269K  19  0  6  0  1 245
#' # 9  rsample   0.0.8 rsample_0.0.8.tar.gz 2020-09-23   261K  59  0  0  0  0 232
#' # 10 rsample   0.0.7 rsample_0.0.7.tar.gz 2020-06-04   248K 101  1  0  0  0 173
#' # 11 rsample   0.0.6 rsample_0.0.6.tar.gz 2020-03-31   299K   7  2  0  0  0  72
#' # 12 rsample   0.0.5 rsample_0.0.5.tar.gz 2019-07-13   297K   4  5  0  0  0  65
#' # 13 rsample   0.0.4 rsample_0.0.4.tar.gz 2019-01-07   254K   0  0  4  0  1  61
#' # 14 rsample   0.0.3 rsample_0.0.3.tar.gz 2018-11-20   254K   7  0  1  0  1  65
#' # 15 rsample   0.0.2 rsample_0.0.2.tar.gz 2017-11-12   339K  24  0  0  0  0  59
#' # 16 rsample   0.0.1 rsample_0.0.1.tar.gz 2017-07-08   180K  35 35  0  0  0  35
#' @export
print.pdiff_score <- function(x, ...) {


  grey60 <- crayon::make_style(grey60 = "#999999")
  cat(grey60("# A stability score: " %+%
               as.character(x$PackageName) %+% " package\n"))

  if (!is.null(x$Repository)) {

    cat(paste0("- Repository: ", x$Repository, "\n"))

  }

  if (!is.null(x$PackageAge))
    cat(paste0("- Age: ", x$PackageAge, "\n"))

  if (!is.null(x$StabilityScore)) {

    scr <- sprintf("%.1f", x$StabilityScore * 100)
    cat(paste0("- Score: ", scr, "\n"))

  }

  if (!is.null(x$Assessment)) {

    cat(paste0("- Assessment: ", x$Assessment, "\n"))

  }

  if (!is.null(x$FirstRelease))
    cat(paste0("- Version Range: ", x$FirstVersion, "/", x$LastVersion, "\n"))

  if (!is.null(x$FirstRelease))
    cat(paste0("- Release Range: ", x$FirstRelease, "/", x$LastRelease, "\n"))

  if (!is.null(x$NumReleases))
    cat(paste0("- Release Count: ", as.character(x$NumReleases), "\n"))

  if (!is.null(x$BreakingReleases))
    cat(paste0("- Breaking Releases: ", as.character(x$BreakingReleases), "\n"))


  if (!is.null(x$StabilityData)) {
    cat(paste0("- Data: \n"))

    print(x$StabilityData)
  }




  invisible(x)
}



# Get Data ----------------------------------------------------------------



#' @noRd
get_github_data <- function(pkgname, releases = NULL, months = NULL) {

  dat <- NULL
  fl <- paste0(pkgname, ".RData")

  # https://github.com/dbosak01/pkgdiffdata/blob/main/data/common.Rdata
  # https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/data/common.Rdata
  # https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/data/common.Rdata
  # https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/data/common.Rdata

  pth <- file.path("https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/data",
                   fl)

  info <- NULL

  ret <- tryCatch({
    murl <- url(pth)
    info <- get(load(gzcon(murl)))
    close(murl)
    TRUE
  }, warning = function(msg) {
    message(paste("Message:", msg))
    return(FALSE)
  }, error = function(cond) {
    return(FALSE)
  })

  if (TRUE & !is.null(info)) {
    dat <- info$stability

    if (!is.null(releases)) {

      if (nrow(dat) >= releases)
        dat <- dat[seq(1, releases), ]
    }

    if (!is.null(months)) {
      dm <- Sys.Date() - ((months + 1) * 30)

      # Handle situation where last package release date
      # is before the requested release range.
      if (max(dat$Release) < dm) {
        dat <- dat[1, ]
      } else {
        dat <- subset(dat, dat$Release >= dm)
      }
    }

    rownames(dat) <- NULL

  }

  return(dat)
}


#' @noRd
get_cran_data <- function(pkgname, releases = NULL, months = NULL) {

  #browser()

  if (is_base(pkgname)) {

    message(paste0("Package '", pkgname, "' is a Base package."))
    # Base package
    dat <- NULL

  } else {

    # Get latest package data
    ldat <- get_latest_data(pkgname)

    # Get archive data
    adat <- get_archive_versions(pkgname)

    if (!is.null(ldat) || !is.null(adat)) {

      # Initialize data
      if (!is.null(adat) && !is.null(ldat))
        dat <- rbind(ldat, adat)
      else if (!is.null(ldat))
        dat <- ldat
      else if (!is.null(adat))
        dat <- adat

      # Initialize count columns
      dat$AF <- NA
      dat$AP <- NA
      dat$RF <- NA
      dat$RP <- NA
      dat$BC <- NA
      dat$TF <- NA

      # Clear out rownames
      rownames(dat) <- NULL

      # Deal with releases parameter
      if (!is.null(releases)) {

        if (nrow(dat) > releases)
          dat <- dat[seq(1, releases + 1), ]
      }

      if (!is.null(months)) {
        dm <- Sys.Date() - ((months + 1) * 30)

        # Handle situation where last package release date
        # is before the requested release range.
        if (max(dat$Release) < dm) {
          dat <- dat[1, ]
        } else {
          dat <- subset(dat, dat$Release >= dm)
        }
      }

      if (nrow(dat) <= 1) {

        dat <- ldat

        inf <- pkg_info(pkgname, "latest")

        dat$AF <- length(inf$Functions)
        dat$AP <- length(inf$Functions)
        dat$RF <- 0
        dat$RP <- 0
        dat$BC <- 0
        dat$TF <- length(inf$Functions)

      } else {

        # browser()

        # Set up loop counter
        idxs <- seq(1, nrow(dat) - 1)

        # Lag earlier info for performance
        laginfo <- NULL

        for (idx in idxs) {

          # Get version strings
          v2 <- dat[idx, "Version"]
          v1 <- dat[idx + 1, "Version"]

          cat(paste0("Comparing ", pkgname, " ", v1, "/", v2, "\n"))

          if (is.null(laginfo)) {
            diff <- tryCatch({
              pkg_diff(pkgname, v1, v2)
            }, error = function(e) {
              NULL
            })
          } else {
            diff <- tryCatch({
              pkg_diff(pkgname, v1, laginfo)
            }, error = function(e) {
              NULL
            })

          }

          if (!is.null(diff)) {

            if (is.null(diff$AddedFunctions))
              dat[idx, "AF"] <- 0
            else
              dat[idx, "AF"] <- length(diff$AddedFunctions)

            if (is.null(diff$AddedParameters))
              dat[idx, "AP"] <- 0
            else
              dat[idx, "AP"] <- length(diff$AddedParameters)

            if (is.null(diff$RemovedFunctions))
              dat[idx, "RF"] <- 0
            else
              dat[idx, "RF"] <- length(diff$RemovedFunctions)

            if (is.null(diff$RemovedParameters))
              dat[idx, "RP"] <- 0
            else
              dat[idx, "RP"] <- length(diff$RemovedParameters)

            if (dat[idx, "RF"] > 0 | dat[idx, "RP"] > 0)
              dat[idx, "BC"] <- 1
            else
              dat[idx, "BC"] <- 0

            if (is.null(diff$AllFunctions))
              dat[idx, "TF"] <- 0
            else
              dat[idx, "TF"] <- length(diff$AllFunctions)

            laginfo <- diff$Version1DiffInfo

          }


        }

      }


      # Take out first release, as it will not be compared.
      if (!is.null(releases)) {

        if (nrow(dat) > releases)
          dat <- dat[seq(1, releases), ]
      }

      # Special handling for first release
      if (nrow(dat) > 0) {

        idx <- nrow(dat)
        if (is.na(dat[[idx, "TF"]])) {
          if (!is.na(dat[[idx, "Version"]])) {

            v1 <- dat[[idx, "Version"]]

            inf <- get_info_cran(pkgname, v1)
            dat[[idx, "AF"]] <- length(inf$Functions)
            dat[[idx, "AP"]] <- length(get_all_parameters(inf))
            dat[[idx, "RF"]] <- 0
            dat[[idx, "RP"]] <- 0
            dat[[idx, "BC"]] <- 0
            dat[[idx, "TF"]] <- length(inf$Functions)

          }
        }
      }

      # Add labels
      common::labels(dat) <- list(AF = "Added Functions",
                                  AP = "Added Parameters",
                                  RF = "Removed Functions",
                                  RP = "Removed Parameters",
                                  BC = "Breaking Changes",
                                  TC = "Total Functions")
    } else {

      # Package not found
      dat <- NULL
    }
  }

  return(dat)

}



# Assume versions are in proper order.
#' @import common
#' @noRd
get_info_data <- function(pkgname, pkginfos, skip_first = FALSE) {

  # browser()

  # Get table data for all versions
  dat <- get_all_versions(pkgname)

  # Extract desired versions
  versions <- names(pkginfos)

  # Subset table for desired versions
  dat <- subset(dat, dat$Version %in% versions)

  if (length(versions) == 1) {

    inf <- pkginfos[[versions]]

    dat$AF <- length(inf$Functions)
    dat$AP <- length(inf$Functions)
    dat$RF <- 0
    dat$RP <- 0
    dat$BC <- 0
    dat$TF <- length(inf$Functions)

  } else {

    # browser()

    idxs <- seq(1, length(versions) - 1)


    for (idx in idxs) {

      v2info <- pkginfos[[idx]]
      v1info <- pkginfos[[idx + 1]]

      v2 <- v2info$Version
      v1 <- v1info$Version

      cat(paste0("Comparing ", pkgname, " ", v1, "/", v2, "\n"))

      diff <- tryCatch({
        pkg_diff(pkgname, v1info, v2info)
      }, error = function(e) {
        NULL
      })


      if (!is.null(diff)) {

        if (is.null(diff$AddedFunctions))
          dat[idx, "AF"] <- 0
        else
          dat[idx, "AF"] <- length(diff$AddedFunctions)

        if (is.null(diff$AddedParameters))
          dat[idx, "AP"] <- 0
        else
          dat[idx, "AP"] <- length(diff$AddedParameters)

        if (is.null(diff$RemovedFunctions))
          dat[idx, "RF"] <- 0
        else
          dat[idx, "RF"] <- length(diff$RemovedFunctions)

        if (is.null(diff$RemovedParameters))
          dat[idx, "RP"] <- 0
        else
          dat[idx, "RP"] <- length(diff$RemovedParameters)

        if (dat[idx, "RF"] > 0 | dat[idx, "RP"] > 0)
          dat[idx, "BC"] <- 1
        else
          dat[idx, "BC"] <- 0

        if (is.null(diff$AllFunctions))
          dat[idx, "TF"] <- 0
        else
          dat[idx, "TF"] <- length(diff$AllFunctions)


      }

    }

  }

  # Deal with first release
  if (nrow(dat) > 0 && !skip_first) {

    idx <- nrow(dat)
    if (is.na(dat[[idx, "TF"]])) {
      if (!is.na(dat[[idx, "Version"]])) {

        v1 <- dat[[idx, "Version"]]

        inf <- pkginfos[[v1]]
        dat[[idx, "AF"]] <- length(inf$Functions)
        dat[[idx, "AP"]] <- length(get_all_parameters(inf))
        dat[[idx, "RF"]] <- 0
        dat[[idx, "RP"]] <- 0
        dat[[idx, "BC"]] <- 0
        dat[[idx, "TF"]] <- length(inf$Functions)

      }
    }
  }

  # Add labels
  common::labels(dat) <- list(AF = "Added Functions",
                              AP = "Added Parameters",
                              RF = "Removed Functions",
                              RP = "Removed Parameters",
                              BC = "Breaking Changes",
                              TC = "Total Functions")

  rownames(dat) <- NULL

  return(dat)

}


# Utilities ---------------------------------------------------------------

#' @noRd
get_stability_assessment <- function(score) {

  ret <- "Unknown"
  if (!is.nan(score)) {
    if (score == 1) {
      ret <- "Perfect"
    } else if (score >= .95) {
      ret <- "Very Stable"
    } else if (score >= .9) {

      ret <- "Stable"
    } else if (score >= .8) {
      ret <- "Somewhat Unstable"
    } else {

      ret <- "Unstable"
    }
  }

  return(ret)
}

#' @noRd
elapsed_months <- Vectorize(function(start_date) {
  ed <- as.POSIXlt(Sys.Date())
  sd <- as.POSIXlt(start_date)
  ret <- 12 * (ed$year - sd$year) + (ed$mon - sd$mon)
  return(ret)
})

#' @noRd
stability_score <- function(dat) {

  # Calculate elapsed months for each release
  el <- elapsed_months(dat$Release)

  # After 10 years, don't count it
  tm <- ifelse(el > 120, 0, 1)

  # Get breaking releases
  br <- dat$BC

  # Weight each breaking release
  clc <- ifelse(br == 1, el * .01, 0)

  # Don't let weights go over 1
  mx <- ifelse(clc > 1, 1, clc)

  # Get numerator
  nd <- sum(br - mx, na.rm = TRUE)

  if (sum(tm) == 0)
    # Denominator is zero
    # Return 100%
    ret <- 1
  else {

    # Calculate percentage
    ret <- 1 - (nd / sum(tm))
  }

  return(ret)
}

dev <- FALSE  # Set to TRUE when testing

test_that("stability1: get_stability_data() basic functionality.", {

  if (dev) {

    res <- get_stability_data("procs")

    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("stability2: get_stability_data() many changes.", {

  if (dev) {
    res <- get_stability_data("admiral")


    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)


  } else {

    expect_equal(TRUE, TRUE)
  }



})


test_that("stability3: get_stability_data() release limit.", {

  if (dev) {

    res <- get_stability_data("fmtr", releases = 5)


    expect_equal(is.data.frame(res), TRUE)

    # Need one extra for comparison
    expect_equal(nrow(res) == 5, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("stability4: get_stability_data() time limit.", {

  if (dev) {
    res <- get_stability_data("fmtr", months = 12)


    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 1, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }


})

# Need to handle
test_that("stability5: get_stability_data() only one release.", {

  if (dev & FALSE) {
    # No archives
    res <- get_stability_data("defineR")

    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) == 2, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})




test_that("stability6: pkg_stability() basic functionality.", {

  if (dev) {

    res <- pkg_stability("procs")

    res

    expect_equal(res$StabilityScore == 1, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("stability7: pkg_stability() stressed functionality.", {

  if (dev) {

    res <- pkg_stability("admiral", releases = 5)

    res

    expect_equal(res$StabilityScore < 1, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("stability8: get_stability_data() no releases in time period.", {

  if (dev) {

    res <- get_stability_data("tibble", months = 1)

    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) == 1, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})

test_that("stability9: pkg_stability() out of range.", {

  if (dev) {

    res <- pkg_stability("tibble", months = 1)

    expect_equal(res$StabilityScore == 1, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("stability10: get_stability_data() boundry conditions.", {


  if (dev) {

    # Unknown package
    pkgs <- c("definer")


    res <- get_stability_data(pkgs)
    expect_equal(is.null(res), TRUE)


    # No archive data
    pkgs <- c("defineR")

    res <- get_stability_data(pkgs)

    expect_equal(is.data.frame(res), TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("stability11: get_stability_data() multiple packages.", {

  if (dev) {

    pkgs <- c("common", "procs", "libr", "defineR")

    res <- get_stability_data(pkgs)


    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("stability12: get_github_data() works as expected.", {

  if (dev) {

    res <- get_github_data("procs")

    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)

    res <- get_github_data("fmtr")

    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)


    res <- get_github_data("fork2")

    expect_equal(is.null(res), TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("stability13: get_info_data() works as expected.", {

  if (dev) {

    infos <- get_all_infos("procs")

    res <- get_info_data("procs", infos)

    expect_equal(is.null(res), FALSE)
    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)
    expect_equal(ncol(res) == 11, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})



test_that("stability14: pkg_stability() one release.", {

  if (dev) {

    res <- pkg_stability("defineR")

    res

    expect_equal(res$StabilityScore == 1, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }


})


test_that("stability15: pkg_stability() package edge cases.", {

  if (dev) {

    # Unknown package
    # Should not give an error
    res <- pkg_stability("forker")

    res
    expect_equal(is.null(res), FALSE)


    # Score coming out NaN
    res <- pkg_stability("KMsurv")

    res
    expect_equal(res$StabilityScore == 1, TRUE)


    # Only one release
    res <- pkg_stability("ards")

    expect_equal(res$StabilityScore == 1, TRUE)

    # No score because no archive versions.
    # Should still be stability 1.
    # Not sure why it isn't coming out.
    res <- pkg_stability("concatenate")

    expect_equal(res$StabilityScore == 1, TRUE)


    # Error on Base Package, formerly contributed
    res <- pkg_stability("tcltk")

    expect_equal(res$Repository, "?")


    # Error on Base Package, formerly contributed
    res <- pkg_stability("Matrix")

    expect_equal(res$Repository, "?")

  } else {
    expect_equal(TRUE, TRUE)
  }


})

# Screen Scraping ---------------------------------------------------------



#' @import rvest
#' @import utils
#' @noRd
get_latest_data <- function(pkgs,
                                 skip_size = FALSE) {



  if (is.null(pkgs)) {
    stop("pkgs parameter cannot be NULL.")
  }

  ret <- NULL

  refresh_package_lists()

  lst <- e$AvailablePackages[e$AvailablePackages$Package %in% pkgs, ]

  for(pkg in pkgs) {

    lrw <- lst[lst$Package == pkg, ]


    if (nrow(lrw) == 0) {

      message(paste0("Package data for '", pkg, "' not found."))
    } else {

      # Get package version
      ver <- lrw$Version

      # Convert dates
      date <- lrw$Release

      # Get package source
      src <- get_file_name(pkg, ver)

      sz <- NA
      if (skip_size == FALSE) {

        # Get temp file name
        flnm <- tempfile()

        currentpath = e$MirrorCurrentPath

        # Get path to package
        url <- file.path(currentpath, src)

        # Download from CRAN
        rc <- tryCatch({utils::download.file(url, flnm, quiet = TRUE)},
                       error = function(e){1})

        # Get size

        if (rc == 0) {
          rsz <- file.size(flnm)
          if (rsz < 1048576) {

            sz <- sprintf("%.0fK", round(rsz / 1024, 0))
          } else if (rsz < 1073741824 ) {

            sz <- sprintf("%.1fM", round(rsz / 1048576, 1))
          } else {
            sz <- sprintf("%.1fG", round(rsz / 1073741824, 1))

          }

          unlink(flnm)
        }
      }

      # print(pkg)
      # print(ver)
      # print(src)
      # print(as.Date(date))
      # print(sz)
      # Create data from from captured info
      rw <- data.frame(Package = pkg, Version = ver, FileName = src,
                       "Release" = as.Date(date, origin = '1970-01-01'),
                       "Size" = sz)

      if (is.null(ret)) {
        ret <- rw
      } else {
        ret <- rbind(ret, rw)
      }

    }
  }

  return(ret)
}

# Retrieve latest info for a package from CRAN
#' @import rvest
#' @import utils
#' @noRd
get_latest_data_back <- function(pkgs, skip_size = FALSE, msg = TRUE) {

  baseurl = e$CranPackagePath

  if (is.null(pkgs)) {
    stop("pkgs parameter cannot be NULL.")
  }

  ret <- NULL

  for(pkg in pkgs) {

    # Concat url
    url <- file.path(baseurl, pkg, "index.html")

    # Read and parse page into data frame
    page <- tryCatch({rvest::read_html(url)},
                     error = function(e){NULL})

    if (!is.null(page)) {
      removed <- grepl("archive", page, fixed = TRUE)[[1]]
      archived <- grepl("archive", page, fixed = TRUE)[[1]]
      tables <- html_elements(page, "table")
    } else {
      removed <- FALSE
      archived <- FALSE
      tables <- NULL
    }


    if (is.null(page)) {

      if (msg)
        message(paste0("Package data for '", pkg, "' not found."))

    } else if (removed && archived && length(tables) == 0) {

      if (msg)
        message(paste0("Package data for '", pkg, "' has been archived."))

      ret <- data.frame(Package = pkg, Version = "archived")


    } else {

      table1 <- html_table(tables[1], fill = TRUE)[[1]]
      table3 <- html_table(tables[3], fill = TRUE)[[1]]

      # Replace non-breaking space
      table1$X1 <- gsub('\xc2\xa0+', " ", table1$X1)
      table3$X1 <- gsub('\xc2\xa0+', " ", table3$X1)

      # Get package version
      ver <- subset(table1, table1$X1 == "Version:")[[2]]

      # Convert dates
      date <- subset(table1, table1$X1 == "Published:")[[2]]

      # Get package source
      src <- subset(table3, grepl("Package\\s+source:", table3$X1))[[2]]

      if (length(src) == 0)
        src <- get_file_name(pkg, ver)

      sz <- NA
      if (skip_size == FALSE) {

        # Get temp file name
        flnm <- tempfile()

        currentpath = e$CranCurrentPath

        # Get path to package
        url <- file.path(currentpath, src)

        # Download from CRAN
        rc <- tryCatch({utils::download.file(url, flnm, quiet = TRUE)},
                       error = function(e){1})

        # Get size

        if (rc == 0) {
          sz <- format(structure(file.size(flnm), class = "object_size"),
                       units = "auto")
          sz <- sub(" Kb", "K", sz, fixed = TRUE)
          sz <- sub(" Mb", "M", sz, fixed = TRUE)
          sz <- sub(" Gb", "G", sz, fixed = TRUE)
          unlink(flnm)
        }
      }

      # print(pkg)
      # print(ver)
      # print(src)
      # print(as.Date(date))
      # print(sz)
      # Create data from from captured info
      rw <- data.frame(Package = pkg, Version = ver, FileName = src,
                       "Release" = as.Date(date, origin = '1970-01-01'),
                       "Size" = sz, Archived = archived)

      if (is.null(ret)) {
        ret <- rw
      } else {
        ret <- rbind(ret, rw)
      }

    }
  }

  return(ret)
}

# Retrieve a table of info for latest version of a package
#' @noRd
get_latest_version <- function(pkgs, msg = TRUE) {

  #browser()
  ret <- c()

  refresh_package_lists()

  lst <- e$AvailablePackages[e$AvailablePackages$Package %in% pkgs, ]

  pos <- 1
  nfpkgs <- c()
  for (pkg in pkgs) {

    if (nrow(lst) == 0) {
      dat <- get_latest_data_back(pkg, skip_size = TRUE, msg = msg)[1, "Version"]
    } else {

      dat <- lst[lst$Package == pkg, "Version"]
    }

    if (!is.null(dat)) {
      if (length(dat) > 0) {
        ret[pos] <- dat
      } else {
        nfpkgs[length(nfpkgs) + 1] <- pkg
      }

    } else {
      nfpkgs[length(nfpkgs) + 1] <- pkg
    }

    pos <- pos + 1
  }

  if (!is.null(ret))
    names(ret) <- pkgs[!pkgs %in% nfpkgs]

  return(ret)
}

# Retrieve the archive versions of a package as a table
#' @import rvest
#' @import common
#' @noRd
get_archive_versions <- function(pkgs) {


  # baseurl = "https://cran.r-project.org/src/contrib/Archive"
  baseurl <- e$CranArchivePath

  if (is.null(pkgs)) {
    stop("pkgs parameter cannot be NULL.")
  }

  ret <- NULL

  for (pkg in pkgs) {

    # Concat url
    url <- file.path(baseurl, pkg)

    # Read and parse page into data frame
    page <- tryCatch({rvest::read_html(url)},
                     error = function(e){NULL})

    if (is.null(page)) {
      message(paste0("Archive versions of '", pkg, "' not available."))
    } else {

      tables <- html_elements(page, "table")
      table1 <- html_table(tables[1], fill = TRUE)[[1]]

      # Get column names
      nms <- names(table1)

      # Clear out unnecessary columns
      if (nms[5] == "Description") {
        table1[[5]] <- NULL
      }
      if (nms[1] == "") {
        table1[[1]] <- NULL
      }


      # Clear out unnecessary rows
      tmp <- subset(table1, table1$Name != "Parent Directory" & table1$Name != "" &
                      table1$Name != "PACKAGES.rds")

      # Convert dates
      tmp[["Last modified"]] <- as.Date(tmp[["Last modified"]],
                                        origin = '1970-01-01')

      # Extract package version
      vers <- get_version(tmp$Name)

      # Create final data frame
      tmp <- data.frame(Package = pkg, Version = vers, FileName = tmp$Name,
                        Release = tmp[["Last modified"]],
                        Size = tmp$Size)

      # Sort descending
      tmp <- sort(tmp, by = "Release", ascending = FALSE)

      if (is.null(ret)) {
        ret <- tmp
      } else {
        ret <- rbind(ret, tmp)
      }
    }
  }

  return(ret)
}

# Get a table of all versions of a package
get_all_versions <- function(pkg, skip_size = FALSE) {

  refresh_package_lists()

  ldat <- get_latest_data(pkg, skip_size = skip_size)
  adat <- get_archive_versions(pkg)


  ret <- NULL

  if (!is.null(adat) && !is.null(ldat)) {

    ret <- rbind(ldat, adat)
  } else if (!is.null(ldat)) {

    ret <- ldat
  } else if (!is.null(adat)) {

    ret <- adat
  }

  rownames(ret) <- NULL

  return(ret)
}

# File Utilities ----------------------------------------------------------



#' @noRd
get_version <- function(pkgname) {

  # Get rid of file extensions
  step1 <- sub(".tar.gz", "", pkgname, fixed = TRUE)

  # Find first underscore
  pos <- regexpr("_", step1, fixed = TRUE)

  ret <- substring(step1, pos + 1)

  return(ret)

}


get_file_name <- function(pkgname, version) {

  ret <- paste0(pkgname, "_", version, ".tar.gz")

  return(ret)
}



# Local Repo Utilities -------------------------------------------------------



#' @noRd
get_current_version <- function(pkgname) {

  dat <- installed_packages(pkgname)

  if (nrow(dat) > 0)
    ret <- dat[["Version"]][1]
  else
    ret <- NULL

  return(ret)
}

#' @noRd
get_package_version <- function(pkg, repo = NULL) {

   p <- suppressWarnings(packageDescription(pkg, repo))

   ret <- p$Version

   return(ret)
}

#' @title Get Installed Packages and Versions
#' @description A utility function to return the installed packages
#' in the current repository and their
#' versions numbers.
#' @param pkgs Optional vector of package names.  If supplied,
#' this vector will be used to filter the repository package list.
#' @param repos The path to the desired repository. This parameter
#' is optional.  If not supplied, the function will use the
#' repository associated with the currently running version of R.
#' @returns A data frame containing each installed
#' package and its version number.
#' @examples
#' # Get all packages in repo
#' res <- installed_packages()
#'
#' # View top 10 results
#' res[1:12, ]
#' #        Package Version
#' # 1        abind   1.4-8
#' # 2         ards   0.1.1
#' # 3      askpass   1.2.0
#' # 4   assertthat   0.2.1
#' # 5    backports   1.5.0
#' # 6    base64enc   0.1-3
#' # 7         bigD   0.3.0
#' # 8  BiocManager 1.30.23
#' # 9          bit   4.0.5
#' # 10       bit64   4.0.5
#' @import utils
#' @import utils
#' @noRd
installed_packages <- function(pkgs = NULL, repos = NULL) {

  # Get pkgs if not passed
  if (is.null(pkgs)) {
    if (is.null(repos)) {
      repos <- .libPaths()
      if (any(grepl("/_build", repos, fixed = TRUE))) {

        repos <- gsub("/_build", "", repos, fixed = TRUE)
      }
    }

    pkgs <- c()
    for (pth in repos) {
      dirs <- common::dir.find(pth, up = -1, down = 1)
      if (length(dirs) > 0) {
        pkgs <- c(pkgs, basename(dirs))
      } else {

        message(paste0("No packages in repo path '", pth, "'"))
      }
    }

    # Take out _build
    pkgs <- pkgs[pkgs != "_build"]
  }

  # Get versions of each package
  vers <- c()
  for (pkg in pkgs) {
    mv <- tryCatch({get_package_version(pkg, repos)},
                   error = function(err) {
                     NA_character_
                   })
    vers[length(vers) + 1] <- mv
  }

  # Create data frame of results
  ret <- data.frame(Package = pkgs, Version = vers)

  # Kill any rownames
  rownames(ret) <- NULL

  return(ret)
}


#' @import common
#' @noRd
base_packages <- function() {

  pth <- .Library

  pths <- common::dir.find(pth, up = -1, down = 1)

  ret <- basename(pths)


  return(ret)
}

#' @noRd
is_base <- function(pkg) {

  ret <- FALSE
  if (pkg %in% e$BasePackages)
    ret <- TRUE

  return(ret)
}

# Difference Utilities ----------------------------------------------------



#' @noRd
get_removed_functions <- function(v1info, v2info) {

  ret <- c()

  idx <- names(v1info$Functions) %in% names(v2info$Functions)

  if (any(idx == FALSE)) {
    ret <- names(v1info$Functions)[!idx]
  }

  return(ret)
}

#' @noRd
get_added_functions <- function(v1info, v2info) {

  ret <- c()

  idx <- !names(v2info$Functions) %in% names(v1info$Functions)

  if (any(idx == TRUE)) {
    ret <- names(v2info$Functions)[idx]
  }

  return(ret)
}

#' @noRd
get_removed_parameters <- function(v1info, v2info) {

  ret <- list()

  ret <- c()

  args1 <- v1info$Functions
  args2 <- v2info$Functions

  # Get function names
  nms <- names(args1)

  # Filter out deprecated functions
  nms <- nms[nms %in% names(args2)]

  ret <- list()

  for (nm in nms) {

    # Get args for both versions
    f1 <- args1[[nm]]
    f2 <- args2[[nm]]

    # Strip out empty string
    f1 <- f1[nchar(f1) > 0]
    f2 <- f2[nchar(f2) > 0]

    # Look for removed functions
    dp <- f1[!f1 %in% f2]

    # if there are any, add to list
    if (length(dp) > 0) {

      ret[[nm]] <- dp
    }
  }


  return(ret)
}




#' @noRd
get_added_parameters <- function(v1info, v2info) {

  ret <- list()

  ret <- c()

  args1 <- v1info$Functions
  args2 <- v2info$Functions

  # Get function names
  nms <- names(args1)

  # Filter out deprecated functions
  nms <- nms[nms %in% names(args2)]

  ret <- list()

  # Loop through functions
  for (nm in nms) {

    f1 <- args1[[nm]]
    f2 <- args2[[nm]]

    # Compare parameters
    dp <- f2[!f2 %in% f1]

    # Add to return value
    if (length(dp) > 0) {
      ret[[nm]] <- dp
    }
  }


  return(ret)
}


#' @noRd
get_all_functions <- function(v1info, v2info) {

  ret <- c()

  if (!is.null(v2info)) {
    ret <- names(v2info$Functions)
  }
  return(ret)
}

#' @noRd
get_all_parameters <- function(info) {

  ret <- list()

  ret <- c()

  args1 <- info$Functions

  # Get function names
  nms <- names(args1)

  ret <- list()

  for (nm in nms) {

    dp <- args1[[nm]]

    # Remove empty string
    # dp <- dp[nchar(dp) > 0]

    if (length(dp) > 0) {

      ret[[nm]] <- dp
    }
  }


  return(ret)
}


#' @noRd
get_all_parameters_back <- function(info) {

  ret <- list()

  ret <- c()

  args1 <- info$FormalArgs

  # Get function names
  nms <- names(args1)

  # Filter out non-exported names
  nms <- nms[nms %in% info$ExportedFunctions]


  ret <- list()

  for (nm in nms) {

    dp <- args1[[nm]]

    if (length(dp) > 0) {

      ret[[nm]] <- dp
    }
  }


  return(ret)
}


get_parameter_count <- function(info) {


  ap <- get_all_parameters(info)

  ret <- 0
  for (nm in names(ap)) {

    ret <- ret + length(ap[[nm]])

  }

  return(ret)
}




# Retrieving Infos --------------------------------------------------------


# Fix
#' @noRd
get_all_infos <- function(pkg, versions = NULL) {

  ret <- NULL

  # print("Debug A")
  lVersion <- get_latest_version(pkg)

  # print("Debug B")

  if (is.null(versions)) {

    adat <- get_archive_versions(pkg)
    versions <- c(lVersion, adat$Version)
  }

  # print("Debug C")

  for (ver in versions) {

      info <- get_info_cran(pkg, ver)

    if (!is.null(info)) {
      ret[[ver]] <- info
    }
  }

  # print("Debug D")

  return(ret)

}

# @noRd
# get_fastest_info <- function(pkgname, version) {
#
#   ret <- NULL
#
#   if (!is.na(github_packages(pkgname))) {
#
#     info <- github_package(pkgname)
#     ret <- info$infos[[version]]
#
#   }
#
#   if (is.null(ret)) {
#
#     ret <- get_all_infos(pkgname, version)[[1]]
#   }
#
#   return(ret)
# }

# Special case needed for two infos
#' @noRd
get_fastest_infos <- function(pkgname, v1, v2, cache = TRUE) {

  inf1 <- NULL
  inf2 <- NULL
  ret <- list()

  if (!is.na(github_packages(pkgname)) && cache == TRUE) {

    info <- github_package(pkgname)
    if ("infos" %in% names(info)) {
      inf1 <- info$infos[[v1]]
      inf2 <- info$infos[[v2]]
      # if (!is.null(inf1) && !is.null(inf2))
      #   print("Retrived from github")
    }
  }

  if (is.null(inf1)) {

    inf1 <- get_info_cran(pkgname, v1)
  }

  if (is.null(inf2)) {

    inf2 <- get_info_cran(pkgname, v2)
  }

  ret[[v1]] <- inf1
  ret[[v2]] <- inf2

  return(ret)
}




# Github Utilities --------------------------------------------------------

#' @noRd
github_packages <- function(pkg = NULL) {

  pth <- "https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/packages.rds"

  murl <- url(pth)
  for (idx in seq(1, 10)) {
    ret <- tryCatch({get(load(gzcon(murl)))},
                    warning = function(cond){NULL},
                    error = function(cond){NULL})
    if (!is.null(ret)) {
      break
    }

    Sys.sleep(1)
  }
  close(murl)

  if (is.null(ret)) {

    stop("Pkgdiff GitHub package list not available.")
  }

  if (!is.null(pkg)) {
    ret <- ret[pkg]
  }

  return(ret)
}

#' @noRd
github_package <- function(pkg) {


  # https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/data/procs.Rdata
  # https://github.com/dbosak01/pkgdiffdata/blob/main/data/procs.Rdata
  # pth <- "https://github.com/dbosak01/pkgdiffdata/blob/main/data"
  pth <- e$GithubPath


  fl <- file.path(pth, paste0(pkg, ".RData"))

  murl <- url(fl)

  for (idx in seq(1, 10)) {
    info <- tryCatch({get(load(gzcon(murl)))},
                    warning = function(cond){NULL},
                    error = function(cond){NULL})
    if (!is.null(info)) {
      break
    }

    Sys.sleep(1)
  }

  close(murl)

  if (is.null(info)) {

    stop(paste0("Pkgdiff GitHub package '", pkg, "' not available."))
  }


  return(info)

}


github_update <- function() {

  pth <- "https://github.com/dbosak01/pkgdiffdata/raw/refs/heads/main/LastUpdate.RData"

  murl <- url(pth)
  # ret <- get(load(gzcon(murl)))
  for (idx in seq(1, 10)) {
    ret <- tryCatch({get(load(gzcon(murl)))},
                    warning = function(cond){NULL},
                    error = function(cond){NULL})
    if (!is.null(ret)) {
      break
    }

    Sys.sleep(1)
  }

  close(murl)

  if (is.null(ret)) {

    stop("Pkgdiff GitHub LastUpdated datestamp not available.")
  }

  return(ret)
}


# Global Lists ------------------------------------------------------------

#' @import utils
#' @import tools
#' @noRd
available_packages <- function() {

  mror <- e$Mirror

  # Get available packages
  flds1 <- c("Package", "Repository")
  apkgs <- as.data.frame(available.packages(repos = mror))[ , flds1]

  flds2 <- c("Package", "Version", "Date/Publication", "Maintainer",
             "Description", "Title", "Depends", "LinkingTo", "Suggests",
             "Enhances", "License")

  # Get CRAN packages
  dpkgs <- tools::CRAN_package_db()[ , flds2]

  # Remove duplicates
  dpkgs <- dpkgs[!duplicated(dpkgs$Package, fromLast = TRUE), ]

  # Merge to add repository
  ret <- merge(dpkgs, apkgs, by.x = c("Package"),
               by.y = c("Package"))

  # Format date
  ret$ReleaseDate <- as.Date(ret$`Date/Publication`)

  return(ret)
}


#' @noRd
#' @import cranlogs
is_popular <- function(pkg, count = 1000) {

  dld <- sum(cranlogs::cran_downloads(pkg,
                                      when = "last-month")[["count"]])

  if (dld >= count)
    ret <- TRUE
  else
    ret <- FALSE

  return(ret)
}


# Other -------------------------------------------------------------------



#' Extract R Package
#' Untar an R package into a temp directory.
#' @param x The compressed (tar.gz) build file of an R package, either local or
#' URL.
#' @return List of files extracted.
#' @import utils
#' @noRd
unzip_package <- function(pth) {
  td <- tempdir()
  if(grepl('^https?:', pth)) {
    url <- pth
    pth <- file.path(td, basename(pth))
    utils::download.file(url, destfile = pth, quiet = TRUE)
  }
  nm <- sub('\\.tar\\.gz', '', basename(pth))
  dir <- file.path(td, nm)
  utils::untar(pth, exdir = dir)
  ret <-  list.files(dir, full.names = TRUE)

  return(ret)
}

# Show a file in the viewer.
# This is used in pkg_report()
#' @noRd
show_viewer <- function(path) {

  pth <- path

  #print(paste0("path: ", path))

  if (interactive()) {

    if (file.exists(path)) {

      viewer <- getOption("viewer")

      if (!is.null(viewer)) {
        viewer(pth)
      } else {
        utils::browseURL(pth)
      }

    }

  }


  return(pth)

}

#' @noRd
sort_data_frame <- function(x, decreasing = FALSE, ..., by = NULL,
                            ascending = TRUE, na.last = TRUE,
                            index.return = FALSE) {

  if (!"data.frame" %in% class(x)) {
    stop("Input object must be derived from a data frame")

  }

  nms <- names(x)
  if (!all(by %in% nms)) {
    lst <- by[!by %in% nms]

    stop(paste0("By value '", lst, "' is not a column on the input data frame."))

  }

  # A temporary list to hold columns
  tmp <- list()

  # Store input dataset in new variable
  df <- x

  # Default by is all variable names
  if (is.null(by))
    by <- names(df)

  # Default ascending
  a <- rep(TRUE, length(by))

  # Set ascending if supplied
  if (!is.null(ascending)) {
    a <- rep(ascending, length(by))
  }
  names(a) <- by

  # Create rank columns to handle custom sorts
  for (nm in by) {

    if (a[nm] == TRUE)
      tmp[[nm]] <- rank(df[[nm]], na.last = na.last)
    else
      tmp[[nm]] <- -rank(df[[nm]], na.last = na.last)
  }

  # Get modified dataframe
  tmp <- as.data.frame(tmp, stringsAsFactors = FALSE)

  # Get row order
  ord <- do.call('order', tmp)

  if (index.return) {
    ret <- ord
  } else {
    # Sort input dataframe
    ret <- df[ord, , drop = FALSE]

    # Restore attributes
    ret <- copy.attributes(x, ret)
  }



  return(ret)


}



parse_date <- function(dstrng) {

  ret <- tryCatch({as.Date(dstrng)},
                  warning = function(cond) {NULL},
                  error = function(cond) { NULL})

  if (is.null(ret) || is.na(ret)) {

    # Extract timezone (last word)
    tz <- sub(".*\\s+", "", dstrng)

    # Remove timezone from the timestamp
    ts <- sub(paste0(" ", tz, "$"), "", dstrng)

    # Parse and convert to Date
    ret <- tryCatch({as.Date(as.POSIXct(ts,
                                        format = " %Y-%m-%d %H:%M:%S",
                                        tz = tz))},
                    warning = function(cond) {NULL},
                    error = function(cond) {NULL})


  }

  if (is.null(ret) || is.na(ret)) {


    # 1. Remove surrounding single quotes and leading/trailing spaces
    clean <- gsub("^[' ]+|[' ]+$", "", dstrng)

    # 2. Extract the numeric timezone (e.g., "-0600")
    offset <- sub(".* ([+-][0-9]{4})$", "\\1", clean)

    # 3. Convert offset to "Etc/GMT" format
    #    (Note: Etc/GMT signs are reversed relative to offsets)
    hours_offset <- suppressWarnings(as.integer(substr(offset, 1, 3)))  # "-06" → -6
    tz <- suppressWarnings(sprintf("Etc/GMT%+d", -hours_offset))  # becomes "Etc/GMT+6"

    # 4. Remove the offset from the datetime string
    datetime <- sub(paste0(" ", offset, "$"), "", clean)

    # 5. Parse into POSIXct using the dynamically generated timezone
    ret <- tryCatch({as.Date(as.POSIXct(datetime,
                     format = "%a, %d %b %Y %H:%M:%S",
                     tz = tz))},
                   warning = function(cond) {NULL},
                   error = function(cond) {NULL})

  }

  if (is.null(ret) || is.na(ret)) {

    ret <- dstrng
  }


  return(ret)

}

dev <- FALSE

test_that("utilities1: get_archive_versions() works as expected.", {

  if (dev) {

    res <- get_archive_versions("fmtr")

    res

    expect_equal(is.data.frame(res), TRUE)

    expect_equal(nrow(res) > 0, TRUE)

    expect_equal(ncol(res) == 5, TRUE)




    if (dev) {

      res <- get_archive_versions(c("fmtr", "common", "rvest"))

      res

      expect_equal(is.data.frame(res), TRUE)

      expect_equal(nrow(res) > 0, TRUE)

      expect_equal(ncol(res) == 5, TRUE)

    }

  } else {
    expect_equal(TRUE, TRUE)
  }
})


test_that("utilities2: get_latest_version() works as expected.", {

  if (dev) {

    res <- get_latest_version("logr")

    expect_equal(is.data.frame(res), FALSE)

    expect_equal(length(res) == 1, TRUE)

    res <- get_latest_version(c("logr", "fmtr", "common"))

    expect_equal(is.data.frame(res), FALSE)

    expect_equal(is.list(res), FALSE)

    expect_equal(length(res) == 3, TRUE)

    res <- get_latest_version("Matrix")

    expect_equal(is.data.frame(res), FALSE)

    expect_equal(is.list(res), FALSE)

    expect_equal(length(res) == 1, TRUE)

    res <- get_latest_version("grid")

    expect_equal(is.data.frame(res), FALSE)

    expect_equal(is.list(res), FALSE)

    expect_equal(length(res) == 1, TRUE)

    expect_equal(res[[1]] == "archived", TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities3: get_version() works as expected.", {


  vers <- c("logr_1.2.3.tar.gz", "logr_1.2.4.tar.gz", "logr_1.3.0.tar.gz")

  if (dev) {

    res <- get_version(vers)

    expect_equal(res[1], "1.2.3")
    expect_equal(res[2], "1.2.4")
    expect_equal(res[3], "1.3.0")

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities4: installed_packages() works as expected.", {

  pkgs <- c("rvest", "curl", "utils", "forker", "common")

  res <- installed_packages(pkgs)

  expect_equal(is.data.frame(res), TRUE)

  expect_equal(nrow(res) == 5, TRUE)

  expect_equal(ncol(res) == 2, TRUE)


  if (dev) {

    res <- installed_packages()

    expect_equal(is.data.frame(res), TRUE)

    expect_equal(nrow(res) > 1, TRUE)

    expect_equal(ncol(res) == 2, TRUE)

    res <- installed_packages(c("logr", "fmtr", "common", "reporter"))

    expect_equal(is.data.frame(res), TRUE)

    expect_equal(nrow(res) > 1, TRUE)

    expect_equal(ncol(res) == 2, TRUE)

    pth <- "C:\\Users\\dbosa\\AppData\\Local\\R\\win-library\\4.2"

    res <- installed_packages(c("logr", "fmtr", "common", "reporter"),
                                  repos = pth)

    expect_equal(is.data.frame(res), TRUE)

    expect_equal(nrow(res) > 1, TRUE)

    expect_equal(ncol(res) == 2, TRUE)
  }

})


test_that("utilities5: get_current_version() works as expected.", {

  if (dev) {

    res <- get_current_version("common")

    expect_equal(is.data.frame(res), FALSE)

    expect_equal(length(res), 1)

    expect_equal(nchar(res) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("utilities6: get_file_name() works as expected.", {

  if (dev) {

    res <- get_file_name("common", "1.2.3")

    expect_equal(res, "common_1.2.3.tar.gz")

  } else {
    expect_equal(TRUE, TRUE)
  }
})


test_that("utilities7: get_latest_data() works as expected.", {

  if (dev) {

    res <- get_latest_data("fmtr")

    expect_equal(is.data.frame(res), TRUE)

    expect_equal(nrow(res) == 1, TRUE)

    expect_equal(ncol(res) == 5, TRUE)

    if (dev) {

       res <- get_latest_data(c("common", "reporter", "purr",
                                "fmtr", "Matrix"), skip_size = FALSE)

       expect_equal(is.data.frame(res), TRUE)

       expect_equal(nrow(res) == 4, TRUE)

       expect_equal(ncol(res) == 5, TRUE)

    }

    res <- get_latest_data("purr")

    expect_equal(is.null(res), TRUE)


    if (dev) {

      res <- get_latest_data("reporter")

      expect_equal(is.data.frame(res), TRUE)

      expect_equal(nrow(res) == 1, TRUE)

      expect_equal(ncol(res) == 5, TRUE)


      res <- get_latest_data("Matrix")

      expect_equal(is.data.frame(res), TRUE)

      expect_equal(nrow(res) == 1, TRUE)

      expect_equal(ncol(res) == 5, TRUE)

    }

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("utilities8: get_removed_functions() works as expected.", {

  if (dev) {

    # Lot of changes
    info1 <- pkg_info("admiral", "0.12.3")
    info2 <- pkg_info("admiral", "1.0.0")

    res <- get_removed_functions(info1, info2)

    expect_equal(length(res) > 0, TRUE)


  } else {

    expect_equal(TRUE, TRUE)
  }

})



test_that("utilities9: get_removed_parameters() works as expected.", {

  if (dev) {

    # Lot of changes
    info1 <- pkg_info("admiral", "0.12.3")
    info2 <- pkg_info("admiral", "1.0.0")

    res <- get_removed_parameters(info1, info2)

    expect_equal(length(res) > 0, TRUE)


  } else {

    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities10: get_added_functions() works as expected.", {

  if (dev) {

    # Lot of changes
    info1 <- pkg_info("admiral", "0.12.3")
    info2 <- pkg_info("admiral", "1.0.0")

    res <- get_added_functions(info1, info2)

    expect_equal(length(res) > 0, TRUE)


  } else {

    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities11: get_added_parameters() works as expected.", {

  if (dev) {

    # Lot of changes
    info1 <- pkg_info("admiral", "0.12.3")
    info2 <- pkg_info("admiral", "1.0.0")

    res <- get_added_parameters(info1, info2)

    expect_equal(length(res) > 0, TRUE)


  } else {

    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities12: get_all_functions() works as expected.", {

  if (dev) {


    # Lot of changes
    info1 <- pkg_info("admiral", "0.12.3")
    info2 <- pkg_info("admiral", "1.0.0")

    res <- get_all_functions(info1, info2)

    expect_equal(length(res) > 0, TRUE)


  } else {

    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities13: pkg_info() works as expected.", {

  if (dev) {

    res <- pkg_info("tibble", "latest")

    expect_equal(is.null(res), FALSE)

    expect_equal("pinfo" %in% class(res), TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("utilities14: pkg_info() works as expected.", {

  if (dev) {

    res <- pkg_info("fmtr", "1.0.1")

    expect_equal(is.null(res), FALSE)

    expect_equal("pinfo" %in% class(res), TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities15: get_all_parameters() works as expected.", {


  if (dev) {

    inf <- pkg_info("fmtr", "1.0.1")

    res <- get_all_parameters(inf)

    res

    expect_equal(is.null(res), FALSE)

    expect_equal("pinfo" %in% class(inf), TRUE)

    expect_equal(length(res) > 0, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities16: get_parameter_count() works as expected.", {

  if (dev) {

    inf <- pkg_info("fmtr", "1.0.1")

    res <- get_parameter_count(inf)

    res

    expect_equal(is.null(res), FALSE)

    expect_equal(res, 36)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities17: get_archive_versions() for negative tests.", {

  if (dev & FALSE) {

    # Non-existent package name
    res <- get_archive_versions("definrr")

    # No error
    expect_equal(is.null(res), TRUE)

    res <- get_archive_versions("defineR")

    res

    expect_equal(is.null(res), TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})


test_that("utilities18: get_all_infos() works as expected.", {


  if (dev) {

    res <- get_all_infos("procs")

    res

    expect_equal(is.list(res), TRUE)
    expect_equal(length(res) > 0, TRUE)
    expect_equal("pinfo" %in% class(res[[1]]), TRUE)

    res <- get_all_infos("procs", c("1.0.3", "1.0.4"))

    res

    expect_equal(is.list(res), TRUE)
    expect_equal(length(res) == 2, TRUE)
    expect_equal("pinfo" %in% class(res[[1]]), TRUE)

    res <- get_all_infos("forker")

    res

    expect_equal(is.null(res), TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities19: github_packages() works as expected.", {

  if (dev) {

    res <- github_packages()

    expect_equal(length(res) > 0, TRUE)
    expect_equal("logr" %in% names(res), TRUE)

    res <- github_packages("logr")

    expect_equal(length(res) == 1, TRUE)
    expect_equal("logr" %in% names(res), TRUE)


    res <- github_packages("loxx")
    expect_equal(is.na(res[[1]]) == TRUE, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities20: get_all_versions() works as expected.", {


  if (dev) {

    res <- get_all_versions("zoo")

    res

    expect_equal(is.data.frame(res), TRUE)
    expect_equal(nrow(res) > 0, TRUE)
    expect_equal(ncol(res), 5)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


test_that("utilities21: github_package() works as expected.", {

  if (dev) {

    res <- github_package("tibble")

    expect_equal(is.data.frame(res$stability), TRUE)
    expect_equal(nrow(res$stability) > 0, TRUE)
    expect_equal("pinfo" %in% class(res$infos[[1]]), TRUE)
    expect_equal(length(res$infos) > 2, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }

})


# test_that("utilities22: get_fastest_info() works as expected.", {
#
#   if (dev) {
#     res <- get_fastest_info("procs", "1.0.6")
#
#     expect_equal("pinfo" %in% class(res), TRUE)
#
#     res <- get_fastest_info("procs", "1.0.5")
#
#     expect_equal("pinfo" %in% class(res), TRUE)
#
#     res <- get_fastest_info("ccRemover", "1.0.4")
#
#     expect_equal("pinfo" %in% class(res), TRUE)
#
#   } else {
#
#     expect_equal(TRUE, TRUE)
#   }
#
# })


test_that("utilities23: get_fastest_infos() works as expected.", {

  if (dev) {

    res <- get_fastest_infos("procs", "1.0.5", "1.0.6")

    expect_equal(length(res), 2)
    expect_equal("pinfo" %in% class(res[[1]]), TRUE)

    # Why no functions?
    res <- get_fastest_infos("ccRemover", "1.0.3", "1.0.4")

    expect_equal(length(res), 2)
    expect_equal("pinfo" %in% class(res[[1]]), TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})


test_that("utilities24: available_packages() works as expected.", {

  res <- available_packages()

  expect_equal(is.null(res), FALSE)
  expect_equal(is.data.frame(res), TRUE)
  expect_equal(nrow(res) > 0, TRUE)

})

test_that("utilities25: refresh_package_lists() works as expected.", {

  if (dev) {

    res <- refresh_package_lists()

    # expect_equal(res, TRUE)
    expect_equal(is.null(e$AvailablePackages), FALSE)
    expect_equal(is.null(e$SavedPackages), FALSE)
    expect_equal(nrow(e$AvailablePackages) > 0, TRUE)
    expect_equal(length(e$SavedPackages) > 0, TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})

test_that("utilities26: is_popular() works as expected.", {

  if (dev) {

    res <- is_popular("logr")

    expect_equal(res, TRUE)

    res <- is_popular("procs")

    expect_equal(res, TRUE)

    res <- is_popular("ards")

    expect_equal(res, FALSE)

    res <- is_popular("ards", 20)

    expect_equal(res, TRUE)


  } else {

    res <- is_popular("dplyr")

    expect_equal(res, TRUE)
  }

})


test_that("utilities27: github_update() works as expected.", {

  res <- github_update()

  expect_equal("POSIXct" %in% class(res), TRUE)

})

test_that("utilities27: github_update() works as expected.", {

  res <- get_package_version("common")

  expect_equal(is.character(res), TRUE)

  expect_error(get_package_version("forker"))

  if (dev) {
    res <- get_package_version("Matrix")

    expect_equal(is.character(res), TRUE)

  }

})


test_that("utilities28: is_base() works as expected.", {

  res <- is_base("grid")

  expect_equal(res, TRUE)

  res <- is_base("utils")

  expect_equal(res, TRUE)

  res <- is_base("common")

  expect_equal(res, FALSE)

  res <- is_base("rvest")

  expect_equal(res, FALSE)

})


test_that("utilities29: get_info_description() works as expected.", {

  res <- get_info_description("grid")

  expect_equal("pinfo" %in% class(res), TRUE)
  expect_equal(res$Package, "grid")

})


test_that("utilities30: parse_date() works as expected.", {



  d1 <- " 2025-02-28 10:27:56 UTC"

  res1 <- parse_date(d1)

  expect_equal(res1, as.Date("2025-02-28"))


  d2 <- " 'Wed, 08 Dec 2021 16:09:23 -0600'"

  res2 <- parse_date(d2)

  expect_equal(res2, as.Date("2021-12-08"))

  d3 <- "bork"

  res3 <- parse_date(d3)

  expect_equal(res3, "bork")


})

dev <- FALSE

test_that("reports1: repo_breakages() works as expected.", {

  if (dev) {

    pkgs <- c("common", "knitr", "rvest")
    vrs <- c("1.1.3", "1.39", "0.3.6")

    if (dev & FALSE) {

      pkgs <- c("admiral", "common", "rvest")

      pth <- "C:\\Users\\dbosa\\AppData\\Local\\R\\win-library\\4.2"
      lst1 <- installed_packages(pkgs,
                                   repos = pth)

    } else {

      lst1 <- data.frame(Package = pkgs, Version = vrs)
    }

    lst2 <- get_latest_data(pkgs)[ , c("Package", "Version")]

    res <- repo_breakages(lst1, lst2)

    expect_equal(is.data.frame(res$Summary), TRUE)

    expect_equal(nrow(res$Summary) == 3, TRUE)

    expect_equal(ncol(res$Summary) == 4, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }


})

test_that("reports2: repo_stability() works as expected.", {

  if (dev) {

    pkgs <- c("common", "fmtr", "libr", "defineR")

    if (dev & FALSE) {

      pkgs <- c("admiral", "common", "rvest")

    }

    res <- repo_stability(pkgs)

    res

    expect_equal(is.data.frame(res), TRUE)

    expect_equal(nrow(res) == 4, TRUE)

    expect_equal(ncol(res) == 9, TRUE)

  } else {
    expect_equal(TRUE, TRUE)
  }


})

test_that("reports3: repo_breakages() handles edge cases.", {


  if (dev & FALSE) {

    r1 <- pkg_repo(ver = "4.2.1")
    r2 <- pkg_repo(ver = "latest")

    # What happens when you upgrade
    # the whole repository
    # Lot of situations to handle
    res <- repo_breakages(r1, r2)

    res$Summary
    res$Details$rmarkdown
    res$Details$admiral$RemovedFunctions

    expect_equal(is.data.frame(res$Summary), TRUE)
    expect_equal(nrow(res$Summary) > 0, TRUE)
    expect_equal(ncol(res$Summary) == 5, TRUE)
    expect_equal(any(res$Summary$Breakages == TRUE), TRUE)

  } else {

    expect_equal(TRUE, TRUE)
  }

})


test_that("reports4: repo_stability() handles edge cases.", {


  if (dev & FALSE) {

    r1 <- pkg_repo(ver = "4.2.1")

    r2 <- pkg_repo()


    # What happens when you get stability scores
    # for a whole repository

    # Failed on appgen (not a real package)
    # Failed on KMsurv (no releases within 10 years)
    # concatenate package getting no score
    # All above fixed now.
    res <- repo_stability(r1$Package)

    expect_equal(is.null(res), FALSE)
    expect_equal(nrow(res) > 0, TRUE)
    expect_equal(ncol(res), 9)


  } else {

    expect_equal(TRUE, TRUE)
  }

})

#' @title Identify Breakages for a Repository
#' @description
#' The \code{repo_breakages} function generates a data frame
#' of breakage information for multiple packages.  This function can be used
#' to identify breakages for a small set of packages, or even an entire repository.
#' @details
#' The \code{repo_breakages} function aims to help with repository upgrades.
#' The purpose is to identify packages that will break if you upgrade
#' your repository.
#'
#' The function accepts two data frames as input.
#' Each data frame should have two columns: "Package" and "Version".
#' The "r1" data frame is for your current repo.  The "r2" data frame
#' is for the upgrade repo.
#'
#' When executed, \code{repo_breakages} will then match up the
#' source and target versions, run a difference operation for each package,
#' and look for breaking changes.
#'
#' The result will be a table of package names, the source and target version,
#' and whether or not there are any breakages. If there are
#' breakages, the difference object for that package will be
#' added to a "details" list.  This difference object can show you
#' exactly which functions or function parameters were removed.
#'
#' The \code{repo_breakages} function is one of primary motivations for
#' the \strong{pkgdiff} package.  This mass-comparison ability did not
#' exist in any other R package prior to \strong{pkgdiff}.
#' @param r1 A data frame that identifies the source repository packages
#' and versions. The default value is "current", which means
#' the function will use the current versions of all packages in
#' the current R repository.
#' @param r2 A data frame that identifies the target repository packages
#' and versions. The default value is "latest", which means
#' the function will use the latest versions of all packages on CRAN.
#' @returns A list containing a summary table and a list of difference
#' objects. The summary table will contain one row for each package. The
#' columns show the package name, source and target versions, and whether
#' there are any breaking changes between the specified versions.  If
#' breaking changes are found, the difference object for that package
#' will be included in the details list.
#' @seealso Use the \code{\link{pkg_repo}} function to help gather
#' version information for each package set.  Also see \code{\link{repo_stability}}
#' to collect stability information on multiple packages.
#' @family prepo
#' @examples
#' # Create package vector
#' pkgs <- c("curl", "dplyr", "crayon", "stringr")
#'
#' # Backdated versions
#' r1 <- data.frame(Package = pkgs,
#'                  Version = c("5.2.1", "1.1.4", "1.5.2", "1.5.0"))
#'
#' # More recent versions
#' r2 <- data.frame(Package = pkgs,
#'                  Version = c("6.2.1", "1.1.4", "1.5.3", "1.5.1"))
#'
#' # Find any breaking changes
#' # res <- repo_breakages(r1, r2) # - Commented out to pass CRAN checks
#' # Comparing curl v5.2.1/v6.2.1
#' # Comparing crayon v1.5.2/v1.5.3
#' # Comparing stringr v1.5.0/v1.5.1
#'
#' # View results
#' # res # - Commented out to pass CRAN checks
#' # # A repo breakages object
#' # - Run Datetime: 2025-03-01 17:47 UTC
#' # - Summary:
#' #   Package Version1 Version2 Breakages
#' # 1    curl    5.2.1    6.2.1      TRUE
#' # 2   dplyr    1.1.4    1.1.4     FALSE
#' # 3  crayon    1.5.2    1.5.3     FALSE
#' # 4 stringr    1.5.0    1.5.1     FALSE
#' # - Details:
#' #   # A difference object: curl package
#' #   - Comparing: v5.2.1/v6.2.1
#' # - Breaking Changes: TRUE
#' # - Added Functions: 1
#' # - Added Parameters: 3
#' # - Removed Parameters:
#' #   - multi_download(): timeout
#' @import common
#' @export
repo_breakages <- function(r1 = "current", r2 = "latest") {

  if (!is.data.frame(r1)) {
    if (r1 == "current") {
      r1 <- installed_packages()
    }
  }

  if (!is.data.frame(r1)) {
    stop("r1 package list must be a data frame.")
  }

  if (!"Package"  %in% names(r1) ||
      !"Version" %in% names(r1)) {
    stop("V1 package list must contain the columns 'Package' and 'Version'.")
  }

  if (!is.data.frame(r2)) {
    if (r2 == "latest") {

      lver <- get_latest_version(r1$Package)

      r2 <- data.frame(Package = r1$Package,
                            Version = lver)

    }
  }

  if (!"Package"  %in% names(r2) ||
      !"Version" %in% names(r2)) {
    stop("V2 package list must contain the columns 'Package' and 'Version'.")
  }

  # Necessary?
  # if (nrow(r1) != nrow(r2)) {
  #   stop("v1 and v2 package lists must be the same size.")
  # }

  dat <- data.frame("Package" = NA, Version1 = NA, Version2 = NA,
                    Breakages = FALSE)
  det <- list()

  # browser()

  for (idx in seq_len(nrow(r1))) {

    pkg <- r1$Package[[idx]]
    v1 <- r1$Version[[idx]]
    v2 <- r2[r2$Package == pkg, "Version"]

    bc <- FALSE
    if (is.na(v1)) {
      message(paste0("Source version for package '", pkg, "' is missing."))
      bc <- NA
    } else if (length(v2) == 0) {
      message(paste0("Target version for package '", pkg, "' is missing."))
      bc <- NA
    } else if (v1 != v2) {
      cat(paste0("Comparing ", pkg, " v", v1, "/v", v2, "\n"))
      diff <- tryCatch({suppressWarnings(pkg_diff(pkg, v1, v2))},
                       error = function(er){NULL})
      if (!is.null(diff))
        bc <- diff$BreakingChanges
      else
        bc <- NA
    }

    dat[[idx, "Package"]] <- pkg
    dat[[idx, "Version1"]] <- v1
    dat[[idx, "Version2"]] <- ifelse(length(v2) == 0, NA, v2)
    dat[[idx, "Breakages"]] <- bc

    if (!is.na(bc)) {
      if (bc) {
        det[[pkg]] <- diff
      }
    }
  }

  ret <- list(Datetime = Sys.time(), Summary = dat, Details = det)

  class(ret) <- c("rbreak", class(ret))

  return(ret)

}

#' @title Print a Repo Breakages Object
#' @description
#' Print routine for a repo breakages object of class "rbreak".
#' @param x The repo breakages object to print.
#' @param ... Follow-on parameters to the print function.
#' @param verbose If FALSE, prints only the difference removals.
#' If TRUE, prints both additions and removals. Default is FALSE.
#' @family prepo
#' @import crayon
#' @returns The repo breakage object, invisibly.
#' @export
print.rbreak <- function(x, ..., verbose = FALSE) {

  grey60 <- crayon::make_style(grey60 = "#999999")
  cat(grey60(paste0("# A repo breakages object\n")))


  if (!is.null(x$Datetime)) {
    tmstmp <- format(x$Datetime, format = "%Y-%m-%d %H:%M UTC")
    cat(paste0("- Run Datetime: ", tmstmp , "\n"))

  }


  if (!is.null(x$Summary)) {
    cat(paste0("- Summary:\n"))
    print(as.data.frame(x$Summary))
  }


  if (!is.null(x$Details)) {
    cat(paste0("- Details:\n"))
    for (d in x$Details) {
      print(d, verbose = verbose)
    }
  }

  invisible(x)
}



#' @title Generate Stability Scores for a Repository
#' @description The function generates stability scores
#' for a vector of packages. If passing a large number of packages,
#' be prepared for the
#' function to run for a considerable amount of time.
#' @details
#' To assess stability for a package, \strong{pkgdiff} has to compare all
#' releases of a package, and identify breaking changes between each release.
#' This comparison takes time, especially for packages that have been
#' active for many years.
#'
#' Therefore, the first step of the function is to compare all releases
#' for each package in the input vector.  The function will send messages
#' to the console to let you know which package it is currently comparing.
#' These messages will let you know how far the function is through the processing.
#'
#' Once all the comparisons are completed, the function will return a data
#' frame summary of stability results.  The data frame will have one
#' row for each package in the input vector.  The data frame will also
#' show some information about the comparison, and give a stability assessment
#' for each package.  The data frame columns are as follows:
#' \itemize{
#'   \item {\strong{Package}: The package name.}
#'   \item {\strong{FV}: The first version of the package.}
#'   \item {\strong{LV}: The last version of the package.}
#'   \item {\strong{FR}: The first release date of the package.}
#'   \item {\strong{LR}: The last release date of the package.}
#'   \item {\strong{TR}: The total number of releases.}
#'   \item {\strong{BR}: The number of breaking releases.}
#'   \item {\strong{Score}: The stability score.}
#'   \item {\strong{Assessment}: The stability assessment.}
#' }
#' To learn how the package stability score and assessment are calculated,
#' see \code{vignette("pkgdiff-stability")}.  Additional information is
#' included in the documentation of the \code{\link{pkg_stability}}
#' function.
#'
#' If a package is not found in the package cache, the function
#' will download and compare each version of the package on CRAN.  This
#' process may increase the processing time.  See \code{vignette("pkg-cache")}
#' and \code{\link{pkg_cache}} for additional information.
#'
#' If a package does not exist on CRAN, a row for that package will still
#' be returned.  However, all data values will be NA.  This situation may
#' occur if a package exists on Github, but has not yet been submitted
#' to CRAN.  The \strong{pkgdiff} stability functions only work with packages
#' that have been published to CRAN.
#' @param pkgs A vector of package names.
#' @param releases An integer number of releases to assess.  The
#' function will then limit the scope of the assessment to the specified
#' number of releases.
#' @param months An integer number of months from the current date.  The
#' function will then limit the scope of the assessment to the specified
#' number of months.
#' @returns A data frame of information regarding the stability of
#' each package in the input vector.
#' @family prepo
#' @import common
#' @examples
#' # Create vector of packages
#' vct <- c("curl", "dplyr", "rvest", "tidymodels")
#'
#' # Get stablity scores
#' res <- repo_stability(vct)
#' # Getting stability score for package 'curl'...
#' # Getting stability score for package 'dplyr'...
#' # Getting stability score for package 'rvest'...
#' # Getting stability score for package 'tidymodels'...
#'
#' # View stability results
#' res
#' # A repo stability object
#' # - Run Datetime: 2025-03-01 17:53 UTC
#' # - Summary:
#' #   Package    FV    LV         FR         LR TR BR Score        Assessment
#' # 1       curl   0.2 6.2.1 2014-11-20 2025-02-19 51  1  98.0       Very Stable
#' # 2      dplyr   0.1 1.1.4 2014-01-16 2023-11-17 45 20  87.5 Somewhat Unstable
#' # 3      rvest 0.1.0 1.0.4 2014-11-22 2024-02-12 14  4  93.2            Stable
#' # 4 tidymodels 0.0.1 1.3.0 2018-07-27 2025-02-21 14  0 100.0           Perfect
#'
#' @export
repo_stability <- function(pkgs, releases = NULL, months = NULL) {

  # Create data frame structure
  dat <- data.frame("Package" = NA, FV = NA, LV = NA,
                    FR = NA, LR = NA,
                    TR = NA, BR = NA, Score = NA, Assessment = NA)

  idx <- 1

  for (pkg in pkgs) {

    cat(paste0("Getting stability score for package '", pkg, "'...\n"))

    # Get stability data for package
    rpt <- pkg_stability(pkg, releases = releases, months = months)

    # Put values into data frame
    dat[[idx, "Package"]] <- pkg
    if (!is.null(rpt$FirstVersion))
      dat[[idx, "FV"]] <- rpt$FirstVersion
    if (!is.null(rpt$LastVersion))
      dat[[idx, "LV"]] <- rpt$LastVersion
    if (!is.null(rpt$FirstRelease))
      dat[[idx, "FR"]] <- as.Date(rpt$FirstRelease, origin = '1970-01-01')
    if (!is.null(rpt$LastRelease))
      dat[[idx, "LR"]] <- as.Date(rpt$LastRelease, origin = '1970-01-01')
    if (!is.null(rpt$NumReleases))
      dat[[idx, "TR"]] <- rpt$NumReleases
    if (!is.null(rpt$BreakingReleases))
      dat[[idx, "BR"]] <- rpt$BreakingReleases
    if (!is.null(rpt$StabilityScore))
      dat[[idx, "Score"]] <- rpt$StabilityScore
    if (!is.null(rpt$StabilityScore))
      dat[[idx, "Assessment"]] <- get_stability_assessment(rpt$StabilityScore)

    idx <- idx + 1
  }

  if (!is.null(dat)) {

    # Assign lables
    common::labels(dat) <- list(FV = "First Version", LV = "Last Version",
                                FR = "First Release", LR = "Last Release",
                                TR = "Total Releases", BR = "Breaking Releases",
                                Score = "Stability Score",
                                Assessment = "Stability Assessment",
                                Package = "Package Name")

    # Ensure release dates come out as dates
    dat$FR <- as.Date(dat$FR, origin = '1970-01-01')
    dat$LR <- as.Date(dat$LR, origin = '1970-01-01')
  }

  ret <- dat

  attr(ret, "Datetime") <- Sys.time()

  class(ret) <- c("rstability", class(ret))

  return(ret)
}

#' @title Print a Repo Stability Object
#' @description
#' Print routine for a repot stability object of class "rstability".
#'
#' @param x The repo stability object to print.
#' @param ... Follow-on parameters to the print function.
#' @family prepo
#' @import crayon
#' @returns The repo stability object, invisibly.
#' @examples
#' # Create vector of packages
#' vct <- c("curl", "dplyr", "rvest", "tidymodels")
#'
#' # Get stablity scores
#' res <- repo_stability(vct)
#'
#' # View stability scores
#' print(res)
#' # # A repo stability object
#' # - Run Datetime: 2025-03-01 19:19 UTC
#' # - Summary:
#' #      Package    FV    LV         FR         LR TR BR Score        Assessment
#' # 1       curl   0.2 6.2.1 2014-11-20 2025-02-19 51  1  98.0       Very Stable
#' # 2      dplyr   0.1 1.1.4 2014-01-16 2023-11-17 45 20  87.5 Somewhat Unstable
#' # 3      rvest 0.1.0 1.0.4 2014-11-22 2024-02-12 14  4  93.2            Stable
#' # 4 tidymodels 0.0.1 1.3.0 2018-07-27 2025-02-21 14  0 100.0           Perfect
#' @export
print.rstability <- function(x, ...) {

  grey60 <- crayon::make_style(grey60 = "#999999")
  cat(grey60(paste0("# A repo stability object\n")))


  dt <- attr(x, "Datetime")

  if (!is.null(dt)) {
    tmstmp <- format(dt, format = "%Y-%m-%d %H:%M UTC")
    cat(paste0("- Run Datetime: ", tmstmp , "\n"))

  }


  if (!is.null(x)) {
    cat(paste0("- Summary:\n"))
    df <- as.data.frame(x)
    scr <- sprintf("%.1f", df$Score * 100)
    df$Score <- scr
    print(df)
  }


  invisible(x)
}

#' @title Generate a Package Summary Report
#' @description
#' The \code{pkg_summary} function creates a package summary report.
#' The report aggregates some key information about the package
#' and a generates a chart showing breaking changes over time.
#' The function provides a nice snapshot that can help you get a feel for a package.
#' @details
#' The \code{pkg_summary} function generates an HTML
#' report that can be sent to the RStudio viewer.  It combines some information
#' from both the \code{pkg_info} and \code{pkg_stability} functions, and displays
#' it in a table and accompanying chart.
#'
#' The table shows a selection of fields
#' that can give you an overall sense of the package: its size, age, number
#' of releases, dependencies, etc.  The chart is a stacked bar chart
#' showing the number of breaking changes for each release.
#'
#' Here is an example of the package summary report:
#' \figure{summary1.png}
#' The chart shows each release issued by the package maintainer
#' over the last ~8.5 years. Releases prior to that are not displayed,
#' as they are not included in the stability score. The bars represent the
#' number of removed functions and removed parameters. The labels on each
#' bar show the release year, month and version.
#'
#' @seealso For additional information, see \code{\link{pkg_info}} and
#' \code{\link{pkg_stability}}.
#' @param pkg A quoted string containing the name of the package to report on.
#' This parameter is required.
#' @param releases An integer indicating the number of releases to collect
#' stability data for. For example, \code{releases = 10} will return stability
#' data for the last 10 releases of the package. Default is NULL, which means
#' the function will return data for all releases.
#' @param months An integer indicating the number of months back to collect
#' stability data for.  For example, \code{months = 24} will collect
#' stability data for the previous 2 years.  Default is NULL, meaning there is
#' no limitation on the number of release months, and the function will collect
#' data from all releases.
#' @param view Whether to display the report in the viewer.  By default,
#' the parameter is TRUE.
#' @param path A path and file name for the report.  If NULL, the function
#' will create a temporary file. Default is NULL.
#' @returns The path to the HTML report, invisibly.  If the \code{view} parameter is TRUE,
#' the function will send the report to the viewer.
#' @examples
#' # View summary report for "stringr" package
#' # - Set view = TRUE to see example
#' pkg_summary("stringr", view = FALSE)
#'
#' @import common
#' @import graphics
#' @export
pkg_summary <- function(pkg, releases = NULL, months = NULL, view = TRUE, path = NULL) {

  # Add parameters for months and releases
  # Automatic limit on 10 years.


  # Get temp directory
  td <- tempdir()

  # Construct temp path
  fpth <- file.path(td, paste0(pkg,  "_stability.html"))

  # Kill if exists
  if (file.exists(fpth)) {
    file.remove(fpth)
  }

  # Create image directory
  idir <- file.path(td, "images")
  if (!dir.exists(idir)) {
    dir.create(idir)
  }

  # Create image path
  ipth <-  file.path(idir, paste0(pkg,  "_stability.svg"))

  # Remove any existing files
  if (file.exists(ipth)) {
    file.remove(ipth)
  }

  # Get html page
  pth <- gen_html(pkg, fpth, ipth, releases, months)


  # Show in viewer
  if (view) {

    show_viewer(pth)

  }

  # Copy to user-requested path
  if (!is.null(path)) {

    # Create user file directory if needed
    udir <- dirname(path)
    if (!dir.exists(udir)) {
      dir.create(udir)
    }

    # Create user images directory if needed
    uidir <- file.path(udir, "images")
    if (!dir.exists(uidir)) {
      dir.create(uidir)
    }

    # Prepare target paths
    ufpth <- file.path(udir, basename(path))
    uipth <- file.path(uidir, basename(ipth))

    # Copy files from temp to user targets
    file.copy(fpth, ufpth)
    file.copy(ipth, uipth)

    pth <- path
  }


  invisible(pth)
}

#' @import grDevices
#' @noRd
gen_chart <- function(pkg, dat, fpth, sbttl) {

  # Extract stability data
  if (!is.null(dat) && nrow(dat) > 0) {

    # Filter for last ~8.5 years
    fltr <- (Sys.Date() - 3080) < as.Date(dat$Release)
    dat <- dat[fltr, ]

    # Get y scale
    bh <- dat$RF + dat$RP
    mx <- max(bh)
    ylm <- c(0, 10)
    if (mx > 10) {
      ylm <- c(0, ceiling(mx / 5) * 5)

    }

    # Create label columns
    dat$Label <- paste0(substring(dat$Release, 3, 7), " v", dat$Version)

    # Sort ascending
    # dat <- sort_by(dat, dat$Release)
    dat <- sort_data_frame(dat, by = "Release")

    # Reorder Releases chronologically (oldest to newest)
    # dat$Release <- factor(dat$Release, levels = dat$Release)

    # Matrix for barplot
    mat <- t(as.matrix(dat[, c("RF", "RP")]))

    # Capture current par settings
    oldmar <- par()$mar

    # Assign output path
    svg(fpth, height = 4, width = 12, pointsize = 12, family = "sans")

    # Set margins
    par(mar = c(5, 4, 4, 4))   # bottom, left, top, right

    # Create title and subtitle
    ttls <- "Breaking Changes by Release"
    if (sbttl != "") {
      ttls <- append(ttls, sbttl)
    }

    # Base R stacked barplot
    barplot(
      mat,
      beside = FALSE,
      col = c("steelblue", "tomato"),
      names.arg = dat$Label,
      las = 2,                        # vertical labels
      cex.names = 0.7,
     # cey.axis = 1.5,
      main = ttls,
      ylab = "Count",
      ylim = ylm
    )

    legend("topright",
           legend = c("Removed Functions", "Removed Parameters"),
           fill = c("steelblue", "tomato"),
           cex = 0.9)


    # Close device context
    dev.off()

    # Restore par
    par(mar = oldmar)

    # Get svg xml
    lns <- readLines(fpth, encoding = "UTF-8")

    # Remove the first line
    # XML definition is not needed and will mess up the HTML
    if (length(lns) > 2) {
      # ret <- paste('<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"',
      #           'width="100%" height="100%" viewBox="0 0 100 100" >')

      lns <- lns[seq(2, length(lns))]
    } else {
      lns <- lns
    }

    # Kill temp file
    if (file.exists(fpth)) {

      file.remove(fpth)
    }

    # Write modified lines
    writeLines(lns, con = fpth)

    ret <- fpth
  } else {

    ret <- ""
  }

  return(ret)

}


#' @noRd
gen_html <- function(pkg, fpth, ipth, rlses, mths) {


  # Get info data
  info <- pkg_info(pkg)

  if (!is.null(info)) {

    # Get stability data
    stb <- pkg_stability(pkg, releases = rlses, months = mths)

    # Modify image path for img tag
    ipth_mod <- file.path("./images", basename(ipth))

    # Construct subtitle
    sttl <- ""
    if (!is.null(rlses)) {
      sttl <- paste0("Releases = ", rlses)
    }
    if (!is.null(mths)) {
      if (sttl != "") {
        sttl <- paste0(sttl, " | ", "Months = ", mths)
      } else {
        sttl <- paste0("Months = ", mths)
      }
    }

    # Get chart path
    cpth <- ""
    if (interactive()) {
      cpth <- gen_chart(pkg, stb$StabilityData, ipth, sttl)
    }

    # Construct HTML
    lns <- c()

    lns[length(lns) + 1] <- "<!DOCTYPE html>"
    lns[length(lns) + 1] <- "<html>"
    lns[length(lns) + 1] <- get_styles()
    lns[length(lns) + 1] <- "<body>"
    lns[length(lns) + 1] <- paste0("<h2>", "Package Summary Report", "</h2>")
    lns[length(lns) + 1] <- get_table(info, stb)
    if (cpth != "") {
      lns[length(lns) + 1] <- paste0("<img src=\"", ipth_mod, "\"",
                                     " width=\"100%\" alt=\"Stability Chart\">")
    }
    lns[length(lns) + 1] <- "</body>"
    lns[length(lns) + 1] <- "</html>"


    # Write html file
    writeLines(lns, fpth)

  } else {


  }

  return(fpth)
}

# Construct style sheet
get_styles <- function() {

  ret <- '<style>
  body {font-family:Arial;
        font-size:11pt}
  </style>'

  return(ret)
}


get_table <- function(info, stab) {

  tblo <- '<table width="100%">'

  # Prepare row1
  scr <- sprintf("%.1f", stab$StabilityScore * 100)
  r1 <- paste0('<tr><td><b>Package Name:</b> ', info$PackageName, '</td>
         <td><b>Current Version:</b> ', info$Version, '</td>
         <td><b>Stability:</b> ', paste0(scr, " (", stab$Assessment, ")"), '</td></tr>')

  # Prepare row2
  r2 <- paste0('<tr><td><b>Maintainer:</b> ', info$Maintainer, '</td>
         <td colspan="2"><b>Title:</b> ', info$Title, '</td></tr>')

  # Prepare row3
  r3 <- paste0('<tr><td><b>Age:</b> ', stab$PackageAge, '</td>
         <td><b>First Release:</b> ', stab$FirstRelease, '</td>
         <td><b>Last Release:</b> ', stab$LastRelease, '</td></tr>')

  # Prepare row4
  r4 <- paste0('<tr><td><b>License:</b> ', info$License, '</td>
         <td><b>Num Releases:</b> ', stab$NumReleases, '</td>
         <td><b>Breaking Releases:</b> ', stab$BreakingReleases, '</td>
         </tr>')

  # Prepare row5
  r5 <- paste0('<tr><td><b>Depends:</b> ', info$Depends, '</td>
         <td><b>Num Functions:</b> ', length(info$Functions), '</td>
         <td><b>Downloads/month:</b> ', format(info$LastMonthDownloads, big.mark = ","), '</td></tr>')

  # Prepare row6
  r6 <- paste0('<tr><td colspan="3"><b>Imports:</b> ', info$Imports, '</td></tr>')

  # Prepare row7
  r7 <- paste0('<tr><td colspan="3"><b>Suggests:</b> ', info$Suggests, '</td></tr>')


  tblc <- '</table>'


  ret <- paste0(tblo, r1, r2, r3, r4, r5, r6, r7, tblc, collapse = "\n")

  return(ret)
}



# # Your data
# dat <- data.frame(
#   Release = c("2025-07-29", "2025-06-25", "2025-01-15", "2024-06-17", "2024-06-07",
#               "2024-03-05", "2024-02-02", "2023-12-16", "2023-10-18",
#               "2023-10-06", "2023-09-25", "2023-09-12", "2023-07-06",
#               "2023-06-09", "2023-04-25", "2023-03-14", "2023-03-08",
#               "2022-12-23", "2022-12-06", "2022-10-14", "2022-10-07",
#               "2022-09-29", "2022-09-21", "2022-09-05", "2022-07-18",
#               "2022-05-31", "2022-02-17"),
#   RF = c(0,0,2,0,0,0,0,15,0,0,0,5,0,8,0,0,7,0,8,0,0,0,0,32,0,13,0),
#   RP = c(0,0,10,0,8,0,0,2,0,0,0,0,0,2,0,0,4,0,0,0,0,0,0,2,0,4,0)
# )

base_path <- "c:/packages/pkgdiff/tests/testthat"
base_path <- "."

dev <- FALSE


test_that("summary1: gen_chart() basic functionality.", {

  if (dev) {

    ipth <- file.path(tempdir(), "stringr_stability.svg")

    sb <- pkg_stability("stringr")

    res <- gen_chart("stringr", sb$StabilityData, ipth, "")


    expect_equal(file.exists(res), TRUE)
  } else {

    expect_equal(TRUE, TRUE)
  }


})

test_that("summary2: gen_html() basic functionality.", {

  idir <- file.path(tempdir(), "images")
  ipth <- file.path(tempdir(), "images/stringr_stability.svg")
  fpth <- file.path(tempdir(), "stringr_stability.html")

  if (!dir.exists(idir)) {
    dir.create(idir)
  }


  # pkg, fpth, ipth
  res <- gen_html("stringr", fpth, ipth, NULL, NULL)


  expect_equal(file.exists(res), TRUE)


  # file.show(res)


})




test_that("summary3: pkg_summary() basic functionality.", {


  if (dev) {

    res <- pkg_summary("stringr")


    expect_equal(file.exists(res), TRUE)


    res <- pkg_summary("stringr", view = FALSE)


    expect_equal(file.exists(res), TRUE)


  } else {
    expect_equal(TRUE, TRUE)
  }

  # file.show(res)


})




test_that("summary4: pkg_summary() more breaking changes.", {

  if (dev) {

    res <- pkg_summary("admiral")


    expect_equal(file.exists(res), TRUE)

    res <- pkg_summary("gtsummary")


    expect_equal(file.exists(res), TRUE)


    # file.show(res)

  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("summary5: pkg_summary() less breaking changes.", {

  if (dev) {

    res <- pkg_summary("fmtr")


    expect_equal(file.exists(res), TRUE)

    res <- pkg_summary("logr")


    expect_equal(file.exists(res), TRUE)


    # file.show(res)

  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("summary6: pkg_summary() few releases.", {

  if (dev) {

    res <- pkg_summary("ards")



    expect_equal(file.exists(res), TRUE)


    # file.show(res)

  } else {
    expect_equal(TRUE, TRUE)
  }


})

test_that("summary7: pkg_summary() no releases.", {



  res <- pkg_summary("bork")



  expect_equal(file.exists(res), FALSE)


  # file.show(res)


})

test_that("summary8: pkg_summary() base package.", {



  res <- pkg_summary("tools")



  expect_equal(file.exists(res), TRUE)


  # file.show(res)


})


test_that("summary9: pkg_summary() path argument.", {


  if (dev) {

    pth <- file.path(base_path, "output/str01.html")

    res <- pkg_summary("stringr", path = pth)



    expect_equal(file.exists(res), TRUE)


    # file.show(res)
  } else {
    expect_equal(TRUE, TRUE)
  }


})


test_that("summary10: pkg_summary() releases and months parameters.", {


  if (dev) {

    res <- pkg_summary("admiral", releases = 10)


    expect_equal(file.exists(res), TRUE)


    res <- pkg_summary("admiral", months = 18)


    expect_equal(file.exists(res), TRUE)


    res <- pkg_summary("admiral", months = 18, releases = 3)


    expect_equal(file.exists(res), TRUE)


    # file.show(res)

  } else {
    expect_equal(TRUE, TRUE)
  }


})

#!/bin/bash

function decompile {
    # Decompile class and include private members. In preparation for the
    # hashing below, strip references on the form #[0-9]+.
    javap -c -p "$1" \
        | sed '/^Compiled from /d' \
        | sed '/^[ ]*}[ ]*$/d' \
        | sed 's/[ ]\(#[1-9][0-9]*\)[ ,]/ /' \
        | sed 's/[ ]\([0-9]*[:]\)//' \
        | sed 's/[ ][ ]*/ /g' \
        | sed '/Code:$/,/^$/s/^[ ]*\([^ ]\)/    \1/g' \
        | sed '/Code:$/d' \
        | sed 's/ *{$//'
}

function hash_bodies {
    # Generate hashes of the instructions making up the method bodies; Yields
    # more concise summaries.
    awk '/^[  ]{0,3}[^ ]/{if (x)print x"\n";x="";}{x=(!x)?$0"\n--":x" "$0;}END{print x;}' \
        | sed '/^--[ ]*$/d' \
        | awk "{ if(/^--/) system(\"printf '    '; (echo '\"\$0\"' | shasum --algorithm 256)\"); else print }" \
        | sed 's/[ ][ ]*-$//'
}

if [[ "$1" == "-s" ]]; then
    shift 1
    decompile "$1" | hash_bodies
else
    javap -c -s -p -verbose "$1"
fi

#!/bin/bash
#
# Synopsis:
#	Show changes between 2 internet-drafts using changebars or html
#	side-by-side diff.
#	
# Usage:
#	rfcdiff [options] file1 file2
#	
#	rfcdiff takes two RFCs or Internet-Drafts in text form as input, and
#	produces output which indicates the differences found in one of various
#	forms, controlled by the options listed below. In all cases, page
#	headers and page footers are stripped before looking for changes.
#	
#	--html		Produce side-by-side .html diff (default)
#	
#	--chbars	Produce changebar marked .txt output
#	
#	--diff		Produce a regular diff output
#	
#	--wdiff		Produce paged wdiff output
#	
#	--hwdiff	Produce html-wrapped coloured wdiff output
#	
#	--oldcolour COLOURNAME	Colour for new file in hwdiff (default is "green")
#	--oldcolor COLORNAME	Color for old file in hwdiff (default is "red")
#
#	--newcolour COLOURNAME	Colour for new file in hwdiff (default is "green")
#	--newcolor COLORNAME	Color for new file in hwdiff (default is "green")
#
#	--larger        Make difference text in hwdiff slightly larger
#
#	--browse	Show html output in browser
#	
#	--keep		Don't delete temporary workfiles
#	
#	--version	Show version
#	
#	--help		Show this help
#	
#	--info "Synopsis|Usage|Copyright|Description|Log"
#			Show various info
#	
#	--width	N	Set a maximum width of N characters for the
#			display of each half of the old/new html diff
#	
#	--linenum	Show linenumbers for each line, not only at the
#			start of each change section
#	
#	--body		Strip document preamble (title, boilerplate and
#			table of contents) and postamble (Intellectual
#			Property Statement, Disclaimer etc)
#	
#	--nostrip	Don't strip headers and footers (or body)
#	
#	--ab-diff	Before/After diff, suitable for rfc-editor
#	--abdiff
#	
#	--stdout	Send output to stdout instead to a file
#	
#       --tmpdiff       Path to intermediate diff file
#
#       --prelines N    Set value for diff -U option
#
#       --minimal       Set value for diff -d option
#
#       --ignore-space-change
#                       Ignore changes in the amount of white space.
#
#       --ignore-all-space
#                       Ignore all white space.
#
#       --ignore-blank-lines
#                       Ignore changes whose lines are all blank.
#       
#
# Copyright:
#	-----------------------------------------------------------------
#	
#	Copyright 2002 Henrik Levkowetz
#	
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#	
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	-----------------------------------------------------------------
#
# Description:
#
#	The purpose of this program is to compare two versions of an
#	internet-draft, and as output produce a diff in one of several
#	formats:
#	
#		- side-by-side html diff
#		- paged wdiff output in a text terminal
#		- a text file with changebars in the left margin
#		- a simple unified diff output
#	
#	In all cases, internet-draft headers and footers are stripped before
#	generating the diff, to produce a cleaner diff.
#	
#	It is called as
#	
#		rfcdiff first-file second-file
#	
#	The latest version is available from
#		http://tools.ietf.org/tools/rfcdiff/
#

export version="1.41"
export prelines="10"
export basename=$(basename $0)
export workdir=`mktemp -d -t rfcdiff.XXXXXXXX`
export pagecache1="$workdir/pagecache1"
export pagecache2="$workdir/pagecache2"

# ----------------------------------------------------------------------
# Utility to find an executable
# ----------------------------------------------------------------------
lookfor() {
    for b in "$@"; do
	found=$(which "$b" 2>/dev/null)
	if [ -n "$found" ]; then
	    if [ -x "$found" ]; then
		echo "$found"
		return
	    fi
	fi
    done
}

AWK=$(lookfor gawk nawk awk)

# ----------------------------------------------------------------------
# Strip headers footers and formfeeds from infile to stdout
# ----------------------------------------------------------------------
strip() {
  $AWK '
				{ gsub(/\r/, ""); }
				{ gsub(/[ \t]+$/, ""); }
				{ pagelength++; }
/\[?[Pp]age [0-9ivx]+\]?[ \t\f]*$/	{
				    match($0, /[Pp]age [0-9ivx]+/);
				    num = substr($0, RSTART+5, RLENGTH-5);
				    print num, outline > ENVIRON["pagecache" ENVIRON["which"]]
				    pagelength = 0;
				}
/\f/				{ newpage=1;
				  pagelength=1;
				}
/\f$/				{
				    # a form feed followed by a \n does not contribute to the
				    # line count.  (But a \f followed by something else does.)
				    pagelength--;
				}
/\f/				{ next; }
/\[?[Pp]age [0-9ivx]+\]?[ \t\f]*$/		{ preindent = indent; next; }

/^ *Internet.Draft.+[12][0-9][0-9][0-9] *$/ && (FNR > 15)	{ newpage=1; next; }
/^ *INTERNET.DRAFT.+[12][0-9][0-9][0-9] *$/ && (FNR > 15)	{ newpage=1; next; }
/^ *Draft.+(  +)[12][0-9][0-9][0-9] *$/	    && (FNR > 15)	{ newpage=1; next; }
/^RFC[ -]?[0-9]+.*(  +).* [12][0-9][0-9][0-9]$/ && (FNR > 15)	{ newpage=1; next; }
/^draft-[-a-z0-9_.]+.*[0-9][0-9][0-9][0-9]$/ && (FNR > 15)	{ newpage=1; next; }
/(Jan|Feb|Mar|March|Apr|April|May|Jun|June|Jul|July|Aug|Sep|Oct|Nov|Dec) (19[89][0-9]|20[0-9][0-9]) *$/ && pagelength < 3  { newpage=1; next; }
newpage && $0 ~ /^ *draft-[-a-z0-9_.]+ *$/ { newpage=1; next; }

/^[ \t]+\[/			{ sentence=1; }
/[^ \t]/			{
				   indent = match($0, /[^ ]/);
				   if (indent < preindent) {
				      sentence = 1;
				   }
				   if (newpage) {
				      if (sentence) {
					 outline++; print "";
				      }
				   } else {
				      if (haveblank) {
					  outline++; print "";
				      }
				   }
				   haveblank=0;
				   sentence=0;
				   newpage=0;

				   line = $0;
				   sub(/^ *\t/, "        ", line);
				   thiscolumn = match(line, /[^ ]/);
				}
/[.:][ \t]*$/			{ sentence=1; }
/\(http:\/\/trustee\.ietf\.org\/license-info\)\./ { sentence=0; }
/^[ \t]*$/			{ haveblank=1; next; }
				{ outline++; print; }
' "$1"
}


# ----------------------------------------------------------------------
# Strip preamble (title, boilerplate and table of contents) and
# postamble (Intellectual Property Statement, Disclaimer etc)
# ----------------------------------------------------------------------
bodystrip() {
    $AWK '
    /^[ \t]*Acknowledgment/		{ inbody = 0; }
    /^(Full )*Copyright Statement$/	{ inbody = 0; }
    /^[ \t]*Disclaimer of Validid/	{ inbody = 0; }
    /^[ \t]*Intellectual Property/	{ inbody = 0; }
    /^Abstract$/			{ inbody = 0; }
    /^Table of Contents$/		{ inbody = 0; }
    /^1.[ \t]*Introduction$/	{ inbody = 1; }

    inbody			{ print; }
    ' "$1"
}


# ----------------------------------------------------------------------
# From two words, find common prefix and differing part, join descriptively
# ----------------------------------------------------------------------
worddiff() {
   $AWK '
BEGIN	{
		w1 = ARGV[1]
		w2 = ARGV[2]
		format = ARGV[3]

		do {
			if (substr(w1,1,1) == substr(w2,1,1)) {
				w1 = substr(w1,2)
				w2 = substr(w2,2)
			} else {
				break;
			}
			prefixlen++;
		} while (length(w1) && length(w2))

		prefix = substr(ARGV[1],1,prefixlen);

		do {
			l1 = length(w1);
			l2 = length(w2);
			if (substr(w1,l1,1) == substr(w2,l2,1)) {
				w1 = substr(w1,1,l1-1)
				w2 = substr(w2,1,l2-1)
			} else {
				break;
			}
		} while (l1 && l2)

		suffix = substr(ARGV[1], prefixlen+length(w1))

		printf format, prefix, w1, w2, suffix;
	}
' "$1" "$2" "$3"
}

# ----------------------------------------------------------------------
# Generate a html page with side-by-side diff from a unified diff
# ----------------------------------------------------------------------
htmldiff() {
   $AWK '
BEGIN	{
           FS = "[ \t,]";

	   # Read pagecache1
	   maxpage[1] = 1
	   pageend[1,0] = 2;
	   while ( getline < ENVIRON["pagecache1"] > 0) {
	      pageend[1,$1] = $2;
	      if ($1+0 > maxpage[1]) maxpage[1] = $1+0;
	   }

	   # Read pagecache2
	   maxpage[2] = 1
	   pageend[2,0] = 2;
	   while ( getline < ENVIRON["pagecache2"] > 0) {
	      pageend[2,$1] = $2;
	      if ($1+0 > maxpage[2]) maxpage[2] = $1+0;
	   }

	   wdiff = ENVIRON["wdiffbin"]
	   base1 = ENVIRON["base1"]
	   base2 = ENVIRON["base2"]
	   optwidth = ENVIRON["optwidth"]
	   optnums =  ENVIRON["optnums"]
	   optlinks = ENVIRON["optlinks"]
	   cmdline = ENVIRON["cmdline"]
	   gsub("--", "- -", cmdline)
	   ENVIRON["cmdline"] = cmdline
	   header(base1, base2)

	   difflines1 = 0
	   difflines2 = 0
	}

function header(file1, file2) {
   url1 = file1;
   url2 = file2;
   if (optlinks) {
      if (file1 ~ /^draft-/) { url1 = sprintf("<a href=\"/html/%s\" style=\"color:#008\">%s</a>", file1, file1); }
      if (file1 ~ /^draft-/) { prev = sprintf("<a href=\"/rfcdiff?url2=%s\" style=\"color:#008; text-decoration:none;\">&lt;</a>", file1); }
      if (file2 ~ /^draft-/) { url2 = sprintf("<a href=\"/html/%s\" style=\"color:#008\">%s</a>", file2, file2); }
      if (file2 ~ /^draft-/) { nxt  = sprintf("<a href=\"/rfcdiff?url1=%s\" style=\"color:#008; text-decoration:none;\">&gt;</a>", file2) }
   }   
   printf "" \
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\"> \n" \
"<!-- Generated by rfcdiff %s: rfcdiff %s --> \n" \
"<!-- <!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional\" > -->\n" \
"<!-- System: %s --> \n" \
"<!-- Using awk: %s: %s --> \n" \
"<!-- Using diff: %s: %s --> \n" \
"<!-- Using wdiff: %s: %s --> \n" \
"<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\"> \n" \
"<head> \n" \
"  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" /> \n" \
"  <meta http-equiv=\"Content-Style-Type\" content=\"text/css\" /> \n" \
"  <title>Diff: %s - %s</title> \n" \
"  <style type=\"text/css\"> \n" \
"    body    { font-size:16px; margin: 0.4ex; margin-right: auto; } \n" \
"    tr      { } \n" \
"    td      { white-space: pre; font-family: Consolas, \"DejaVu Sans Mono\", \"Droid Sans Mono\", Monaco, Monospace; vertical-align: top; font-size: 0.86em;} \n" \
"    th      { font-size: 0.86em; } \n" \
"    .small  { font-size: 0.6em; font-style: italic; font-family: Verdana, Helvetica, sans-serif; } \n" \
"    .left   { background-color: #EEE; } \n" \
"    .right  { background-color: #FFF; } \n" \
"    .diff   { background-color: #CCF; } \n" \
"    .lblock { background-color: #BFB; } \n" \
"    .rblock { background-color: #FF8; } \n" \
"    .insert { background-color: #8FF; } \n" \
"    .delete { background-color: #ACF; } \n" \
"    .void   { background-color: #FFB; } \n" \
"    .cont   { background-color: #EEE; } \n" \
"    .linebr { background-color: #AAA; } \n" \
"    .lineno { color: red; background-color: #FFF; font-size: 0.7em; text-align: right; padding: 0 2px; } \n" \
"    .elipsis{ background-color: #AAA; } \n" \
"    .left .cont { background-color: #DDD; } \n" \
"    .right .cont { background-color: #EEE; } \n" \
"    .lblock .cont { background-color: #9D9; } \n" \
"    .rblock .cont { background-color: #DD6; } \n" \
"    .insert .cont { background-color: #0DD; } \n" \
"    .delete .cont { background-color: #8AD; } \n" \
"    .stats, .stats td, .stats th { background-color: #EEE; padding: 2px 0; } \n" \
"  </style> \n" \
"</head> \n" \
"<body > \n" \
"  <table border=\"0\" cellpadding=\"0\" cellspacing=\"0\"> \n" \
"  <tr bgcolor=\"orange\"><th></th><th>%s&nbsp;%s&nbsp;</th><th> </th><th>&nbsp;%s&nbsp;%s</th><th></th></tr> \n" \
"", ENVIRON["version"], ENVIRON["cmdline"], ENVIRON["uname"], ENVIRON["awkbin"], ENVIRON["awkver"], ENVIRON["diffbin"], ENVIRON["diffver"], ENVIRON["wdiffbin"], ENVIRON["wdiffver"], file1, file2, prev, url1, url2, nxt;
}

function worddiff(w1, w2) {
   prefixlen = 0;
   word1 = w1;
   do {
      if (substr(w1,1,1) == substr(w2,1,1)) {
	 w1 = substr(w1,2);
	 w2 = substr(w2,2);
      } else {
	 break;
      }
      prefixlen++;
   } while (length(w1) && length(w2));

   prefix = substr(word1,1,prefixlen);

   do {
      l1 = length(w1);
      l2 = length(w2);
      if (substr(w1,l1,1) == substr(w2,l2,1)) {
	 w1 = substr(w1,1,l1-1);
	 w2 = substr(w2,1,l2-1);
      } else {
	 break;
      }
   } while (l1 && l2);

   suffix = substr(word1, prefixlen+length(w1)+1);

   wordpart[0] = prefix;
   wordpart[1] = w1;
   wordpart[2] = w2;
   wordpart[3] = suffix;
}

function numdisplay(which, line) {
    if (optnums && (line != prevline[which])) {
	prevline[which] = line;
	return line-1;
    }
    return "";
}

function fixesc(line) {
#    Making this a no-op for now -- the change in line-breaking
#    "<br><span...>" => "\n" should make this less necessary.
#    line = gensub(/&(<[^>]*>)/, "\\1\\&", "g", line);

#   We still have to handle cases where we have a broken up "&lt;" / "&gt;"
    gsub(/&l<\/span>t;/, "\\&lt;</span>", line);
    gsub(/&g<\/span>t;/, "\\&gt;</span>", line);

    gsub(/&<span class="delete">amp;/, "<span class=\"delete\">\\&", line)
    gsub(/&<span class="insert">amp;/, "<span class=\"insert\">\\&", line)

    gsub(/&<span class="delete">lt;/, "<span class=\"delete\">\\&lt;", line);
    gsub(/&<span class="delete">gt;/, "<span class=\"delete\">\\&gt;", line);
    gsub(/&<span class="insert">lt;/, "<span class=\"insert\">\\&lt;", line);
    gsub(/&<span class="insert">gt;/, "<span class=\"insert\">\\&gt;", line);

    gsub(/&<span class="delete">l<\/span>t;/, "<span class=\"delete\">\\&lt;</span>", line);
    gsub(/&<span class="delete">g<\/span>t;/, "<span class=\"delete\">\\&gt;</span>", line);
    gsub(/&<span class="insert">l<\/span>t;/, "<span class=\"insert\">\\&lt;</span>", line);
    gsub(/&<span class="insert">g<\/span>t;/, "<span class=\"insert\">\\&gt;</span>", line);


    return line;
}

function chunkdiff(chunk) {
   if (difflines1 == 0 && difflines2 == 0) return;

   chunkfile1= sprintf("1/chunk%04d", chunk);
   chunkfile2= sprintf("2/chunk%04d", chunk);
   printf "" > chunkfile1;
   printf "" > chunkfile2;
   for (l = 0; l < difflines1; l++) { print stack1[l] >> chunkfile1; }
   for (l = 0; l < difflines2; l++) { print stack2[l] >> chunkfile2; }
   close(chunkfile1);
   close(chunkfile2);

   cmd1 = sprintf("%s -n -2 -w \"<span class=\\\"delete\\\">\"  -x \"</span>\" %s %s", wdiff, chunkfile1, chunkfile2);
   cmd2 = sprintf("%s -n -1 -y \"<span class=\\\"insert\\\">\"  -z \"</span>\" %s %s", wdiff, chunkfile1, chunkfile2);

   l=0; while (cmd1 | getline > 0) { stack1[l] = fixesc($0); l++; }
   difflines1 = l;
   l=0; while (cmd2 | getline > 0) { stack2[l] = fixesc($0); l++; }
   difflines2 = l;

   close(cmd1);
   close(cmd2);
}

function flush() {
   if (difflines1 || difflines2) {
      difftag++;
      multiline = (difflines1 > 1) || (difflines2 > 1);
      if (multiline && (wdiff != "")) chunkdiff(difftag);

      printf "      <tr><td><a name=\"diff%04d\" /></td></tr>\n", difftag;
      for (l = 0; l < difflines1 || l < difflines2; l++) {
	 if (l in stack1) {
	    line1 = stack1[l];
	    delete stack1[l];
	    linenum1++
	    if (line1 == "")
	       if (optwidth > 0) {
		   line1 = substr("                                                                                                                                                                ",0,optwidth);
	       } else {
		   line1 = "                                                                         ";
	       }
	 } else {
	    line1 = "";
	 }
	 if (l in stack2) {
	    line2 = stack2[l];
	    delete stack2[l];
	    linenum2++;
	    if (line2 == "")
	       if (optwidth > 0) {
		   line2 = substr("                                                                                                                                                                ",0,optwidth);
	       } else {
		   line2 = "                                                                         ";
	       }
	 } else {
	    line2 = "";
	 }

	 if (!multiline || (wdiff == "")) {
	    worddiff(line1, line2);
	    line1 = fixesc(sprintf("%s<span class=\"delete\">%s</span>%s", wordpart[0], wordpart[1], wordpart[3]));
	    line2 = fixesc(sprintf("%s<span class=\"insert\">%s</span>%s", wordpart[0], wordpart[2], wordpart[3]));
	    # Clean up; remove empty spans
	    sub(/<span class="delete"><\/span>/,"", line1);
	    sub(/<span class="insert"><\/span>/,"", line2);
	 }
	 left  = sprintf("<td class=\"lineno\" valign=\"top\">%s</td><td class=\"lblock\">%s</td>", numdisplay(1, linenum1), line1);
	 right = sprintf("<td class=\"rblock\">%s</td><td class=\"lineno\" valign=\"top\">%s</td>", line2, numdisplay(2, linenum2));
	 printf "      <tr>%s<td> </td>%s</tr>\n", left, right;
      }
   }
}

function getpage(which, line) {
    line = line + ENVIRON["prelines"];
    page = "?";
    for (p=1; p <= maxpage[which]; p++) {
	if (pageend[which,p] == 0) continue;
	if (line <= pageend[which,p]) {
	    page = p;
	    break;
	}
    }
    return page;
}

function getpageline(which, line, page) {
    if (page == "?") {
	return line + ENVIRON["prelines"];
    } else {
	if (pageend[which,page-1]+0 != 0) {
	    return line + ENVIRON["prelines"] - pageend[which,page-1] + 3; # numbers of header lines stripped
	} else {
	    return "?";
	}
    }
}

function htmlesc(line) {
    gsub("&", "\\&amp;", line);
    gsub("<", "\\&lt;", line);
    gsub(">", "\\&gt;", line);
    return line;
}

function expandtabs(line) {
    spaces = "        ";
    while (pos = index(line, "\t")) {
	sub("\t", substr(spaces, 0, (8-pos%8)), line);
    }
    return line;
}

function maybebreakline(line,    width) {
    width = optwidth;
    new = "";
    if (width > 0) {
	line = expandtabs(line);
	while (length(line) > width) {
	    new = new htmlesc(substr(line, 1, width)) "\n";
	    line = substr(line, width+1);
	}
    }
    line = new htmlesc(line) ;
    return line;
}

/^@@/	{
	   linenum1 = 0 - $2;
	   linenum2 = 0 + $4;
	   diffnum ++;
	   if (linenum1 > 1) {
	      printf "      <tr><td class=\"lineno\"></td><td class=\"left\"></td><td> </td><td class=\"right\"></td><td class=\"lineno\"></td></tr>\n";
	      page1 = getpage(1,linenum1);
	      page2 = getpage(2,linenum2);
	      if (page1 == "?") {
		 posinfo1 = sprintf("<a name=\"part-l%s\" /><small>skipping to change at</small><em> line %s</em>", diffnum, getpageline(1, linenum1, page1));
	      } else {
		 posinfo1 = sprintf("<a name=\"part-l%s\" /><small>skipping to change at</small><em> page %s, line %s</em>", diffnum, page1, getpageline(1, linenum1, page1));
	      }

	      if (page2 == "?") {
		 posinfo2 = sprintf("<a name=\"part-r%s\" /><small>skipping to change at</small><em> line %s</em>", diffnum, getpageline(2, linenum2, page2));
	      } else {
		 posinfo2 = sprintf("<a name=\"part-r%s\" /><small>skipping to change at</small><em> page %s, line %s</em>", diffnum, page2, getpageline(2, linenum2, page2));
	      }

	      printf "      <tr bgcolor=\"gray\" ><td></td><th>%s</th><th> </th><th>%s</th><td></td></tr>\n", posinfo1, posinfo2;
	   }
	}

/^---/	{  next; }
/^[+][+][+]/	{  next; }
/^[ ]/	{
	   line = substr($0, 2);
	   line = maybebreakline(line);

	   flush();
	   linenum1++;
	   linenum2++;
	   printf "      <tr><td class=\"lineno\" valign=\"top\">%s</td><td class=\"left\">%s</td><td> </td>", numdisplay(1, linenum1), line;
	   printf "<td class=\"right\">%s</td><td class=\"lineno\" valign=\"top\">%s</td></tr>\n", line, numdisplay(2, linenum2);
	   diffcount1 += difflines1
	   difflines1 = 0
	   diffcount2 += difflines2
	   difflines2 = 0
	}
/^-/	{
	   line = substr($0, 2);
	   line = maybebreakline(line);

	   stack1[difflines1] = line;
	   difflines1++;
	}
/^[+]/	{
	   line = substr($0, 2);
	   line = maybebreakline(line);

	   stack2[difflines2] = line;
	   difflines2++;
	}

END	{
	   flush();
	   printf("\n" \
"     <tr><td></td><td class=\"left\"></td><td> </td><td class=\"right\"></td><td></td></tr>\n" \
"     <tr bgcolor=\"gray\"><th colspan=\"5\" align=\"center\"><a name=\"end\">&nbsp;%s. %s change blocks.&nbsp;</a></th></tr>\n" \
"     <tr class=\"stats\"><td></td><th><i>%s lines changed or deleted</i></th><th><i> </i></th><th><i>%s lines changed or added</i></th><td></td></tr>\n" \
"     <tr><td colspan=\"5\" align=\"center\" class=\"small\"><br/>This html diff was produced by rfcdiff %s. The latest version is available from <a href=\"http://www.tools.ietf.org/tools/rfcdiff/\" >http://tools.ietf.org/tools/rfcdiff/</a> </td></tr>\n" \
"   </table>\n" \
"   </body>\n" \
"   </html>\n", diffnum?"End of changes":"No changes", difftag, diffcount1, diffcount2, ENVIRON["version"]);
	}
' "$1"
}

# ----------------------------------------------------------------------
# Generate before/after text output from a context diff
# ----------------------------------------------------------------------
abdiff() {
$AWK '
BEGIN	{
	   # Read pagecache1
	   maxpage[1] = 1
	   pageend[1,0] = 2;
	   while ( getline < ENVIRON["pagecache1"] > 0) {
	      pageend[1,$1] = $2;
	      if ($1+0 > maxpage[1]) maxpage[1] = $1+0;
	   }

	   # Read pagecache2
	   maxpage[2] = 1
	   pageend[2,0] = 2;
	   while ( getline < ENVIRON["pagecache2"] > 0) {
	      pageend[2,$1] = $2;
	      if ($1+0 > maxpage[2]) maxpage[2] = $1+0;
	   }

	   base1 = ENVIRON["base1"]
	   base2 = ENVIRON["base2"]

	   section = "INTRODUCTION";
	   para = 0;

	}
/^\+\+/ {
	   next;
	}
/^\-\-/ {
	   next;
	}
/^ Appendix ./	{
	   section = $1 " " $2;
	   para = 0;
	}
/^  ? ? ?[0-9]+(\.[0-9]+)*\.? /	{
	   section = "Section " $1;
	   para = 0;
	}
/^ ?$/	{
	   if (inpara) {
	      printf "\n%s, paragraph %s:\n", section, para;
	      print "OLD:\n"
	      print oldpara
	      print "NEW:\n"
	      print newpara
	   }
	   oldpara = "";
	   newpara = "";
	   para ++;
	   inpara = 0
	}
/^ ./	{
	   oldpara = oldpara $0 "\n"
	   newpara = newpara $0 "\n"
	}
/^\-/	{
	   sub(/^./, " ");
	   oldpara = oldpara $0 "\n"
	   inpara++;
	}
/^\+/	{
	   sub(/^./, " ");
	   newpara = newpara $0 "\n"
	   inpara++;
	}
END	{
	   if (inpara) {
	      printf "\n%s, paragraph %s:\n", section, para;
	      print "OLD:\n"
	      print oldpara
	      print "NEW:\n"
	      print newpara
	   }	
	}
'
}


# ----------------------------------------------------------------------
# Utility to extract keyword info
# ----------------------------------------------------------------------
extract() {
    $AWK -v keyword="$1" '
	BEGIN {
	    # print "Keyword", keyword;
	}
	/^# [A-Z]/ {
	    # print "New key", $2;
	    if ($2 == keyword ":" ) { output=1; } else { output=0; }
	    # print "Output", output;
	}
	/^#\t/	{
	    # print "Content", output, $0;
	    if ( output ) {
		sub(/^#/,"");
		print;
	    }
	}
	{
	    next;
	}

    ' "$2"
}
# ----------------------------------------------------------------------
# Utility to start a browser
# ----------------------------------------------------------------------

browse() {
    if   [ "$(uname)" = "Darwin" ]; then
	open "$@"
    else
	browser=$(lookfor firefox phoenix MozillaFirebird mozilla opera Netscape netscape dillo)

	if [ -z "$browser" ]; then
	    echo "Couldn't find any browser, can't display $*."
	    exit 1
	fi

	# make sure file name is absolute
	if [ "${1#/}" = "$1" ]; then
	    # not absolute path, add pwd
	    arg="file://$PWD/$1"
	else
	    arg="file://$1"
	fi


	# see if a browser is running, act accordingly
	$browser -remote "ping()" >/dev/null 2>&1
	if [ $? -eq 0 ]; then
	    # use running instance
	    $browser -raise -remote "openurl($arg, new-tab)"
	else
	    # error exit: no running instance
	    echo "Starting web browser."

	    $browser "$arg" >/dev/null 2>&1 &
	fi
    fi
}


# ----------------------------------------------------------------------
# Utility for error exit
# ----------------------------------------------------------------------
die() {
   echo $*;
   exit 1;
}

# ----------------------------------------------------------------------
# Process options
# ----------------------------------------------------------------------

# Default values
opthtml=1; optdiff=0; optchbars=0; optwdiff=0; optshow=0; optnowdiff=0;
optkeep=0; optinfo=0; optwidth=0;  optnums=0;  optbody=0; optabdiff=0;
optstrip=1; opthwdiff=0; optlinks=0;
optoldcolour="red"; optnewcolour="green"; optlarger=""
optstdout=0;
opttmpdiff=0; tmpdiff=$workdir/diff;

while [ $# -gt 0 ]; do
   case "$1" in
      --html)   opthtml=1; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=0;;
      --diff)   opthtml=0; optdiff=1; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=0;;
      --chbars) opthtml=0; optdiff=0; optchbars=1; optwdiff=0; opthwdiff=0; optabdiff=0;;
      --wdiff)  opthtml=0; optdiff=0; optchbars=0; optwdiff=1; opthwdiff=0; optabdiff=0;;
      --hwdiff) opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=1; optabdiff=0;;
      --changes)opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --abdiff)	opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --ab-diff)opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --rfc-editor-diff)opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --version)echo -e "$basename\t$version"; exit 0;;
      --browse) optshow=1;;
      --nowdiff)optnowdiff=1;;
      --keep)	optkeep=1;;
      --info)	optinfo=1; keyword=$2; shift;;
      --help)	optinfo=1; keyword="Usage";;
      --width)	optwidth=$2; shift;;
      --oldcolor)     optoldcolour=$2; shift;;
      --oldcolour)    optoldcolour=$2; shift;;
      --newcolor)     optnewcolour=$2; shift;;
      --newcolour)    optnewcolour=$2; shift;;
      --larger)       optlarger='size="+1"';;
      --linenum)optnums=1;;
      --body)	optbody=1;;
      --nostrip)optstrip=0; optbody=0;;
      --stdout) optstdout=1;;
      --links)  optlinks=1;;
      --ignore-space-change) optnospacechange=1;;
      --ignore-all-space) optignorewhite=1;;
      --ignore-blank-lines) optignoreblank=1;;
      --wdiff-args) optwdiffargs=$2; shift;;
      --tmpdiff) opttmpdiff=1; tmpdiff=$2; shift;;
      --prelines)  export prelines=$2; shift;;
      --minimal) optminimal=1;;
      --)	shift; break;;

      -v) echo "$basename $version"; exit 0;;
      -*) echo "Unrecognized option: $1";
	  exit 1;;
      *)  break;;
   esac
   shift
done

export optwidth
export optnums
export optlinks

# ----------------------------------------------------------------------
# Determine output file name. Maybe output usage and exit.
# ----------------------------------------------------------------------
#set -x

if [ $optinfo -gt 0 ]; then
   extract $keyword $0
   exit
fi
if [ $# -ge 2 ]; then
   if [ "$1" = "$2" ]; then
      echo "The files are the same file"
      exit
   fi
   export base1=$(basename "$1")
   export base2=$(basename "$2")
   outbase=$(worddiff "$base2" "$base1" "%s%s-from-%s")
else
   extract Usage $0
   exit 1
fi


# ----------------------------------------------------------------------
# create working directory.
# ----------------------------------------------------------------------
mkdir $workdir/1 || die "$0: Error: Failed to create temporary directory '$workdir/1'."
mkdir $workdir/2 || die "$0: Error: Failed to create temporary directory '$workdir/2'."

# ----------------------------------------------------------------------
# If any of the files is an http or ftp URL we download it, else copy it
# ----------------------------------------------------------------------

wgetbin=$(lookfor wget)
dowgetarg1=0
dowgetarg2=0

if [ -n "$wgetbin" ]; then
   if [ "${1#http://}" != "$1" ]; then dowgetarg1=1; fi
   if [ "${1#ftp://}" != "$1" ]; then dowgetarg1=1; fi

   if [ "${2#http://}" != "$2" ]; then dowgetarg2=1; fi
   if [ "${2#ftp://}" != "$2" ]; then dowgetarg2=1; fi
fi


if [ $dowgetarg1 -gt 0 ]; then
   $wgetbin -nv "$1" -O $workdir/1/"$base1"
else
   cp "$1" $workdir/1/"$base1"
fi

if [ $dowgetarg2 -gt 0 ]; then
   $wgetbin -nv "$2" -O $workdir/2/"$base2"
else
   cp "$2" $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Maybe strip headers/footers from both files
# ----------------------------------------------------------------------

if [ $optstrip -gt 0 ]; then
   export which=1
   strip $workdir/1/"$base1" > $workdir/1/"$base1".stripped
   mv -f $workdir/1/"$base1".stripped $workdir/1/"$base1"
   export which=2
   strip $workdir/2/"$base2" > $workdir/2/"$base2".stripped
   mv -f $workdir/2/"$base2".stripped $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Maybe do html quoting
# ----------------------------------------------------------------------

if [ $opthwdiff -gt 0 ]; then
   sed -e 's/&/&amp;/g' -e 's/</\&lt;/g' -e 's/>/\&gt;/g' $workdir/1/"$base1" > $workdir/1/"$base1".quoted
   mv -f $workdir/1/"$base1".quoted $workdir/1/"$base1"
   sed -e 's/&/&amp;/g' -e 's/</\&lt;/g' -e 's/>/\&gt;/g' $workdir/2/"$base2" > $workdir/2/"$base2".quoted
   mv -f $workdir/2/"$base2".quoted $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Maybe strip preamble (title, boilerplate and table of contents) and
# postamble (Intellectual Property Statement, Disclaimer etc)
# ----------------------------------------------------------------------
if [ $optbody -gt 0 ]; then
   bodystrip $workdir/1/"$base1" > $workdir/1/"$base1".stripped
   mv $workdir/1/"$base1".stripped $workdir/1/"$base1"
   bodystrip $workdir/2/"$base2" > $workdir/2/"$base2".stripped
   mv $workdir/2/"$base2".stripped $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Get output file name
# ----------------------------------------------------------------------
if [ "$3" ]; then
  outfile="$3"
else
    if [ $opthtml -gt 0 ]; then
      outfile=./"$outbase".diff.html
    fi
    if [ $optchbars -gt 0 ]; then
      outfile=./"$outbase".chbar
    fi
    if [ $optdiff -gt 0 ]; then
      outfile=./"$outbase".diff
    fi
    if [ $optabdiff -gt 0 ]; then
      outfile=./"$outbase".changes
    fi
    if [ $opthwdiff -gt 0 ]; then
      outfile=./"$outbase".wdiff.html
    fi
fi
if [ "$outfile" ]; then
   tempout=./$(basename "$outfile")
fi

# ----------------------------------------------------------------------
# Check if we can use wdiff for block diffs
# ----------------------------------------------------------------------
if [ $optnowdiff -eq 0 ]; then
   wdiffbin=$(lookfor wdiff)
   if [ -n "$wdiffbin" ]; then
      wdiffver=$($wdiffbin --version 2>/dev/null | egrep "(wdiff|GNU).+[0-9]\.[0-9]")
      if [ -z "$wdiffver" ]; then
        wdiffbin="";
	echo -en "\n  Found wdiff, but it reported no recognisable version."
      fi
   else
      echo -en "\n  Couldn't find wdiff."
   fi
   if [ -z "$wdiffbin" ]; then echo " Falling back to builtin diff colouring..."; fi
   export wdiffbin
   export wdiffver
   #echo "Found wdiff at $wdiffbin"
fi

# ----------------------------------------------------------------------
# Get some misc. info
# ----------------------------------------------------------------------
uname=$(uname -a)
export uname
awkbin=$AWK
export awkbin
awkver=$( { $AWK --version 2>/dev/null || $AWK -V 2>/dev/null; } | head -n 1)
export awkver
diffbin=$(lookfor diff)
export diffbin
diffver=$(diff --version | head -n 1)
export diffver

# ----------------------------------------------------------------------
# Check that we don't have a broken awk
# ----------------------------------------------------------------------
if [ $opthtml -gt 0 -a "${uname%% *}" == "Darwin" -a "$awkver" == "awk version 20070501" ]; then
    echo -e  "\n  Oops.  Awk version 20070501 on OS X doesn't work with rfcdiff's html mode.\n  To make rfcdiff work, you could install Gnu Awk (gawk), for instance using\n  MacPorts, http://www.macports.org/."
    exit 1
fi

# ----------------------------------------------------------------------
# Do diff
# ----------------------------------------------------------------------

origdir=$PWD
cd $workdir
if cmp 1/"$base1" 2/"$base2" >/dev/null; then
   echo ""
   echo "The files are identical."
fi

if [ $opthtml -gt 0 ]; then
   diff ${optignoreblank:+-B} ${optminimal:+-d} ${optnospacechange:+-b} ${optignorewhite:+-w} -U $prelines 1/"$base1" 2/"$base2" | tee $tmpdiff | htmldiff > "$tempout"
fi
if [ $optchbars -gt 0 ]; then
   diff -Bwd -U 10000 1/"$base1" 2/"$base2" | tee $tmpdiff | grep -v "^-" | tail -n +3 | sed 's/^+/|/' > "$tempout"
fi
if [ $optdiff -gt 0 ]; then
   diff -Bwd -U $prelines 1/"$base1" 2/"$base2" | tee $tmpdiff > "$tempout"
fi
if [ $optabdiff -gt 0 ]; then
   diff -wd -U 1000 1/"$base1" 2/"$base2" | tee $tmpdiff | abdiff
fi
if [ $optwdiff -gt 0 ]; then
   wdiff -a $optwdiffargs 1/"$base1" 2/"$base2"
fi
if [ $opthwdiff -gt 0 ]; then
    echo "<html><head><title>wdiff "$base1" "$base2"</title></head><body>"	>  "$tempout"
    echo "<pre>"								>> "$tempout"
    wdiff -w "<strike><font color='$optoldcolour' $optlarger>" -x "</font></strike>"	\
          -y "<strong><font color='$optnewcolour' $optlarger>" -z "</font></strong>"	\
	  1/"$base1" 2/"$base2"							>> "$tempout"
    echo "</pre>"								>> "$tempout"
    echo "</body></html>"							>> "$tempout"
fi

if [ $optstdout -gt 0 ]; then
  cat "$tempout"
  rm  "$tempout"
else
  cd "$origdir"; if [ -f $workdir/"$tempout" ]; then mv $workdir/"$tempout" "$outfile"; fi
fi

if [ $optshow -gt 0 ]; then
   browse "$outfile"
fi

if [ $optkeep -eq 0 ]; then
   if [ -f $pagecache1 ]; then rm $pagecache1; fi
   if [ -f $pagecache2 ]; then rm $pagecache2; fi
   rm -fr $workdir/1
   rm -fr $workdir/2
   if [ -f $tmpdiff ]; then
      if [ $opttmpdiff -eq 0 ]; then
          rm $tmpdiff
      fi
   fi
   rmdir $workdir
else
   cd /tmp
   tar czf $basename-$$.tgz $basename-$$
   echo "
   Temporary workfiles have been left in $workdir/, and packed up in $workdir.tgz"
fi

#!/bin/bash
#
# Synopsis:
#	Show changes between 2 internet-drafts using changebars or html
#	side-by-side diff.
#	
# Usage:
#	rfcdiff [options] file1 file2
#	
#	rfcdiff takes two RFCs or Internet-Drafts in text form as input, and
#	produces output which indicates the differences found in one of various
#	forms, controlled by the options listed below. In all cases, page
#	headers and page footers are stripped before looking for changes.
#	
#	--html		Produce side-by-side .html diff (default)
#	
#	--chbars	Produce changebar marked .txt output
#	
#	--diff		Produce a regular diff output
#	
#	--wdiff		Produce paged wdiff output
#	
#	--hwdiff	Produce html-wrapped coloured wdiff output
#	
#	--oldcolour COLOURNAME	Colour for new file in hwdiff (default is "green")
#	--oldcolor COLORNAME	Color for old file in hwdiff (default is "red")
#
#	--newcolour COLOURNAME	Colour for new file in hwdiff (default is "green")
#	--newcolor COLORNAME	Color for new file in hwdiff (default is "green")
#
#	--larger        Make difference text in hwdiff slightly larger
#	
#	--keep		Don't delete temporary workfiles
#	
#	--version	Show version
#	
#	--help		Show this help
#	
#	--info "Synopsis|Usage|Copyright|Description|Log"
#			Show various info
#	
#	--width	N	Set a maximum width of N characters for the
#			display of each half of the old/new html diff
#	
#	--linenum	Show linenumbers for each line, not only at the
#			start of each change section
#	
#	--body		Strip document preamble (title, boilerplate and
#			table of contents) and postamble (Intellectual
#			Property Statement, Disclaimer etc)
#	
#	--nostrip	Don't strip headers and footers (or body)
#	
#	--ab-diff	Before/After diff, suitable for rfc-editor
#	--abdiff
#	
#	--stdout	Send output to stdout instead to a file
#	
#       --tmpdiff       Path to intermediate diff file
#
#       --prelines N    Set value for diff -U option
#
#       --minimal       Set value for diff -d option
#
#       --ignore-space-change
#                       Ignore changes in the amount of white space.
#
#       --ignore-all-space
#                       Ignore all white space.
#
#       --ignore-blank-lines
#                       Ignore changes whose lines are all blank.
#       
#
# Copyright:
#	-----------------------------------------------------------------
#	
#	Copyright 2002 Henrik Levkowetz
#	
#	This program is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; either version 2 of the License, or
#	(at your option) any later version.
#	
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	-----------------------------------------------------------------
#
# Description:
#
#	The purpose of this program is to compare two versions of an
#	internet-draft, and as output produce a diff in one of several
#	formats:
#	
#		- side-by-side html diff
#		- paged wdiff output in a text terminal
#		- a text file with changebars in the left margin
#		- a simple unified diff output
#	
#	In all cases, internet-draft headers and footers are stripped before
#	generating the diff, to produce a cleaner diff.
#	
#	It is called as
#	
#		rfcdiff first-file second-file
#	
#	The latest version is available from
#		http://tools.ietf.org/tools/rfcdiff/
#

export version="1.41"
export prelines="10"
export basename=$(basename $0)
export workdir=`mktemp -d -t rfcdiff.XXXXXXXX`
export pagecache1="$workdir/pagecache1"
export pagecache2="$workdir/pagecache2"

# ----------------------------------------------------------------------
# Utility to find an executable
# ----------------------------------------------------------------------
lookfor() {
    for b in "$@"; do
	found=$(which "$b" 2>/dev/null)
	if [ -n "$found" ]; then
	    if [ -x "$found" ]; then
		echo "$found"
		return
	    fi
	fi
    done
}

AWK=awk

# ----------------------------------------------------------------------
# Strip headers footers and formfeeds from infile to stdout
# ----------------------------------------------------------------------
strip() {
  $AWK '
				{ gsub(/\r/, ""); }
				{ gsub(/[ \t]+$/, ""); }
				{ pagelength++; }
/\[?[Pp]age [0-9ivx]+\]?[ \t\f]*$/	{
				    match($0, /[Pp]age [0-9ivx]+/);
				    num = substr($0, RSTART+5, RLENGTH-5);
				    print num, outline > ENVIRON["pagecache" ENVIRON["which"]]
				    pagelength = 0;
				}
/\f/				{ newpage=1;
				  pagelength=1;
				}
/\f$/				{
				    # a form feed followed by a \n does not contribute to the
				    # line count.  (But a \f followed by something else does.)
				    pagelength--;
				}
/\f/				{ next; }
/\[?[Pp]age [0-9ivx]+\]?[ \t\f]*$/		{ preindent = indent; next; }

/^ *Internet.Draft.+[12][0-9][0-9][0-9] *$/ && (FNR > 15)	{ newpage=1; next; }
/^ *INTERNET.DRAFT.+[12][0-9][0-9][0-9] *$/ && (FNR > 15)	{ newpage=1; next; }
/^ *Draft.+(  +)[12][0-9][0-9][0-9] *$/	    && (FNR > 15)	{ newpage=1; next; }
/^RFC[ -]?[0-9]+.*(  +).* [12][0-9][0-9][0-9]$/ && (FNR > 15)	{ newpage=1; next; }
/^draft-[-a-z0-9_.]+.*[0-9][0-9][0-9][0-9]$/ && (FNR > 15)	{ newpage=1; next; }
/(Jan|Feb|Mar|March|Apr|April|May|Jun|June|Jul|July|Aug|Sep|Oct|Nov|Dec) (19[89][0-9]|20[0-9][0-9]) *$/ && pagelength < 3  { newpage=1; next; }
newpage && $0 ~ /^ *draft-[-a-z0-9_.]+ *$/ { newpage=1; next; }

/^[ \t]+\[/			{ sentence=1; }
/[^ \t]/			{
				   indent = match($0, /[^ ]/);
				   if (indent < preindent) {
				      sentence = 1;
				   }
				   if (newpage) {
				      if (sentence) {
					 outline++; print "";
				      }
				   } else {
				      if (haveblank) {
					  outline++; print "";
				      }
				   }
				   haveblank=0;
				   sentence=0;
				   newpage=0;

				   line = $0;
				   sub(/^ *\t/, "        ", line);
				   thiscolumn = match(line, /[^ ]/);
				}
/[.:][ \t]*$/			{ sentence=1; }
/\(http:\/\/trustee\.ietf\.org\/license-info\)\./ { sentence=0; }
/^[ \t]*$/			{ haveblank=1; next; }
				{ outline++; print; }
' "$1"
}


# ----------------------------------------------------------------------
# Strip preamble (title, boilerplate and table of contents) and
# postamble (Intellectual Property Statement, Disclaimer etc)
# ----------------------------------------------------------------------
bodystrip() {
    $AWK '
    /^[ \t]*Acknowledgment/		{ inbody = 0; }
    /^(Full )*Copyright Statement$/	{ inbody = 0; }
    /^[ \t]*Disclaimer of Validid/	{ inbody = 0; }
    /^[ \t]*Intellectual Property/	{ inbody = 0; }
    /^Abstract$/			{ inbody = 0; }
    /^Table of Contents$/		{ inbody = 0; }
    /^1.[ \t]*Introduction$/	{ inbody = 1; }

    inbody			{ print; }
    ' "$1"
}


# ----------------------------------------------------------------------
# From two words, find common prefix and differing part, join descriptively
# ----------------------------------------------------------------------
worddiff() {
   $AWK '
BEGIN	{
		w1 = ARGV[1]
		w2 = ARGV[2]
		format = ARGV[3]

		do {
			if (substr(w1,1,1) == substr(w2,1,1)) {
				w1 = substr(w1,2)
				w2 = substr(w2,2)
			} else {
				break;
			}
			prefixlen++;
		} while (length(w1) && length(w2))

		prefix = substr(ARGV[1],1,prefixlen);

		do {
			l1 = length(w1);
			l2 = length(w2);
			if (substr(w1,l1,1) == substr(w2,l2,1)) {
				w1 = substr(w1,1,l1-1)
				w2 = substr(w2,1,l2-1)
			} else {
				break;
			}
		} while (l1 && l2)

		suffix = substr(ARGV[1], prefixlen+length(w1))

		printf format, prefix, w1, w2, suffix;
	}
' "$1" "$2" "$3"
}

# ----------------------------------------------------------------------
# Generate a html page with side-by-side diff from a unified diff
# ----------------------------------------------------------------------
htmldiff() {
   $AWK '
BEGIN	{
           FS = "[ \t,]";

	   # Read pagecache1
	   maxpage[1] = 1
	   pageend[1,0] = 2;
	   while ( getline < ENVIRON["pagecache1"] > 0) {
	      pageend[1,$1] = $2;
	      if ($1+0 > maxpage[1]) maxpage[1] = $1+0;
	   }

	   # Read pagecache2
	   maxpage[2] = 1
	   pageend[2,0] = 2;
	   while ( getline < ENVIRON["pagecache2"] > 0) {
	      pageend[2,$1] = $2;
	      if ($1+0 > maxpage[2]) maxpage[2] = $1+0;
	   }

	   wdiff = ENVIRON["wdiffbin"]
	   base1 = ENVIRON["base1"]
	   base2 = ENVIRON["base2"]
	   optwidth = ENVIRON["optwidth"]
	   optnums =  ENVIRON["optnums"]
	   optlinks = ENVIRON["optlinks"]
	   header(base1, base2)

	   difflines1 = 0
	   difflines2 = 0
	}

function header(file1, file2) {
   url1 = file1;
   url2 = file2;
   if (optlinks) {
      if (file1 ~ /^draft-/) { url1 = sprintf("<a href=\"/html/%s\" style=\"color:#008\">%s</a>", file1, file1); }
      if (file1 ~ /^draft-/) { prev = sprintf("<a href=\"/rfcdiff?url2=%s\" style=\"color:#008; text-decoration:none;\">&lt;</a>", file1); }
      if (file2 ~ /^draft-/) { url2 = sprintf("<a href=\"/html/%s\" style=\"color:#008\">%s</a>", file2, file2); }
      if (file2 ~ /^draft-/) { nxt  = sprintf("<a href=\"/rfcdiff?url1=%s\" style=\"color:#008; text-decoration:none;\">&gt;</a>", file2) }
   }   
   printf "" \
"<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\"> \n" \
"<!-- <!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional\" > -->\n" \
"<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\"> \n" \
"<head> \n" \
"  <meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" /> \n" \
"  <meta http-equiv=\"Content-Style-Type\" content=\"text/css\" /> \n" \
"  <title>Diff: %s - %s</title> \n" \
"  <style type=\"text/css\"> \n" \
"    body    { font-family: Arial, sans-serif; margin: 0.4ex; margin-right: auto; } \n" \
"    table   { width: 100%; } \n" \
"    tr      { } \n" \
"    td      { white-space: pre; font-family: Consolas, \"DejaVu Sans Mono\", \"Droid Sans Mono\", Monaco, Monospace; vertical-align: top;} \n" \
"    th      { font-size: 0.86em; } \n" \
"    .small  { font-size: 0.6em; font-style: italic; font-family: Verdana, Helvetica, sans-serif; } \n" \
"    .left   { background-color: #EEE; } \n" \
"    .right  { background-color: #FFF; } \n" \
"    .diff   { background-color: #CCF; } \n" \
"    .lblock { background-color: #BFB; } \n" \
"    .rblock { background-color: #FF8; } \n" \
"    .insert { background-color: #8FF; } \n" \
"    .delete { background-color: #ACF; } \n" \
"    .void   { background-color: #FFB; } \n" \
"    .cont   { background-color: #EEE; } \n" \
"    .linebr { background-color: #AAA; } \n" \
"    .lineno { color: red; background-color: #FFF; font-size: 0.7em; text-align: right; padding: 0 2px; } \n" \
"    .elipsis{ background-color: #AAA; } \n" \
"    .left .cont { background-color: #DDD; } \n" \
"    .right .cont { background-color: #EEE; } \n" \
"    .lblock .cont { background-color: #9D9; } \n" \
"    .rblock .cont { background-color: #DD6; } \n" \
"    .insert .cont { background-color: #0DD; } \n" \
"    .delete .cont { background-color: #8AD; } \n" \
"    .stats, .stats td, .stats th { background-color: #EEE; padding: 2px 0; } \n" \
"  </style> \n" \
"</head> \n" \
"<body > \n" \
"  <table border=\"0\" cellpadding=\"0\" cellspacing=\"0\"> \n" \
"  <tr bgcolor=\"orange\"><th></th><th>%s&nbsp;%s&nbsp;</th><th> </th><th>&nbsp;%s&nbsp;%s</th><th></th></tr> \n" \
"", file1, file2, prev, url1, url2, nxt;
}

function worddiff(w1, w2) {
   prefixlen = 0;
   word1 = w1;
   do {
      if (substr(w1,1,1) == substr(w2,1,1)) {
	 w1 = substr(w1,2);
	 w2 = substr(w2,2);
      } else {
	 break;
      }
      prefixlen++;
   } while (length(w1) && length(w2));

   prefix = substr(word1,1,prefixlen);

   do {
      l1 = length(w1);
      l2 = length(w2);
      if (substr(w1,l1,1) == substr(w2,l2,1)) {
	 w1 = substr(w1,1,l1-1);
	 w2 = substr(w2,1,l2-1);
      } else {
	 break;
      }
   } while (l1 && l2);

   suffix = substr(word1, prefixlen+length(w1)+1);

   wordpart[0] = prefix;
   wordpart[1] = w1;
   wordpart[2] = w2;
   wordpart[3] = suffix;
}

function numdisplay(which, line) {
    if (optnums && (line != prevline[which])) {
	prevline[which] = line;
	return line-1;
    }
    return "";
}

function fixesc(line) {
#    Making this a no-op for now -- the change in line-breaking
#    "<br><span...>" => "\n" should make this less necessary.
#    line = gensub(/&(<[^>]*>)/, "\\1\\&", "g", line);

#   We still have to handle cases where we have a broken up "&lt;" / "&gt;"
    gsub(/&l<\/span>t;/, "\\&lt;</span>", line);
    gsub(/&g<\/span>t;/, "\\&gt;</span>", line);

    gsub(/&<span class="delete">amp;/, "<span class=\"delete\">\\&", line)
    gsub(/&<span class="insert">amp;/, "<span class=\"insert\">\\&", line)

    gsub(/&<span class="delete">lt;/, "<span class=\"delete\">\\&lt;", line);
    gsub(/&<span class="delete">gt;/, "<span class=\"delete\">\\&gt;", line);
    gsub(/&<span class="insert">lt;/, "<span class=\"insert\">\\&lt;", line);
    gsub(/&<span class="insert">gt;/, "<span class=\"insert\">\\&gt;", line);

    gsub(/&<span class="delete">l<\/span>t;/, "<span class=\"delete\">\\&lt;</span>", line);
    gsub(/&<span class="delete">g<\/span>t;/, "<span class=\"delete\">\\&gt;</span>", line);
    gsub(/&<span class="insert">l<\/span>t;/, "<span class=\"insert\">\\&lt;</span>", line);
    gsub(/&<span class="insert">g<\/span>t;/, "<span class=\"insert\">\\&gt;</span>", line);


    return line;
}

function chunkdiff(chunk) {
   if (difflines1 == 0 && difflines2 == 0) return;

   chunkfile1= sprintf("1/chunk%04d", chunk);
   chunkfile2= sprintf("2/chunk%04d", chunk);
   printf "" > chunkfile1;
   printf "" > chunkfile2;
   for (l = 0; l < difflines1; l++) { print stack1[l] >> chunkfile1; }
   for (l = 0; l < difflines2; l++) { print stack2[l] >> chunkfile2; }
   close(chunkfile1);
   close(chunkfile2);

   cmd1 = sprintf("%s -n -2 -w \"<span class=\\\"delete\\\">\"  -x \"</span>\" %s %s", wdiff, chunkfile1, chunkfile2);
   cmd2 = sprintf("%s -n -1 -y \"<span class=\\\"insert\\\">\"  -z \"</span>\" %s %s", wdiff, chunkfile1, chunkfile2);

   l=0; while (cmd1 | getline > 0) { stack1[l] = fixesc($0); l++; }
   difflines1 = l;
   l=0; while (cmd2 | getline > 0) { stack2[l] = fixesc($0); l++; }
   difflines2 = l;

   close(cmd1);
   close(cmd2);
}

function flush() {
   if (difflines1 || difflines2) {
      difftag++;
      multiline = (difflines1 > 1) || (difflines2 > 1);
      if (multiline && (wdiff != "")) chunkdiff(difftag);

      printf "      <tr><td><a name=\"diff%04d\" /></td></tr>\n", difftag;
      for (l = 0; l < difflines1 || l < difflines2; l++) {
	 if (l in stack1) {
	    line1 = stack1[l];
	    delete stack1[l];
	    linenum1++
	    if (line1 == "")
	       if (optwidth > 0) {
		   line1 = substr("                                                                                                                                                                ",0,optwidth);
	       } else {
		   line1 = "                                                                         ";
	       }
	 } else {
	    line1 = "";
	 }
	 if (l in stack2) {
	    line2 = stack2[l];
	    delete stack2[l];
	    linenum2++;
	    if (line2 == "")
	       if (optwidth > 0) {
		   line2 = substr("                                                                                                                                                                ",0,optwidth);
	       } else {
		   line2 = "                                                                         ";
	       }
	 } else {
	    line2 = "";
	 }

	 if (!multiline || (wdiff == "")) {
	    worddiff(line1, line2);
	    line1 = fixesc(sprintf("%s<span class=\"delete\">%s</span>%s", wordpart[0], wordpart[1], wordpart[3]));
	    line2 = fixesc(sprintf("%s<span class=\"insert\">%s</span>%s", wordpart[0], wordpart[2], wordpart[3]));
	    # Clean up; remove empty spans
	    sub(/<span class="delete"><\/span>/,"", line1);
	    sub(/<span class="insert"><\/span>/,"", line2);
	 }
	 left  = sprintf("<td class=\"lineno\" valign=\"top\">%s</td><td class=\"lblock\">%s</td>", numdisplay(1, linenum1), line1);
	 right = sprintf("<td class=\"rblock\">%s</td><td class=\"lineno\" valign=\"top\">%s</td>", line2, numdisplay(2, linenum2));
	 printf "      <tr>%s<td> </td>%s</tr>\n", left, right;
      }
   }
}

function getpage(which, line) {
    line = line + ENVIRON["prelines"];
    page = "?";
    for (p=1; p <= maxpage[which]; p++) {
	if (pageend[which,p] == 0) continue;
	if (line <= pageend[which,p]) {
	    page = p;
	    break;
	}
    }
    return page;
}

function getpageline(which, line, page) {
    if (page == "?") {
	return line + ENVIRON["prelines"];
    } else {
	if (pageend[which,page-1]+0 != 0) {
	    return line + ENVIRON["prelines"] - pageend[which,page-1] + 3; # numbers of header lines stripped
	} else {
	    return "?";
	}
    }
}

function htmlesc(line) {
    gsub("&", "\\&amp;", line);
    gsub("<", "\\&lt;", line);
    gsub(">", "\\&gt;", line);
    return line;
}

function expandtabs(line) {
    spaces = "        ";
    while (pos = index(line, "\t")) {
	sub("\t", substr(spaces, 0, (8-pos%8)), line);
    }
    return line;
}

function maybebreakline(line,    width) {
    width = optwidth;
    new = "";
    if (width > 0) {
	line = expandtabs(line);
	while (length(line) > width) {
	    new = new htmlesc(substr(line, 1, width)) "\n";
	    line = substr(line, width+1);
	}
    }
    line = new htmlesc(line) ;
    return line;
}

/^@@/	{
	   linenum1 = 0 - $2;
	   linenum2 = 0 + $4;
	   diffnum ++;
	   if (linenum1 > 1) {
	      printf "      <tr><td class=\"lineno\"></td><td class=\"left\"></td><td> </td><td class=\"right\"></td><td class=\"lineno\"></td></tr>\n";
	      page1 = getpage(1,linenum1);
	      page2 = getpage(2,linenum2);
	      if (page1 == "?") {
		 posinfo1 = sprintf("<a name=\"part-l%s\" /><small>skipping to change at</small><em> line %s</em>", diffnum, getpageline(1, linenum1, page1));
	      } else {
		 posinfo1 = sprintf("<a name=\"part-l%s\" /><small>skipping to change at</small><em> page %s, line %s</em>", diffnum, page1, getpageline(1, linenum1, page1));
	      }

	      if (page2 == "?") {
		 posinfo2 = sprintf("<a name=\"part-r%s\" /><small>skipping to change at</small><em> line %s</em>", diffnum, getpageline(2, linenum2, page2));
	      } else {
		 posinfo2 = sprintf("<a name=\"part-r%s\" /><small>skipping to change at</small><em> page %s, line %s</em>", diffnum, page2, getpageline(2, linenum2, page2));
	      }

	      printf "      <tr bgcolor=\"#c0c0c0\" ><td></td><th>%s</th><th> </th><th>%s</th><td></td></tr>\n", posinfo1, posinfo2;
	   }
	}

/^---/	{  next; }
/^[+][+][+]/	{  next; }
/^[ ]/	{
	   line = substr($0, 2);
	   line = maybebreakline(line);

	   flush();
	   linenum1++;
	   linenum2++;
	   printf "      <tr><td class=\"lineno\" valign=\"top\">%s</td><td class=\"left\">%s</td><td> </td>", numdisplay(1, linenum1), line;
	   printf "<td class=\"right\">%s</td><td class=\"lineno\" valign=\"top\">%s</td></tr>\n", line, numdisplay(2, linenum2);
	   diffcount1 += difflines1
	   difflines1 = 0
	   diffcount2 += difflines2
	   difflines2 = 0
	}
/^-/	{
	   line = substr($0, 2);
	   line = maybebreakline(line);

	   stack1[difflines1] = line;
	   difflines1++;
	}
/^[+]/	{
	   line = substr($0, 2);
	   line = maybebreakline(line);

	   stack2[difflines2] = line;
	   difflines2++;
	}

END	{
	   flush();
	   printf("\n" \
"     <tr><td></td><td class=\"left\"></td><td> </td><td class=\"right\"></td><td></td></tr>\n" \
"     <tr bgcolor=\"#c0c0c0\"><th colspan=\"5\" align=\"center\"><a name=\"end\">&nbsp;%s. %s change block(s).&nbsp;</a></th></tr>\n" \
"     <tr class=\"stats\"><td></td><th><i>%s line(s) changed or deleted</i></th><th><i> </i></th><th><i>%s line(s) changed or added</i></th><td></td></tr>\n" \
"     <tr><td colspan=\"5\" align=\"center\" class=\"small\"><br/>This html diff was produced by rfcdiff %s</td></tr>\n" \
"   </table>\n" \
"   </body>\n" \
"   </html>\n", diffnum?"End of changes":"No changes", difftag, diffcount1, diffcount2, ENVIRON["version"]);
	}
' "$1"
}

# ----------------------------------------------------------------------
# Generate before/after text output from a context diff
# ----------------------------------------------------------------------
abdiff() {
$AWK '
BEGIN	{
	   # Read pagecache1
	   maxpage[1] = 1
	   pageend[1,0] = 2;
	   while ( getline < ENVIRON["pagecache1"] > 0) {
	      pageend[1,$1] = $2;
	      if ($1+0 > maxpage[1]) maxpage[1] = $1+0;
	   }

	   # Read pagecache2
	   maxpage[2] = 1
	   pageend[2,0] = 2;
	   while ( getline < ENVIRON["pagecache2"] > 0) {
	      pageend[2,$1] = $2;
	      if ($1+0 > maxpage[2]) maxpage[2] = $1+0;
	   }

	   base1 = ENVIRON["base1"]
	   base2 = ENVIRON["base2"]

	   section = "INTRODUCTION";
	   para = 0;

	}
/^\+\+/ {
	   next;
	}
/^\-\-/ {
	   next;
	}
/^ Appendix ./	{
	   section = $1 " " $2;
	   para = 0;
	}
/^  ? ? ?[0-9]+(\.[0-9]+)*\.? /	{
	   section = "Section " $1;
	   para = 0;
	}
/^ ?$/	{
	   if (inpara) {
	      printf "\n%s, paragraph %s:\n", section, para;
	      print "OLD:\n"
	      print oldpara
	      print "NEW:\n"
	      print newpara
	   }
	   oldpara = "";
	   newpara = "";
	   para ++;
	   inpara = 0
	}
/^ ./	{
	   oldpara = oldpara $0 "\n"
	   newpara = newpara $0 "\n"
	}
/^\-/	{
	   sub(/^./, " ");
	   oldpara = oldpara $0 "\n"
	   inpara++;
	}
/^\+/	{
	   sub(/^./, " ");
	   newpara = newpara $0 "\n"
	   inpara++;
	}
END	{
	   if (inpara) {
	      printf "\n%s, paragraph %s:\n", section, para;
	      print "OLD:\n"
	      print oldpara
	      print "NEW:\n"
	      print newpara
	   }	
	}
'
}


# ----------------------------------------------------------------------
# Utility to extract keyword info
# ----------------------------------------------------------------------
extract() {
    $AWK -v keyword="$1" '
	BEGIN {
	    # print "Keyword", keyword;
	}
	/^# [A-Z]/ {
	    # print "New key", $2;
	    if ($2 == keyword ":" ) { output=1; } else { output=0; }
	    # print "Output", output;
	}
	/^#\t/	{
	    # print "Content", output, $0;
	    if ( output ) {
		sub(/^#/,"");
		print;
	    }
	}
	{
	    next;
	}

    ' "$2"
}


# ----------------------------------------------------------------------
# Utility for error exit
# ----------------------------------------------------------------------
die() {
   echo $*;
   exit 1;
}

# ----------------------------------------------------------------------
# Process options
# ----------------------------------------------------------------------

# Default values
opthtml=1; optdiff=0; optchbars=0; optwdiff=0; optnowdiff=0;
optkeep=0; optinfo=0; optwidth=0;  optnums=0;  optbody=0; optabdiff=0;
optstrip=1; opthwdiff=0; optlinks=0;
optoldcolour="red"; optnewcolour="green"; optlarger=""
optstdout=0;
opttmpdiff=0; tmpdiff=$workdir/diff;

while [ $# -gt 0 ]; do
   case "$1" in
      --html)   opthtml=1; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=0;;
      --diff)   opthtml=0; optdiff=1; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=0;;
      --chbars) opthtml=0; optdiff=0; optchbars=1; optwdiff=0; opthwdiff=0; optabdiff=0;;
      --wdiff)  opthtml=0; optdiff=0; optchbars=0; optwdiff=1; opthwdiff=0; optabdiff=0;;
      --hwdiff) opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=1; optabdiff=0;;
      --changes)opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --abdiff)	opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --ab-diff)opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --rfc-editor-diff)opthtml=0; optdiff=0; optchbars=0; optwdiff=0; opthwdiff=0; optabdiff=1;;
      --version)echo -e "$basename\t$version"; exit 0;;
      --nowdiff)optnowdiff=1;;
      --keep)	optkeep=1;;
      --info)	optinfo=1; keyword=$2; shift;;
      --help)	optinfo=1; keyword="Usage";;
      --width)	optwidth=$2; shift;;
      --oldcolor)     optoldcolour=$2; shift;;
      --oldcolour)    optoldcolour=$2; shift;;
      --newcolor)     optnewcolour=$2; shift;;
      --newcolour)    optnewcolour=$2; shift;;
      --larger)       optlarger='size="+1"';;
      --linenum)optnums=1;;
      --body)	optbody=1;;
      --nostrip)optstrip=0; optbody=0;;
      --stdout) optstdout=1;;
      --links)  optlinks=1;;
      --ignore-space-change) optnospacechange=1;;
      --ignore-all-space) optignorewhite=1;;
      --ignore-blank-lines) optignoreblank=1;;
      --wdiff-args) optwdiffargs=$2; shift;;
      --tmpdiff) opttmpdiff=1; tmpdiff=$2; shift;;
      --prelines)  export prelines=$2; shift;;
      --minimal) optminimal=1;;
      --)	shift; break;;

      -v) echo "$basename $version"; exit 0;;
      -*) echo "Unrecognized option: $1";
	  exit 1;;
      *)  break;;
   esac
   shift
done

export optwidth
export optnums
export optlinks

# ----------------------------------------------------------------------
# Determine output file name. Maybe output usage and exit.
# ----------------------------------------------------------------------
#set -x

if [ $optinfo -gt 0 ]; then
   extract $keyword $0
   exit
fi
if [ $# -ge 2 ]; then
   if [ "$1" = "$2" ]; then
      echo "The files are the same file"
      exit
   fi
   export base1=$(basename "$1")
   export base2=$(basename "$2")
   outbase=$(worddiff "$base2" "$base1" "%s%s-from-%s")
else
   extract Usage $0
   exit 1
fi


# ----------------------------------------------------------------------
# create working directory.
# ----------------------------------------------------------------------
mkdir $workdir/1 || die "$0: Error: Failed to create temporary directory '$workdir/1'."
mkdir $workdir/2 || die "$0: Error: Failed to create temporary directory '$workdir/2'."

# ----------------------------------------------------------------------
# Copy files to a work directory
# ----------------------------------------------------------------------
cp "$1" $workdir/1/"$base1"
cp "$2" $workdir/2/"$base2"


# ----------------------------------------------------------------------
# Maybe strip headers/footers from both files
# ----------------------------------------------------------------------

if [ $optstrip -gt 0 ]; then
   export which=1
   strip $workdir/1/"$base1" > $workdir/1/"$base1".stripped
   mv -f $workdir/1/"$base1".stripped $workdir/1/"$base1"
   export which=2
   strip $workdir/2/"$base2" > $workdir/2/"$base2".stripped
   mv -f $workdir/2/"$base2".stripped $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Maybe do html quoting
# ----------------------------------------------------------------------

if [ $opthwdiff -gt 0 ]; then
   sed -e 's/&/&amp;/g' -e 's/</\&lt;/g' -e 's/>/\&gt;/g' $workdir/1/"$base1" > $workdir/1/"$base1".quoted
   mv -f $workdir/1/"$base1".quoted $workdir/1/"$base1"
   sed -e 's/&/&amp;/g' -e 's/</\&lt;/g' -e 's/>/\&gt;/g' $workdir/2/"$base2" > $workdir/2/"$base2".quoted
   mv -f $workdir/2/"$base2".quoted $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Maybe strip preamble (title, boilerplate and table of contents) and
# postamble (Intellectual Property Statement, Disclaimer etc)
# ----------------------------------------------------------------------
if [ $optbody -gt 0 ]; then
   bodystrip $workdir/1/"$base1" > $workdir/1/"$base1".stripped
   mv $workdir/1/"$base1".stripped $workdir/1/"$base1"
   bodystrip $workdir/2/"$base2" > $workdir/2/"$base2".stripped
   mv $workdir/2/"$base2".stripped $workdir/2/"$base2"
fi

# ----------------------------------------------------------------------
# Get output file name
# ----------------------------------------------------------------------
if [ "$3" ]; then
  outfile="$3"
else
    if [ $opthtml -gt 0 ]; then
      outfile=./"$outbase".diff.html
    fi
    if [ $optchbars -gt 0 ]; then
      outfile=./"$outbase".chbar
    fi
    if [ $optdiff -gt 0 ]; then
      outfile=./"$outbase".diff
    fi
    if [ $optabdiff -gt 0 ]; then
      outfile=./"$outbase".changes
    fi
    if [ $opthwdiff -gt 0 ]; then
      outfile=./"$outbase".wdiff.html
    fi
fi
if [ "$outfile" ]; then
   tempout=./$(basename "$outfile")
fi

# ----------------------------------------------------------------------
# Check if we can use wdiff for block diffs
# ----------------------------------------------------------------------
if [ $optnowdiff -eq 0 ]; then
   wdiffbin=$(lookfor wdiff)
   export wdiffbin
fi

# ----------------------------------------------------------------------
# Do diff
# ----------------------------------------------------------------------

origdir=$PWD
cd $workdir
if cmp 1/"$base1" 2/"$base2" >/dev/null; then
   echo ""
   echo "The files are identical."
fi

if [ $opthtml -gt 0 ]; then
   diff ${optignoreblank:+-B} ${optminimal:+-d} ${optnospacechange:+-b} ${optignorewhite:+-w} -U $prelines 1/"$base1" 2/"$base2" | tee $tmpdiff | htmldiff > "$tempout"
fi
if [ $optchbars -gt 0 ]; then
   diff -Bwd -U 10000 1/"$base1" 2/"$base2" | tee $tmpdiff | grep -v "^-" | tail -n +3 | sed 's/^+/|/' > "$tempout"
fi
if [ $optdiff -gt 0 ]; then
   diff -Bwd -U $prelines 1/"$base1" 2/"$base2" | tee $tmpdiff > "$tempout"
fi
if [ $optabdiff -gt 0 ]; then
   diff -wd -U 1000 1/"$base1" 2/"$base2" | tee $tmpdiff | abdiff
fi
if [ $optwdiff -gt 0 ]; then
   wdiff -a $optwdiffargs 1/"$base1" 2/"$base2"
fi
if [ $opthwdiff -gt 0 ]; then
    echo "<html><head><title>wdiff "$base1" "$base2"</title></head><body>"	>  "$tempout"
    echo "<pre>"								>> "$tempout"
    wdiff -w "<strike><font color='$optoldcolour' $optlarger>" -x "</font></strike>"	\
          -y "<strong><font color='$optnewcolour' $optlarger>" -z "</font></strong>"	\
	  1/"$base1" 2/"$base2"							>> "$tempout"
    echo "</pre>"								>> "$tempout"
    echo "</body></html>"							>> "$tempout"
fi

if [ $optstdout -gt 0 ]; then
  cat "$tempout"
  rm  "$tempout"
else
  cd "$origdir"; if [ -f $workdir/"$tempout" ]; then mv $workdir/"$tempout" "$outfile"; fi
fi

if [ $optkeep -eq 0 ]; then
   if [ -f $pagecache1 ]; then rm $pagecache1; fi
   if [ -f $pagecache2 ]; then rm $pagecache2; fi
   rm -fr $workdir/1
   rm -fr $workdir/2
   if [ -f $tmpdiff ]; then
      if [ $opttmpdiff -eq 0 ]; then
          rm $tmpdiff
      fi
   fi
   rm -f $workdir/$tmpdiff
   rmdir $workdir
else
   cd /tmp
   tar czf $basename-$$.tgz $basename-$$
   echo "
   Temporary workfiles have been left in $workdir/, and packed up in $workdir.tgz"
fi

---
title: "Stability Score"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Stability Score}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The function `pkg_stability()` provides a **Stability Score** for a selected 
package.  This function looks at the entire history of package releases,
and compares them to identify breaking changes.  It then uses the number
of breaking changes to calculate a **Stability Score**.  This vignette 
describes how the score is calculated.

### Breaking Changes

Before describing how the **Stability Score** is calculated, we first have to 
understand what is meant by a "breaking change".  In the context of **pkgdiff**,
a "breaking change" is a function that exists in one release, but is removed in the next.
It can also be a function parameter that exists in one release, but is removed
in the next.  There are, therefore, two kinds of breaking changes tracked
by **pkgdiff**: removed functions and removed parameters.  Either one of these
will count as a breaking change.

Note that there are other kinds of changes not tracked by **pkgdiff**.  Functions
that are changed, but not removed, are not tracked.  Also, functions that 
are deprecated, but not removed, are not tracked.  That means you can start 
getting a deprecation warning on a function, but as long as it is not actually
removed, **pkgdiff** does not count that as a breaking change.

### Percentage of Non-Breaking Changes

A package version with at least one breaking change from the previous version is called 
a "breaking release". A package version that is compatible with the previous 
version is called a "non-breaking release".  

The **Stability Score** is calculated as the percentage of non-breaking releases.
In other words, the **Stability Score** is the number of breaking releases 
divided by the total number of releases, subtracted from one.  The basic 
formula for the stability score is thus: 1 - (BR / TR), where 
BR is Breaking Releases and TR is Total Releases.

For example, if a package has 10 releases, and one of them is a breaking 
release, the **Stability Score** will be .9 or 90%.

### Weighting Over Time

Breaking releases are weighted over time.  This consideration reflects the 
perception that breaking releases many years ago matter less than recent breaking
releases.  Breaking releases therefore start with a value of 
one (1), and that value is reduced by .01 for each month away from the present
date.  That means a breaking release in the last month will count as 1, and
a breaking release two years ago will count as 1 - (.01 * 24) = 0.76.

Observe that this weighting over time will eventually cause the breaking release
to not be counted at all.  The discounting of a breaking release occurs at 8.333
years.  That means breaking releases older than 8.333 years will not be considered
in the **Stability Score**.

### Other Considerations

Also note that the number of releases eventually age out of the calculation. 
Any release older
than 10 years from the current date will be discarded from the Stability Score.
These releases will still be shown in the Stability Data. Keeping them in 
the data helps to give you an accurate understanding of the package history.

## Stability Assessment

In addition to the above Stability Score, the `pkg_stability()` function also provides
a **Stability Assessment**.  The **Stability Assessment** is a categorization of
the Stability Score.  The Stability Assessment is a way to give you an
English interpretation of the Stability Score.  The Stability Assessment is 
determined as follows:

  - 100%: Perfect
  - &gt;= 95%: Very Stable
  - &gt;= 90%: Stable
  - &gt;= 80%: Somewhat Unstable
  - &lt; 80%: Unstable
  
The above categorization sets a high bar. The purpose of setting a high bar is 
simply to bring the 
R Open Source language up to the stability level of proprietary 
languages.  It was designed to promote stability and encourage package
developers to reduce the number of breaking changes in their Open Source packages.
A high stability score is something that R package developers should be proud of, 
and R users should look for when selecting packages to include in their programs.

## CRAN Packages Only

The **Stability Score** and **Stability Assessment** only work with packages 
published on CRAN.
The reason is because CRAN preserves all releases of a package, and those archived
releases are easily available to **pkgdiff**.  Releases on Github are not reliable,
as they are subject to whether the package developer remembers to create 
a release version or not.  

---
title: "Frequently Asked Questions"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{FAQ}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

options(rmarkdown.html_vignette.check_title = FALSE)

```
Below are some frequently asked questions about the **pkgdiff** package. Click
on the links below to navigate to the full question and answer content.

## Index{#top}

* [What is **pkgdiff**?](#overview)
* [Do I need internet access to use **pkgdiff**?](#internet)
* [How are you doing the version comparisons?](#comparisons)
* [Do package comparisons have to be between sequential versions?](#sequential)
* [What counts as a breaking change?](#breaks)
* [How is the stability score calculated?](#stability)
* [Why is pkg_stability() so slow?](#performance)
* [Which packages are in the cache?](#cache)
* [Where is the cache located?](#location)
* [How can I add a package to the cache?](#addcache)
* [Can I create a report with **pkgdiff**?](#report)
* [Can I use **pkgdiff** with packages that are not on CRAN?](#github)
* [Can I use **pkgdiff** with Base R packages?](#base)

## Content

### What is **pkgdiff**? {#overview}

**Q:** I don't quite understand this package.  What it is for?  Why 
would anyone care about differences between two versions of a package?

**A:** The **pkgdiff** package arose as an attempt to deal with the 
problem of breaking changes.  A breaking change occurs when 
a function or function parameter exists in one version of a package,
but is then removed in a subsequent version.  If you happen to be using
that function in a program, the removal breaks your code.  

Breaking changes 
are especially problematic when you have large numbers
of programs that use many R packages. In this situation, upgrading
your R repository can cause a cascade of breaking changes that can
cripple your entire system.  

One aim of **pkgdiff** is to help identify stable packages.
The most sure way to avoid breaking changes is to avoid using unstable packages
in the first place.  The functions `pkg_info()`, `pkg_stability()`, and
`repo_stability()` can
help you assess the stability of your packages.

The second aim of **pkgdiff** is to anticipate 
breaking changes for the packages already in your repository. The functions 
`pkg_diff()` and `repo_breakages()` let you know if a package is going to break
when it is upgraded, and specifically which functions or parameters will
be removed.  This information will allow you determine what the impact 
of a version upgrade will be to your system.  Knowing the potential impact
will allow you to plan your upgrade more effectively.

[top](#top) 

******

### Do I need internet access to use **pkgdiff**? {#internet}

**Q:** At my company the R server we use is offline, and not connected
to the internet.  Can I use **pkgdiff** on an offline system? Do I have
to have an internet connection?

**A:** Yes. **pkgdiff** requires an internet connection.  It uses the
connection to pull package information from CRAN, the RStudio CRAN
mirror, and Github. In this scenario, you should use **pkgdiff** 
on a connected system to run queries about the packages on your 
disconnected system.

[top](#top) 

******

### How are you doing the version comparisons? {#comparisons}

**Q:** When you compare package versions, what exactly are you comparing?
Is this a full code comparison?  What about the documentation, etc.?

**A:** Package comparisons are focused on two things: exported functions and 
exported function parameters.  The comparison does not look at anything else.
It does not look at code or documentation. The goal of the comparison 
is to come up with a list of functions and parameters that have been added or 
removed between releases.  These items can directly affect any programs
that rely on that package.

Note that sometimes code changes can cause a breaking change, even when
the function itself is not removed or the parameters altered.  **pkgdiff** currently
cannot detect this type of change.  This feature may be included
in a future release.

[top](#top) 

******

### Do package comparisons have to be between sequential versions? {#sequential}

**Q:** I've been using the same version of a package for over a year, and 
now I want to upgrade.  There have been three releases in the last year.  Can 
I still get differences for releases that are not sequential?

**A:** Yes.  The `pkg_diff()` function will compare any two versions of a package.
The versions do not have to be sequential.

[top](#top) 

******

### What counts as a breaking change? {#breaks}

**Q:** What do you mean by a breaking change?  What exactly counts as a 
breaking change?

**A:** There are two types of breaking change tracked by **pkgdiff**: a)
a removed function, and b) a removed function parameter.  Either one
counts as a breaking change.  If the function code changes but the
signature stays the same, it does not 
count as a breaking change. Likewise, if a function parameter is added 
to the end of the signature, and no parameters are removed, it does not 
count as a breaking change.

Also note that a "breaking release" is any release with at least one
breaking change. Whether a release has one function removed or 10 functions
removed, it still counts as one breaking release.

[top](#top) 

******

### How is the stability score calculated? {#stability}

**Q:** It's not clear how you came up with the stability score. What goes into
this calculation?

**A:** The stability score is the percentage of non-breaking releases.  **pkgdiff**
looks at the entire history of a package, counts how many releases there
are, and determines how many of them were breaking releases. 
The stability score is the number of breaking releases divided by the 
total number of releases.  Some adjustments are made for the age of the 
breaking release.  See `vignette("pkgdiff-stability")` for a complete 
description of how the stability score is calculated. 

[top](#top) 

******

### Why is `pkg_stability()` so slow? {#performance}

**Q:** For most packages I'm looking at, `pkg_stability()` runs pretty fast.
For a few packages, it is very slow, and appears to be doing a lot of additional
processing.  What is going on?

**A:** The difference is that some packages have been cached, and some have not.
About 20% of the top CRAN downloads have been cached.  This subset constitutes
the most commonly used R packages, with over 1000 downloads per month.  For 
packages with less than 1000 downloads per month, the **pkgdiff** functions
will still run. But the package information has to be downloaded and extracted
from the CRAN mirror. The additional processing slows down `pkg_stability()`
significantly.  

For more information on the package cache, 
see `vignette("pkgdiff-cache")`.

[top](#top) 

******

### Which packages are in the cache? {#cache}

**Q:** I understand that some packages have been cached, and some have not.
How do I know which packages have been cached?  Is there a list someplace?

**A:** You can get a list of cached packages by running the function
`pkg_cache()`.  If you have a specific package in mind, and want to know 
if that package is in the cache, you can pass the package name to 
`pkg_cache()` as the first parameter.  If the package exists in the cache, 
the version number will be shown in the "Version" column.  If the package
does not exist in the cache, the version will be NA.

[top](#top) 

******

### Where is the cache located? {#location}

**Q:** Reading the documentation for **pkgdiff**, it appears that some
package information has been cached to speed up the queries.  Where
exactly is this cache located?

**A:** The **pkgdiff** package cache is located on Github. The 
URL is "https://github.com/dbosak01/pkgdiffdata". The package cache is 
usually updated on a daily basis.  

[top](#top) 

******

### How can I add a package to the cache? {#addcache}

**Q:** I have a package that I want to run a stability score on,
but that package is not in the cache, and runs too slow.  Is there
any way to add the package to the cache?

**A:** Yes. You may request that the package be added to the cache.
Use the **pkgdiff** issue list located 
[here](https://github.com/dbosak01/pkgdiff/issues).

[top](#top) 

******

### Can I create a report with pkgdiff? {#report}

**Q:** I like the console output from **pkgdiff**. But I want to create
a permanent report.  Is there a way to get a permanent report that I can 
save and print?

**A:** Yes. Use **Quarto** or **RMarkdown**.  You can call **pkgdiff**
functions from within a markdown document.  Then when you knit the document,
the console output will be emitted as text. The document can then be saved and 
printed as desired.

[top](#top) 

******

### Can I use **pkgdiff** with packages that are not on CRAN? {#github}

**Q:** My company uses some packages on Github, and also has a couple 
packages that have been developed internally.  Can I use **pkgdiff**
for these packages?

**A:** No.  **pkgdiff** only works with packages published on CRAN. The reason
is that CRAN retains all releases of a package in an archive. **pkgdiff**
uses this archive to retrieve information about a specific version of a 
package, and calculate stability scores.  Packages on Github or 
locally developed packages do not have a predictable and easily accessible archive.

Note that there is another package on CRAN called **packageDiff** that allows you
to compare two packages locally. This package can generate version differences.
It will not, however, generate stability scores, or assess an entire repository.

[top](#top) 

******

### Can I use **pkgdiff** with Base R packages? {#base}

**Q:** I see that **pkgdiff** does not seem to work with some very commonly 
used packages.
For instance, `pkg_stabilty()` produces no results for "tools" and "utils".  Does
**pkgdiff** work with Base R packages?

**A:** No.  **pkgdiff** only works with contributed packages. It does not 
work with Base R packages, as these packages are distributed and archived in 
a completely different way.  

Some Base R "recommended" packages are published
as contributed CRAN packages, and will therefore work with **pkgdiff**.  But 
packages such as "base", "tools", "utils", "compiler", "datasets", "graphics", 
"grDevices", "grid", "methods", "stats", etc. will give limited or no results
for most functions in **pkgdiff**.  

[top](#top) 

******



<!-- ### How do I format a data frame? {#overview} -->

<!-- **Q:** I have a data frame with different types of variables: -->
<!-- numbers, dates, character values.  How can I format them without messing -->
<!-- up the original data? -->


<!-- **A:** With the **fmtr** package you can assign the formats to the  -->
<!-- "format" attribute of the dataframe, and the apply the formats using the  -->
<!-- `fdata()` function, returning the result to a new data frame.  Here is  -->
<!-- a simple example: -->
<!-- ```{r eval=FALSE, echo=TRUE} -->
<!-- # Create sample data -->
<!-- dat <- data.frame(SUBJ = c(1, 2, 3), -->
<!--                   BDATE = c(as.Date("1945-10-17"),  -->
<!--                             as.Date("1967-09-04"), -->
<!--                             as.Date("1998-04-28")), -->
<!--                   SEX = c("M", "F", "M"), -->
<!--                   WEIGHT = c(77.1107, 64.2848, 85.9842)) -->

<!-- # View data                   -->
<!-- dat -->
<!-- #   SUBJ      BDATE SEX  WEIGHT -->
<!-- # 1    1 1945-10-17   M 77.1107 -->
<!-- # 2    2 1967-09-04   F 64.2848 -->
<!-- # 3    3 1998-04-28   M 85.9842 -->

<!-- # Assign formats -->
<!-- formats(dat) <- list(BDATE = "%Y/%m/%d", -->
<!--                      SEX = c("M" = "Male", "F" = "Female"), -->
<!--                      WEIGHT = "%1.1f kg") -->

<!-- # Apply formats to new data frame -->
<!-- dat2 <- fdata(dat) -->

<!-- # View new data frame -->
<!-- dat2 -->
<!-- #   SUBJ      BDATE    SEX  WEIGHT -->
<!-- # 1    1 1945/10/17   Male 77.1 kg -->
<!-- # 2    2 1967/09/04 Female 64.3 kg -->
<!-- # 3    3 1998/04/28   Male 86.0 kg -->
<!-- ``` -->
<!-- [top](#top)  -->

<!-- ****** -->

---
title: "Package Cache"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Package Cache}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

To improve performance of difference operations, and especially 
of stability scores, **pkgdiff** uses a cache of pre-processed package
information on Github.  This cache greatly speeds up the operation
of **pkgdiff** functions.  The cache is created by a periodic process
that extracts information from packages on CRAN, and stores it in a way
that the information can be quickly retrieved when a **pkgdiff** function
is called.

Note the following about the **pkgdiff** cache:  

1) To access the cache, the computer **pkgdiff** is running on must have 
access to the internet and, specifically, Github.  If your machine is 
isolated from the internet, or blocked from Github, **pkgdiff** will error.

2) The cache does not contain all CRAN packages.  It contains packages that
exceed 1000 downloads per month.  These packages constitute approximately 20%
of the top CRAN downloads, or about 4000 packages. For most R users, any package 
you might be interested in will be included in this subset.

3) To find out which packages are in the **pkgdiff** cache, 
you may call the function `pkg_cache()`. See the documentation of that
function for additional information.

4) If a package is not in the cache, all **pkgdiff** functions will still work. 
However, they will be much slower than packages
that are included in the cache. 

5) You may request that a package be added to the cache by submitting 
an issue to the **pkgdiff** issue log 
[here](https://github.com/dbosak01/pkgdiff/issues).

6) Observe that the `pkg_info()` function also returns a parameter
on whether or not the selected package is in the cache.  It is recommended
to run `pkg_info()` or `pkg_cache()` first, and check if the package is 
in the cache before running `pkg_stability()`.

7) If a package is not in the cache, you may speed up `pkg_stability()` by 
setting either the "releases" or "months" parameters.  Either of these parameters
will limit the scope of the call to `pkg_stability()`.  For example, if a package has
a lifespan of 10 years, you can greatly reduce the amount of stability processing
by reducing the comparison period to 36 months instead of the entire 10 years.
Do note, however, that such scope reductions may also have an effect on 
the calculated score and assessment.

8) The package cache is updated periodically, usually every day.  The update 
looks for new versions of packages that are already in the cache.  If a new
version is found, the update routine will add the new version to the stored
infos for that package, and also update the stability data.  The update
routine also looks for new packages that have recently met the 1000 downloads per month
threshold.  If such a package is found, it will be added to the cache automatically.
Packages that fall below the 1000 downloads per month threshold are not removed.
Once a package has been added to the cache, it will be retained indefinitely.

---
title: "pkgdiff"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{pkgdiff}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(pkgdiff)
```
The purpose of **pkgdiff** is to identify changes between package versions,
with the aim of assessing package stability.  Here is a list of the major
functions in **pkgdiff**, and a short explanation of each:

- `pkg_info()`: Displays general information about a package,
including the description, maintainer, and number of downloads in
the last month.
- `pkg_versions()`: Returns a data frame of all versions of a package.
- `pkg_diff()`: Compares two versions of a package, and shows
which functions and parameters have been added or removed.
- `pkg_stability()`: Looks at the history of a package,
identifies breaking changes, and assesses the package for stability.
- `pkg_summary()`: Creates an HTML summary of information about a package,
and sends it to the RStudio viewer. 
- `pkg_repo()`: Generates a table of the packages in a repository
and their versions.
- `repo_breakages()`: Looks at a package repository and tries to identify
which packages will break when upgraded.
- `repo_stability()`: Looks at a package repository and calculates
a stability score for each package.

Below is a further elaboration and demo for each of the above functions.

## General Package Information

The first thing you may want to do when exploring a package is to get
some general information about it.  General information about a package
can be achieved using the `pkg_info()` function.  All you need to 
pass is the package name. The function will then return basic 
information such as the maintainer, description,
and imported packages.  Here is a simple demonstration:

```{r eval=FALSE, echo=TRUE}

# Examine tidymodels package
pkg_info("tidymodels")
# # A package info object: tidymodels package
# - Version: v1.2.0
# - Release Date: 2024-03-25
# - Title: Easily Install and Load the 'Tidymodels' Packages
# - Maintainer: Max Kuhn <max@posit.co>
# - License: MIT + file LICENSE
# - Description: The tidy modeling "verse" is a collection of packages for
# modeling and statistical analysis that share the underlying design
# philosophy, grammar, and data structures of the tidyverse.
# - Depends: R (>= 3.6)
# - Imports: broom (>= 1.0.5), cli (>= 3.6.2), conflicted (>= 1.2.0), 
# dials (>= 1.2.1), dplyr (>= 1.1.4), ggplot2 (>= 3.5.0), hardhat (>= 1.3.1), 
# infer (>= 1.0.6), modeldata (>= 1.3.0), parsnip (>= 1.2.1), purrr (>= 1.0.2), 
# recipes (>= 1.0.10), rlang (>= 1.1.3), rsample (>= 1.2.1), rstudioapi (>= 0.16.0), 
# tibble (>= 3.2.1), tidyr (>= 1.3.1), tune (>= 1.2.0), workflows (>= 1.1.4), 
# workflowsets (>= 1.1.0), yardstick (>= 1.3.1)
# - Suggests: covr, glue, knitr, rmarkdown, testthat (>= 3.0.0), xml2
# - Repository: CRAN
# - Cached: TRUE
# - Functions: 9

```
The return value of this function is actually an object that can be manipulated
programmatically.  Here we will extract some of the data values from this object:

```{r eval=FALSE, echo=TRUE}
# Capture package info object
res <- pkg_info("tidymodels")

# View version
res$Version
# [1] "1.3.0"

# View title
res$Title
# [1] "Easily Install and Load the 'Tidymodels' Packages"

# View release date
res$ReleaseDate
# [1] "2025-02-21"

```
Note that, by default, the version of the package returned is the version of the package
located on the current machine, in the current version of the R repository.
If the package is not installed in the current R repository, the function
will return the latest version on CRAN.

Also note that the object provides a complete list of all functions exported 
by the package. The function list can be accessed programmatically, and also viewed using
the `verbose` parameter of the `print()` function:

```{r eval=FALSE, echo=TRUE}
# Get package info
res <- pkg_info("tidymodels")

# View functions
print(res, verbose = TRUE)
# # A package info object: tidymodels package
# - Version: v1.3.0
# - Release Date: 2025-02-21
# - Title: Easily Install and Load the 'Tidymodels' Packages
# - Maintainer: Max Kuhn <max@posit.co>
# - License: MIT + file LICENSE
# - Description: The tidy modeling "verse" is a collection of packages for
# modeling and statistical analysis that share the underlying design
# philosophy, grammar, and data structures of the tidyverse.
# - Depends: R (>= 3.6)
# - Imports: broom (>= 1.0.7), cli (>= 3.6.4), conflicted (>= 1.2.0), 
# dials (>= 1.4.0), dplyr (>= 1.1.4), ggplot2 (>= 3.5.1), hardhat (>= 1.4.1), 
# infer (>= 1.0.7), modeldata (>= 1.4.0), parsnip (>= 1.3.0), purrr (>= 1.0.4), 
# recipes (>= 1.1.1), rlang (>= 1.1.5), rsample (>= 1.2.1), rstudioapi (>= 0.17.1), 
# tibble (>= 3.2.1), tidyr (>= 1.3.1), tune (>= 1.3.0), workflows (>= 1.2.0), 
# workflowsets (>= 1.1.0), yardstick (>= 1.3.2)
# - Suggests: covr, glue, knitr, rmarkdown, testthat (>= 3.0.0), xml2
# - Downloads/Month: 31789
# - Repository: CRAN
# - Cached: TRUE
# - Functions: 
#   - pkg_deps(): x, recursive
#   - print.tidymodels_conflicts(): x, ..., startup
#   - tag_attach(): tag
#   - tag_show()
#   - tag_update(): tag
#   - tidymodels_conflicts()
#   - tidymodels_packages(): include_self
#   - tidymodels_prefer(): quiet
#   - tidymodels_update(): pkg, recursive, ...
```

## Stability Assessment

Besides general information about a package, you may also view the "stability"
of a package.  The `pkg_stability()` function looks at the entire history
of the package, and attempts to assess its overall stability in terms of 
breaking changes.  Here, a "breaking change" means either a removed function
or removed function parameter.  The `pkg_stability()` function is very useful when you are
writing production code, and want to select the most stable packages possible.
The results of this function can help you make that determination: 

```{r eval=FALSE, echo=TRUE}

# View package stability
pkg_stability("tidymodels")
# # A stability score: tidymodels package
# - Age: 6.58 years
# - Score: 100.0
# - Assessment: Perfect
# - Version Range: 0.0.1/1.3.0
# - Release Range: 2018-07-27/2025-02-21
# - Release Count: 14
# - Breaking Releases: 0
# - Data: 
#       Package Version                FileName    Release  Size AF AP RF RP BC TF
# 1  tidymodels   1.3.0 tidymodels_1.3.0.tar.gz 2025-02-21 46.4K  0  0  0  0  0  9
# 2  tidymodels   1.2.0 tidymodels_1.2.0.tar.gz 2024-03-25 49.6K  0  0  0  0  0  9
# 3  tidymodels   1.1.1 tidymodels_1.1.1.tar.gz 2023-08-24   50K  0  0  0  0  0  9
# 4  tidymodels   1.1.0 tidymodels_1.1.0.tar.gz 2023-05-01   49K  0  0  0  0  0  9
# 5  tidymodels   1.0.0 tidymodels_1.0.0.tar.gz 2022-07-13   75K  0  0  0  0  0  9
# 6  tidymodels   0.2.0 tidymodels_0.2.0.tar.gz 2022-03-19   75K  0  0  0  0  0  9
# 7  tidymodels   0.1.4 tidymodels_0.1.4.tar.gz 2021-10-01   17K  0  0  0  0  0  9
# 8  tidymodels   0.1.3 tidymodels_0.1.3.tar.gz 2021-04-19   15K  1  0  0  0  0  9
# 9  tidymodels   0.1.2 tidymodels_0.1.2.tar.gz 2020-11-22   26K  0  0  0  0  0  8
# 10 tidymodels   0.1.1 tidymodels_0.1.1.tar.gz 2020-07-14   25K  0  0  0  0  0  8
# 11 tidymodels   0.1.0 tidymodels_0.1.0.tar.gz 2020-02-16   57K  0  0  0  0  0  8
# 12 tidymodels   0.0.3 tidymodels_0.0.3.tar.gz 2019-10-05   19K  0  0  0  0  0  8
# 13 tidymodels   0.0.2 tidymodels_0.0.2.tar.gz 2018-11-27   18K  0  0  0  0  0  8
# 14 tidymodels   0.0.1 tidymodels_0.0.1.tar.gz 2018-07-27   18K  8  8  0  0  0  8

```

The stability score is calculated generally as the percentage of non-breaking
releases.  Breaking releases are also weighted to reduce their impact over time.
The above score of 100 for **tidymodels** means that the package has never
had a breaking change.  It is therefore given a stability assessment of "Perfect".

Now let's look at a package with some breaking changes:
```{r eval=FALSE, echo=TRUE}

# View package stability
pkg_stability("dbplyr")
# # A stability score: dbplyr package
# - Age: 7.67 years
# - Score: 73.7
# - Assessment: Unstable
# - Version Range: 1.0.0/2.5.0
# - Release Range: 2017-06-09/2024-03-19
# - Release Count: 23
# - Breaking Releases: 10
# - Data: 
#    Package Version            FileName    Release   Size  AF  AP RF RP BC  TF
# 1   dbplyr   2.5.0 dbplyr_2.5.0.tar.gz 2024-03-19 752.6K  32   6 12  0  1 611
# 2   dbplyr   2.4.0 dbplyr_2.4.0.tar.gz 2023-10-26   743K  74  80 45 33  1 591
# 3   dbplyr   2.3.4 dbplyr_2.3.4.tar.gz 2023-09-27   706K   0   0  0  0  0 562
# 4   dbplyr   2.3.3 dbplyr_2.3.3.tar.gz 2023-07-07   706K   0   0  0  0  0 562
# 5   dbplyr   2.3.2 dbplyr_2.3.2.tar.gz 2023-03-21   707K   0   1  0  1  1 562
# 6   dbplyr   2.3.1 dbplyr_2.3.1.tar.gz 2023-02-24   707K   5   6  1  0  1 562
# 7   dbplyr   2.3.0 dbplyr_2.3.0.tar.gz 2023-01-16   688K  33  16  5  1  1 558
# 8   dbplyr   2.2.1 dbplyr_2.2.1.tar.gz 2022-06-27   655K   4   0  0  0  0 530
# 9   dbplyr   2.2.0 dbplyr_2.2.0.tar.gz 2022-06-05   655K 136  47 53  1  1 526
# 10  dbplyr   2.1.1 dbplyr_2.1.1.tar.gz 2021-04-06   568K   5   0  0  0  0 443
# 11  dbplyr   2.1.0 dbplyr_2.1.0.tar.gz 2021-02-03   566K  15   2  0  0  0 438
# 12  dbplyr   2.0.0 dbplyr_2.0.0.tar.gz 2020-11-03   522K 139  33 94 11  1 423
# 13  dbplyr   1.4.4 dbplyr_1.4.4.tar.gz 2020-05-27   417K   5   0  0  0  0 378
# 14  dbplyr   1.4.3 dbplyr_1.4.3.tar.gz 2020-04-19   413K   9   3  2  2  1 373
# 15  dbplyr   1.4.2 dbplyr_1.4.2.tar.gz 2019-06-17   296K   1   0  0  0  0 366
# 16  dbplyr   1.4.1 dbplyr_1.4.1.tar.gz 2019-06-06   296K   1   0  0  0  0 365
# 17  dbplyr   1.4.0 dbplyr_1.4.0.tar.gz 2019-04-24   295K  24  30 17 13  1 364
# 18  dbplyr   1.3.0 dbplyr_1.3.0.tar.gz 2019-01-09   263K   2   0  0  0  0 357
# 19  dbplyr   1.2.2 dbplyr_1.2.2.tar.gz 2018-07-25   258K   0   0  0  0  0 355
# 20  dbplyr   1.2.1 dbplyr_1.2.1.tar.gz 2018-02-19   257K   0   0  0  0  0 355
# 21  dbplyr   1.2.0 dbplyr_1.2.0.tar.gz 2018-01-03   257K  43   6  2  2  1 355
# 22  dbplyr   1.1.0 dbplyr_1.1.0.tar.gz 2017-06-27   241K  23   2  0  0  0 314
# 23  dbplyr   1.0.0 dbplyr_1.0.0.tar.gz 2017-06-09   235K 291 291  0  0  0 291

```
As seen above, the **dbplyr** package has had a lot of breaking changes.  Which releases
had breaking changes are shown in the "BC" column of the data grid.  The "AF" and
"AP" columns stand for "Added Functions" and "Added Parameters", while
"RF" and "RP" mean "Removed Functions" and "Removed Parameters".  "TF" is the total
number of exported functions in that version of the package.

Since this package has issued breaking changes in 10 out of 23 releases, it gets
a stability score of 73.7, and an assessment of "Unstable".  A stability assessment
of "Unstable" does not necessarily mean you should not use the package.  Whether
or not you use the package depends on your situation, and the amount of 
risk you can tolerate.


## Create a Summary Report

The information from the above functions can also be combined into a nice report
using `pkg_summary()`. `pkg_summary()` pulls key data from `pkg_info()` and 
`pkg_stability()` and generates an HTML page, which it then sends to the 
RStudio viewer.  

For example, you can run a package summary report for the **dplyr** package like
this:

```{r eval=FALSE, echo=TRUE}
# Get package summary report
pkg_summary("dplyr")
```
The following will then appear in the RStudio viewer:

<img src="../man/images/summary2.png" alt = "Summary Report 2"/>

The chart shows the number of breaking changes per release.  As you can see,
the **dplyr** package had a large number of breaking changes between version
"0.8.5" and version "1.0.0", but has been relatively stable since then.  

The Package Summary Report provides an easy way to get an overview of a package.
For additional information, see the `pkg_summary()` documentation.

## Comparing Package Versions

If you find a package with breaking changes, you would then want to know
what those breaking changes are.  The breaking changes between package versions
can be discovered with the `pkg_diff()` function.  This function accepts 
a package name, and the two versions of the package you wish to compare.

Looking at our stability data above, let's examine the breaking changes 
between versions "2.4.0" and "2.5.0" of **dbplyr**:

```{r eval=FALSE, echo=TRUE}
# Compare package versions
pkg_diff("dbplyr", "2.4.0", "2.5.0")
# # A difference object: dbplyr package
# - Comparing: v2.4.0/v2.5.0
# - Breaking Changes: TRUE
# - Added Functions: 
#   - [.dbplyr_table_path()
#   - [[.dbplyr_table_path()
#   - as_table_path()
#   - c.dbplyr_table_path()
#   - check_table_path()
#   - complete.tbl_lazy()
#   - db_explain.Oracle()
#   - db_explain.OraConnection()
#   - db_sql_render.Microsoft SQL Server()
#   - escape.dbplyr_table_path()
#   - escape.default()
#   - expand.tbl_lazy()
#   - fill.tbl_lazy()
#   - group_by_drop_default.tbl_lazy()
#   - intersect.tbl_lazy()
#   - is_table_path()
#   - pivot_longer.tbl_lazy()
#   - pivot_wider.tbl_lazy()
#   - print.dbplyr_table_path()
#   - reframe.tbl_lazy()
#   - replace_na.tbl_lazy()
#   - setdiff.OraConnection()
#   - setdiff.tbl_lazy()
#   - setdiff.tbl_Oracle()
#   - simulate_mariadb()
#   - sql_escape_ident.default()
#   - sql_escape_ident.MySQLConnection()
#   - sql_escape_string.default()
#   - table_path_components()
#   - table_path_components.default()
#   - table_path_name()
#   - union.tbl_lazy()
# - Added Parameters: 
#   - flatten_query.base_query(): con
#   - flatten_query.multi_join_query(): con
#   - flatten_query.select_query(): con
#   - flatten_query.union_query(): con
#   - flatten_query.values_query(): con
#   - sql_translation.MySQL(): con
# - Removed Functions: 
#   - as_table_ident.character()
#   - as_table_ident.dbplyr_catalog()
#   - as_table_ident.dbplyr_schema()
#   - as_table_ident.dbplyr_table_ident()
#   - as_table_ident.Id()
#   - as_table_ident.ident()
#   - as_table_ident.ident_q()
#   - as_table_ident.sql()
#   - escape.dbplyr_table_ident()
#   - format.dbplyr_table_ident()
#   - sql_escape_ident.DBIConnection()
#   - sql_escape_string.DBIConnection()

```
As you can see, many new functions and function parameters have been added
between these two versions.  In addition, there are 12 functions that 
have been removed.  It is up to you to determine whether and how these 
removals will affect your programs.  It is very convenient, however,
to have an accurate accounting of package changes.  These changes
can help you plan your package upgrades and anticipate breakages.

## Listing Package Versions

Another question you may ask when evaluating a package is: How many versions of
a package are there?  **pkgdiff** provides two tools to find version 
information: `pkg_versions()` and `pkg_repo()`.  

`pkg_versions()` returns all the versions of a package in a data frame, along
with release dates for each version. It works like this:

```{r eval=FALSE, echo=TRUE}
# Get package versions
pkg_versions("stringr")
#    Package Version              FileName    Release   Size
# 1  stringr   1.5.1  stringr_1.5.1.tar.gz 2023-11-14   172K
# 2  stringr   1.5.0  stringr_1.5.0.tar.gz 2022-12-02   172K
# 3  stringr   1.4.1  stringr_1.4.1.tar.gz 2022-08-21   133K
# 4  stringr   1.4.0  stringr_1.4.0.tar.gz 2019-02-10   133K
# 5  stringr   1.3.1  stringr_1.3.1.tar.gz 2018-05-10   116K
# 6  stringr   1.3.0  stringr_1.3.0.tar.gz 2018-02-19   116K
# 7  stringr   1.2.0  stringr_1.2.0.tar.gz 2017-02-18    92K
# 8  stringr   1.1.0  stringr_1.1.0.tar.gz 2016-08-19    62K
# 9  stringr   1.0.0  stringr_1.0.0.tar.gz 2015-04-30    34K
# 10 stringr   0.6.2  stringr_0.6.2.tar.gz 2012-12-06    20K
# 11 stringr   0.6.1  stringr_0.6.1.tar.gz 2012-07-25    20K
# 12 stringr     0.6    stringr_0.6.tar.gz 2011-12-08    20K
# 13 stringr     0.5    stringr_0.5.tar.gz 2011-06-30    18K
# 14 stringr     0.4    stringr_0.4.tar.gz 2010-08-24    16K
# 15 stringr     0.3    stringr_0.3.tar.gz 2010-02-15    11K
# 16 stringr     0.2    stringr_0.2.tar.gz 2009-11-16    10K
# 17 stringr  0.1.10 stringr_0.1.10.tar.gz 2009-11-09   6.8K

```
Observe that the **stringr** package has had 17 releases. The first 
release was in 2009, and the most recent in 2023.  This information is 
scraped from the CRAN archive site.  It is useful for understanding
the release history of a package, and knowing which versions
of a package are available.

Next, you may want to know which versions of packages you have installed
in your R repository. This information can be queried using the 
`pkg_repo()` function.  Let's take a look:


```{r eval=FALSE, echo=TRUE}

# Check version in repo
pkg_repo("stringr")
# # A package repo object
# - Repo Version: 4.4
# - Packages:
#   Package Version
# 1 stringr   1.5.1

``` 
Base on the above information, it appears we have the latest version of 
**stringr** installed in our R 4.4 repo.  Good to know!

Can `pkg_repo()` be used to get versions for multiple packages? Yes! Here
is how:

```{r eval=FALSE, echo=TRUE}

# Create vector of packages
v1 <- c("curl", "dplyr", "fmtr", "purrr")

# Get current versions
res1 <- pkg_repo(v1, ver = "current")

# View results
res1
# # A package repo object
# - Repo Version: 4.4
# - Packages:
#   Package Version
# 1    curl   5.2.1
# 2   dplyr   1.1.4
# 3    fmtr   1.6.5
# 4   purrr   1.0.2

# Get latest versions
res2 <- pkg_repo(v1, ver = "latest")

# View results
res2
# # A package repo object
# - Repo Version: latest
# - Packages:
#   Package Version
# 1    curl   6.2.1
# 2   dplyr   1.1.4
# 3    fmtr   1.6.8
# 4   purrr   1.0.4

```
Moreover, if you run `pkg_repo()` without any parameters, it will return
all the R packages you have installed, along with their versions.  Nice!

The `pkg_repo()` function is very useful when dealing with large numbers of 
packages.  This function feeds directly into the next two topics, which discuss
operations on entire repositories.

## Anticipating Breakages

One of the main motivations for creating **pkgdiff** was to find a way to
better handle repository upgrades.  As it stands, repository upgrades can be a 
chaotic experience. When you have several hundred packages in your repository,
you will surely have some of them break when you go to upgrade.  Wouldn't it
be nice to at least know which packages are going to break?  That is the topic
of this next section.

### Small Set of Packages

First, let's see how we can anticipate breaking changes on a small number of 
packages.  

For this exercise, we will use the `res1` and `res2` 
data frames from the previous section.  These data frames were created using
`pkg_repo()`, and contain version numbers for four R packages.  These data
frames are in fact perfect inputs to our next function: `repo_breakages()`.

The `repo_breakages()` function can compare packages in bulk.  It will 
perform difference operations on any number of packages, and tell you whether
or not there will be any breaking changes.  As you can imagine, this function
is well suited for repository upgrades.  Here is an example:

```{r eval=FALSE, echo=TRUE}
# View upgrade breakages
repo_breakages(res1, res2) 
# Comparing curl v5.2.1/v6.2.0
# Comparing fmtr v1.6.5/v1.6.8
# Comparing purrr v1.0.2/v1.0.4
# # A repo breakages object
# - Run Datetime: 2025-03-01 17:58 UTC
# - Summary:
#   Package Version1 Version2 Breakages
# 1    curl    5.2.1    6.2.1      TRUE
# 2   dplyr    1.1.4    1.1.4     FALSE
# 3    fmtr    1.6.5    1.6.8     FALSE
# 4   purrr    1.0.2    1.0.4     FALSE
# - Details:
# # A difference object: curl package
# - Comparing: v5.2.1/v6.2.1
# - Breaking Changes: TRUE
# - Added Functions: 1
# - Added Parameters: 3
# - Removed Parameters: 
#   - multi_download(): timeout
```
The `repo_breakages()` function returns a list with two major items: "Summary"
and "Details".  The "Summary" item is a data frame showing the input packages,
the input package versions, and a TRUE or FALSE flag that describes whether
there are any breaking changes between the two package versions.  The "details" 
list item then provides the difference object for any package that had 
a breaking change. This function tells you exactly what you need to know
before a repository upgrade.  If desired, you can then use the difference 
information to scan your code, and determine whether those removals will
break anything.

### Entire Repository

The above demonstration showed how to use `repo_breakages()` on a small number
of packages.  Now let's see how to use the function on a large number of packages.
In fact, we will perform an upgrade analyse on our entire R repository:

```{r eval=FALSE, echo=TRUE}

# Get current versions from local repo
r1 <- pkg_repo(ver = "current")

# View Results
r1
# # A package repo object
# - Repo Version: 4.2
# - Packages:
#            Package              Version
# 1            abind                1.4-5
# 2          admiral               0.10.1
# 3       admiraldev                0.3.0
# 4      admiralonco                0.3.0
# 5             ards                0.1.0
# 6            arrow             11.0.0.3
# 7          askpass                  1.1
# 8       assertthat                0.2.1
# 9           aws.s3               0.3.21
# 10   aws.signature                0.6.0
# 11       backports                1.4.1
# 12       base64enc                0.1-3
# 13             bit                4.0.4
# 14           bit64                4.0.5
# 15          bitops                1.0-7
# 16            blob                1.2.3
# ...

# Get latest versions of current repo on CRAN
r2 <- pkg_repo(r1$Package, ver = "latest")

# View Results
r2
# # A package repo object
# - Repo Version: latest
# - Packages:
#            Package    Version
# 1            abind      1.4-8
# 2          admiral      1.2.0
# 3       admiraldev      1.2.0
# 4      admiralonco      1.2.0
# 5             ards      0.1.1
# 6            arrow   18.1.0.1
# 7          askpass      1.2.1
# 8       assertthat      0.2.1
# 9           aws.s3     0.3.21
# 10   aws.signature      0.6.0
# 11       backports      1.5.0
# 12       base64enc      0.1-3
# 13             bit    4.5.0.1
# 14           bit64    4.6.0-1
# 15          bitops      1.0-9
# 16            blob      1.2.4
# ...


# Check for breakages on whole repository
res <- repo_breakages(r1, r2)
# Comparing abind v1.4-5/v1.4-8
# Comparing admiral v0.10.1/v1.2.0
# Comparing admiraldev v0.3.0/v1.2.0
# Comparing admiralonco v0.3.0/v1.2.0
# Comparing ards v0.1.0/v0.1.1
# Archive versions of 'ards' not available.
# Comparing arrow v11.0.0.3/v18.1.0.1
# Comparing askpass v1.1/v1.2.1
# Comparing backports v1.4.1/v1.5.0
# Comparing bit v4.0.4/v4.5.0.1
# Comparing bit64 v4.0.5/v4.6.0-1
# Comparing bitops v1.0-7/v1.0-9
# Comparing blob v1.2.3/v1.2.4
# Comparing box v1.1.3/v1.2.0
# Comparing brew v1.0-8/v1.0-10
# Comparing brio v1.1.3/v1.1.5
# Comparing broom v1.0.1/v1.0.7
# Comparing bslib v0.4.0/v0.9.0
# Comparing cachem v1.0.6/v1.1.0
# Comparing callr v3.7.3/v3.7.6
# ...
```
When the processing finishes, you can examine the results using dollar sign ($)
notation:

```{r eval=FALSE, echo=TRUE}
# View Summary Table
res$Summary
#            Package             Version1   Version2 Breakages
# 1            abind                1.4-5      1.4-8     FALSE
# 2          admiral               0.10.1      1.2.0      TRUE
# 3       admiraldev                0.3.0      1.2.0      TRUE
# 4      admiralonco                0.3.0      1.2.0      TRUE
# 5             ards                0.1.0      0.1.1        NA
# 6            arrow             11.0.0.3   18.1.0.1     FALSE
# 7          askpass                  1.1      1.2.1     FALSE
# 8       assertthat                0.2.1      0.2.1     FALSE
# 9           aws.s3               0.3.21     0.3.21     FALSE
# 10   aws.signature                0.6.0      0.6.0     FALSE
# 11       backports                1.4.1      1.5.0     FALSE
# 12       base64enc                0.1-3      0.1-3     FALSE
# 13             bit                4.0.4    4.5.0.1     FALSE
# 14           bit64                4.0.5    4.6.0-1      TRUE
# 15          bitops                1.0-7      1.0-9     FALSE
# 16            blob                1.2.3      1.2.4     FALSE
# ...

# View Difference Details
res$Details$admiralonco
# # A difference object: admiralonco package
# - Comparing: v0.3.0/v1.2.0
# - Breaking Changes: TRUE
# - Added Functions: 
#   - bor_cr()
#   - bor_ne()
#   - bor_non_crpd()
#   - bor_pd()
#   - bor_pr()
#   - bor_sd()
#   - cb_y()
#   - cbor_cr()
#   - cbor_pr()
#   - crsp_y_cr()
#   - crsp_y_pr()
#   - no_data_missing()
#   - no_data_n()
#   - rsp_y()
# - Removed Functions: 
#   - call_aval_fun()
#   - derive_param_lasta()

```
What is convenient is that you can run
these functions before any new package versions are downloaded and installed.  
They work on archive packages stored on CRAN, not on your local machine.
In this way, prior to any actual upgrade, 
you can make a determination as to a) whether to proceed with 
the upgrade or not, and b) how much effort it might take to fix any breakages.

## Assessing Entire Repository

In addition to anticipating breakages for a large number of packages,
**pkgdiff** also allows you to gather stability scores for a large number of 
packages.  Bulk stability scores are achieved using `repo_stability()`.  Here is
how to do it:

```{r eval=FALSE, echo=TRUE}
# Create vector of packages
v1 <- c("curl", "dplyr", "dbplyr", "purrr")

# Assess stability 
res <- repo_stability(v1)
# Getting stability score for package 'curl'...
# Getting stability score for package 'dplyr'...
# Getting stability score for package 'dbplyr'...
# Getting stability score for package 'purrr'...

# View results
res
# # A repo stability object
# - Run Datetime: 2025-03-01 17:46 UTC
# - Summary:
#   Package    FV    LV         FR         LR TR BR Score        Assessment
# 1    curl   0.2 6.2.1 2014-11-20 2025-02-19 51  1  98.0       Very Stable
# 2   dplyr   0.1 1.1.4 2014-01-16 2023-11-17 45 20  87.5 Somewhat Unstable
# 3  dbplyr 1.0.0 2.5.0 2017-06-09 2024-03-19 23 10  74.1          Unstable
# 4   purrr 0.1.0 1.0.4 2015-09-28 2025-02-05 18  6  92.1            Stable

```
The above result shows each package on the input vector, the first and last
versions, the first and last release dates, and the total number of releases.
It also shows the number of breaking releases, a stability score, and 
a stablity assessment for each package.  This information is very helpful 
to get an overall sense of the stability of your repository, and where 
breaking changes might occur when they do.

Like `repo_breakages()`, `repo_stability()` can be used to assess an entire
repository.  Simply use `pkg_repo()` to get the packages in your current repo,
and pass the "Packages" column into `repo_stability()`.  Remember to give 
`repo_stability()` plenty of uninterrupted processing time if you are 
assessing a large number of packages.


## Next Steps

For addition research on **pkgdiff**, you may wish to browse the following 
articles:

* [Stability Score](pkgdiff-stability.html)
* [Package Cache](pkgdiff-cache.html)
* [Frequently Asked Questions](pkgdiff-faq.html)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgsummary.R
\name{pkg_summary}
\alias{pkg_summary}
\title{Generate a Package Summary Report}
\usage{
pkg_summary(pkg, releases = NULL, months = NULL, view = TRUE, path = NULL)
}
\arguments{
\item{pkg}{A quoted string containing the name of the package to report on.
This parameter is required.}

\item{releases}{An integer indicating the number of releases to collect
stability data for. For example, \code{releases = 10} will return stability
data for the last 10 releases of the package. Default is NULL, which means
the function will return data for all releases.}

\item{months}{An integer indicating the number of months back to collect
stability data for.  For example, \code{months = 24} will collect
stability data for the previous 2 years.  Default is NULL, meaning there is
no limitation on the number of release months, and the function will collect
data from all releases.}

\item{view}{Whether to display the report in the viewer.  By default,
the parameter is TRUE.}

\item{path}{A path and file name for the report.  If NULL, the function
will create a temporary file. Default is NULL.}
}
\value{
The path to the HTML report, invisibly.  If the \code{view} parameter is TRUE,
the function will send the report to the viewer.
}
\description{
The \code{pkg_summary} function creates a package summary report.
The report aggregates some key information about the package
and a generates a chart showing breaking changes over time.
The function provides a nice snapshot that can help you get a feel for a package.
}
\details{
The \code{pkg_summary} function generates an HTML
report that can be sent to the RStudio viewer.  It combines some information
from both the \code{pkg_info} and \code{pkg_stability} functions, and displays
it in a table and accompanying chart.

The table shows a selection of fields
that can give you an overall sense of the package: its size, age, number
of releases, dependencies, etc.  The chart is a stacked bar chart
showing the number of breaking changes for each release.

Here is an example of the package summary report:
\figure{summary1.png}
The chart shows each release issued by the package maintainer
over the last ~8.5 years. Releases prior to that are not displayed,
as they are not included in the stability score. The bars represent the
number of removed functions and removed parameters. The labels on each
bar show the release year, month and version.
}
\examples{
# View summary report for "stringr" package
# - Set view = TRUE to see example
pkg_summary("stringr", view = FALSE)

}
\seealso{
For additional information, see \code{\link{pkg_info}} and
\code{\link{pkg_stability}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgInfo.R
\name{pkg_versions}
\alias{pkg_versions}
\title{Retrieves All Versions of a Package}
\usage{
pkg_versions(pkg)
}
\arguments{
\item{pkg}{A package name to return versions for.  Pass the package
name as a quoted string.  This parameter is required.}
}
\value{
A data frame with one row per package release, showing the
version and date of each release. If the package is not found on CRAN,
a message will be generated and the function will return NULL.
}
\description{
The \code{pkg_versions} function queries the CRAN archive,
and returns a data frame of package version information for
all releases of the package. This function may be used to understand
what versions of a package are available. It is often used in
conjunction with \code{\link{pkg_info}} to get a general overview
of a package.
}
\examples{
# View package versions
pkg_versions("rsample")
#    Package Version             FileName    Release   Size
# 1  rsample   1.2.1 rsample_1.2.1.tar.gz 2024-03-25 320.9K
# 2  rsample   1.2.0 rsample_1.2.0.tar.gz 2023-08-23   321K
# 3  rsample   1.1.1 rsample_1.1.1.tar.gz 2022-12-07   318K
# 4  rsample   1.1.0 rsample_1.1.0.tar.gz 2022-08-08   306K
# 5  rsample   1.0.0 rsample_1.0.0.tar.gz 2022-06-24   268K
# 6  rsample   0.1.1 rsample_0.1.1.tar.gz 2021-11-08   274K
# 7  rsample   0.1.0 rsample_0.1.0.tar.gz 2021-05-08   274K
# 8  rsample   0.0.9 rsample_0.0.9.tar.gz 2021-02-17   269K
# 9  rsample   0.0.8 rsample_0.0.8.tar.gz 2020-09-23   261K
# 10 rsample   0.0.7 rsample_0.0.7.tar.gz 2020-06-04   248K
# 11 rsample   0.0.6 rsample_0.0.6.tar.gz 2020-03-31   299K
# 12 rsample   0.0.5 rsample_0.0.5.tar.gz 2019-07-13   297K
# 13 rsample   0.0.4 rsample_0.0.4.tar.gz 2019-01-07   254K
# 14 rsample   0.0.3 rsample_0.0.3.tar.gz 2018-11-20   254K
# 15 rsample   0.0.2 rsample_0.0.2.tar.gz 2017-11-12   339K
# 16 rsample   0.0.1 rsample_0.0.1.tar.gz 2017-07-08   180K
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgInfo.R
\name{pkg_cache}
\alias{pkg_cache}
\title{Queries the Package Cache}
\usage{
pkg_cache(pkgs = NULL)
}
\arguments{
\item{pkgs}{A vector of package names to retrieve cache information about.
Default is NULL, which means to return all packages in the cache.}
}
\value{
An data frame showing the package name and latest package
version stored in the cache.  If the package is not stored in the cache,
the package version will be NA.  The "Last Update" time stamp is the
last day and time the cache was updated.
}
\description{
The \code{pkg_cache} function queries the package cache, and
returns information on which packages are included in the cache.
The function also returns the last version of the package cached.
}
\details{
The \strong{pkgdiff} cache is used to speed up \strong{pkgdiff} functions.
Information about each package is pre-processed and stored in the cache
on Github.  The functions then retrieve this pre-processed information
instead of pulling packages down from CRAN.  To learn more about the cache,
refer to \code{vignette('pkgdiff-cache')}.
}
\examples{
# View single package
pkg_cache("dplyr")
# # A package cache object
# - Last Update: 2025-02-25 14:58 UTC
# - Packages:
#   Package Version
# 1   dplyr   1.1.4

# View multiple packages
pkgs <- c("dplyr", "tidyr", "stringr")
pkg_cache(pkgs)
# # A package cache object
# - Last Update: 2025-02-25 14:58 UTC
# - Packages:
#   Package Version
# 1   dplyr   1.1.4
# 2   tidyr   1.3.1
# 3 stringr   1.5.1

# Get all packages
res <- pkg_cache()

# View first 10
res[1:10, ]
# A package cache object
# - Last Update: 2025-02-25 14:58 UTC
# - Packages:
#        Package Version
# 1           A3   1.0.0
# 2   abbreviate     0.1
# 3     abc.data     1.1
# 4          abc   2.2.2
# 5  ABCanalysis   1.2.1
# 6          abd   0.2-8
# 7        abind   1.4-8
# 8      acepack   1.6.1
# 9       actuar   3.3-5
# 10         ada   2.0-5
}
\seealso{
Other pdiff: 
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stability.R
\name{print.pdiff_score}
\alias{print.pdiff_score}
\title{Print a Stability Score Object}
\usage{
\method{print}{pdiff_score}(x, ...)
}
\arguments{
\item{x}{The package stability score to print}

\item{...}{Follow-on parameters to the print function}
}
\value{
The stability score object, invisibly.
}
\description{
Print routine for a package stability score object of class "pdiff_score".
}
\examples{
# View package stability
res <- pkg_stability("rsample")

# Print stability scores
print(res)
# # A stability score: rsample package
# - Age: 7.67 years
# - Score: 80.7
# - Assessment: Somewhat Unstable
# - Version Range: 0.0.1/1.2.1
# - Release Range: 2017-07-08/2024-03-25
# - Release Count: 16
# - Breaking Releases: 6
# - Data:
#    Package Version             FileName    Release   Size  AF AP RF RP BC  TF
# 1  rsample   1.2.1 rsample_1.2.1.tar.gz 2024-03-25   321K   1  0  0  0  0 401
# 2  rsample   1.2.0 rsample_1.2.0.tar.gz 2023-08-23   321K  43  6  1  3  1 400
# 3  rsample   1.1.1 rsample_1.1.1.tar.gz 2022-12-07   318K  37  5  0  0  0 358
# 4  rsample   1.1.0 rsample_1.1.0.tar.gz 2022-08-08   306K  72  1  0  0  0 321
# 5  rsample   1.0.0 rsample_1.0.0.tar.gz 2022-06-24   268K   1  0  1  0  1 249
# 6  rsample   0.1.1 rsample_0.1.1.tar.gz 2021-11-08   274K   4  1  0  1  1 249
# 7  rsample   0.1.0 rsample_0.1.0.tar.gz 2021-05-08   274K   0  5  0  0  0 245
# 8  rsample   0.0.9 rsample_0.0.9.tar.gz 2021-02-17   269K  19  0  6  0  1 245
# 9  rsample   0.0.8 rsample_0.0.8.tar.gz 2020-09-23   261K  59  0  0  0  0 232
# 10 rsample   0.0.7 rsample_0.0.7.tar.gz 2020-06-04   248K 101  1  0  0  0 173
# 11 rsample   0.0.6 rsample_0.0.6.tar.gz 2020-03-31   299K   7  2  0  0  0  72
# 12 rsample   0.0.5 rsample_0.0.5.tar.gz 2019-07-13   297K   4  5  0  0  0  65
# 13 rsample   0.0.4 rsample_0.0.4.tar.gz 2019-01-07   254K   0  0  4  0  1  61
# 14 rsample   0.0.3 rsample_0.0.3.tar.gz 2018-11-20   254K   7  0  1  0  1  65
# 15 rsample   0.0.2 rsample_0.0.2.tar.gz 2017-11-12   339K  24  0  0  0  0  59
# 16 rsample   0.0.1 rsample_0.0.1.tar.gz 2017-07-08   180K  35 35  0  0  0  35
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgInfo.R
\name{print.pcache}
\alias{print.pcache}
\title{Print a Package Cache Object}
\usage{
\method{print}{pcache}(x, ...)
}
\arguments{
\item{x}{The package cache to print.}

\item{...}{Follow-on parameters to the print function.}
}
\value{
The package cache object, invisibly.
}
\description{
Print routine for a package cache object of class "pcache".
}
\examples{
# Create vector of packages
pkgs <- c("ggplot2", "patchwork", "gt")

# Capture cache versions
res <- pkg_cache(pkgs)

# Print cache versions
print(res)
# # A package cache object
# - Last Update: 2025-03-01 20:51 UTC
# - Packages:
#     Package Version
# 1   ggplot2   3.5.1
# 2 patchwork   1.3.0
# 3        gt  0.11.1
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgrepo.R
\name{print.prepo}
\alias{print.prepo}
\title{Print a Package Repo Object}
\usage{
\method{print}{prepo}(x, ...)
}
\arguments{
\item{x}{The package repo to print.}

\item{...}{Follow-on parameters to the print function.}
}
\value{
The package repo object, invisibly.
}
\description{
Print routine for a package repo object of class "prepo".
}
\examples{
# Create vector of packages
pkgs <- c("tidymodels", "rsample", "parsnip", "recipes", "workflows")

# Retrieve latest versions
res <- pkg_repo(pkgs, ver = "latest")

# Print results
print(res)
# # A package repo object
# - Repo Version: latest
# - Packages:
#      Package Version
# 1    parsnip   1.3.0
# 2    recipes   1.1.1
# 3    rsample   1.2.1
# 4 tidymodels   1.3.0
# 5  workflows   1.2.0
}
\seealso{
Other prepo: 
\code{\link{pkg_repo}()},
\code{\link{print.rbreak}()},
\code{\link{print.rstability}()},
\code{\link{repo_breakages}()},
\code{\link{repo_stability}()}
}
\concept{prepo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgInfo.R
\name{pkg_info}
\alias{pkg_info}
\title{Get Information for a Package}
\usage{
pkg_info(pkg, ver = "current", cache = TRUE)
}
\arguments{
\item{pkg}{The package name as a quoted string. This parameter is
required.}

\item{ver}{The version of the package to retrieve information for.
Pass the version as a quoted string.
Special values are "current" and
"latest".  The value "current" is the current version of the package
running on the machine.  The value "latest" is the latest version
of the package from CRAN.}

\item{cache}{Whether to retrieve the info from the Github cache, or
from CRAN.  If TRUE, the function will first search the cache, and
return the info if available.  If the info is not available in the
Github cache, or the cache parameter is set to FALSE, the info will be
retrieved from CRAN.}
}
\value{
A package information object of class "pinfo".  This object contains
a set of general information about the package, such as the version,
release date, maintainer, title, etc.  Most of this information comes
from the package description file.  In addition, the info object also
contains a list of functions in the package and their parameters.

If the package has been archived on CRAN, info will be returned,
but the "Archived" flag will be set to TRUE. If the
package is not found on CRAN, the function will emit a message and return
NULL.

If the package is a Base R package, a reduced number of fields will be returned,
and the package will be marked with \code{Base Package: TRUE}.  Function
lists are not available for Base packages.
}
\description{
Combines information related to a package. Objects of this class are
used by several other functions in the \strong{pkgdiff} system.  This
class also makes a compact storage format for pre-processed package
information.
}
\details{
Package information is unique for each version of a package.  Since
a package can have multiple versions, you must select which version
you wish to return information for.  By default, the version
returned is the current version of the package on the current machine.
You may also specify a previous version number from the CRAN archive.
Another option is to specify the latest version on CRAN.  See the
"ver" parameter for additional details on how to select these
different package versions.

The package info object contains a list of functions and function
parameters associated with the specified version of the package.
You may access this list using the \code{$Functions} item name.

Most other information contained in the info object is retrieved
from the package description file. One exception is the downloads
per month.  This information is retrieved from CRAN logs.

In addition, the package cache status is appended to the info object.
The package cache status indicates whether the package info
has been stored in the \strong{pkgdiff} Github cache.  Packages that
have been stored in the cache enjoy faster results from
\strong{pkgdiff} functions.

Note that \code{pkg_info} and other \strong{pkgdiff} functions
only work with contributed CRAN packages.  They do not
work with Base R packages.
}
\examples{
# View package info
pkg_info("glue")
# A package info object: glue package
# - Version: v1.7.0
# - Release Date: 2024-01-09
# - Title: Interpreted String Literals
# - Maintainer: Jennifer Bryan <jenny@posit.co>
# - License: MIT + file LICENSE
# - Description: An implementation of interpreted string literals, inspired by
# Python's Literal String Interpolation
# <https://www.python.org/dev/peps/pep-0498/> and Docstrings
# <https://www.python.org/dev/peps/pep-0257/> and Julia's Triple-Quoted
# String Literals
# <https://docs.julialang.org/en/v1.3/manual/strings/#Triple-Quoted-String-Literals-1>.
#   - Depends: R (>= 3.6)
# - Imports: methods
# - Suggests: crayon, DBI (>= 1.2.0), dplyr, knitr, magrittr, rlang,
# rmarkdown, RSQLite, testthat (>= 3.2.0), vctrs (>= 0.3.0),
# waldo (>= 0.3.0), withr
# - Downloads/Month: 1463244
# - Repository: CRAN
# - Cached: TRUE
# - Functions: 24

# Get info object
res <- pkg_info("glue")

# Extract package version
res$Version
# [1] "1.7.0"

# Extract maintainer
res$Maintainer
# [1] "Jennifer Bryan <jenny@posit.co>"

# Extract function list
res$Functions
# $`[.glue`
# [1] "x"   "i"   "..."
#
# $`[[.glue`
# [1] "x"   "i"   "..."
#
# $`+.glue`
# [1] "e1" "e2"
#
# $as.character.glue
# [1] "x"   "..."
#
# $as_glue
# [1] "x"   "..."
#
# $as_glue.character
# [1] "x"   "..."
#
# $as_glue.default
# [1] "x"   "..."
#
# $as_glue.glue
# [1] "x"   "..."
#
# $backtick
# [1] "x"
#
# $double_quote
# [1] "x"
#
# $glue
# [1] "..."          ".sep"         ".envir"
# [4] ".open"        ".close"       ".na"
# [7] ".null"        ".comment"     ".literal"
# [10] ".transformer" ".trim"
# ...
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgrepo.R
\name{pkg_repo}
\alias{pkg_repo}
\title{Retrieve Package Versions from a Repository}
\usage{
pkg_repo(pkgs = NULL, ver = "current", libpaths = NULL)
}
\arguments{
\item{pkgs}{A vector of package names used to subset the repository list.
Default is NULL, which means all packages in the repository will be returned.}

\item{ver}{The R version of the repository. Pass the R version
as a quoted string. Default is "current",
meaning the current version of the repository. The value "latest"
will return the latest versions on CRAN.}

\item{libpaths}{A vector of paths specifying the locations of the repositories
to query. Default
is NULL, meaning the function will use the default R locations on the machine.
These default paths are identified by the Base R function \code{.libPaths()}.}
}
\value{
An object of class "prepo".  The object will contain a data
frame of R packages with corresponding version numbers.  This table
may then be passed to \code{\link{repo_breakages}} or
\code{\link{repo_stability}}.

If a package is not found in the repo,
a row will still be returned for the package.  However, the version
will appear as NA.
}
\description{
The \code{pkg_repo} function combines information related to a package repository.
The function retrieves all the packages in a repository for a specified version
of R.  Results will list the package name and version.
}
\examples{
# Create vector of packages
pkgs <- c("common", "dplyr", "rvest", "stringr")

# Retrieve latest versions
pkg_repo(pkgs, ver = "latest")
# # A package repo object
# - Repo Version: latest
# - Packages:
#   Package Version
# 1  common   1.1.3
# 2   dplyr   1.1.4
# 3   rvest   1.0.4
# 4 stringr   1.5.1

}
\seealso{
Other prepo: 
\code{\link{print.prepo}()},
\code{\link{print.rbreak}()},
\code{\link{print.rstability}()},
\code{\link{repo_breakages}()},
\code{\link{repo_stability}()}
}
\concept{prepo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{print.pdiff}
\alias{print.pdiff}
\title{Print a Package Difference Object}
\usage{
\method{print}{pdiff}(x, ..., verbose = TRUE)
}
\arguments{
\item{x}{The package difference to print}

\item{...}{Follow-on parameters to the print function}

\item{verbose}{Whether to print all added and removed functions
and function parameters.  Default is TRUE.  If FALSE, the function
will print removed functions and parameters, but only print
counts of added functions and parameters.}
}
\value{
The package difference object, invisibly.
}
\description{
Print routine for a package difference object of class "pdiff".
}
\examples{
# View versions
pkg_versions("rvest")
#    Package Version           FileName    Release Size
# 1    rvest   1.0.4 rvest_1.0.4.tar.gz 2024-02-12 113K
# 2    rvest   1.0.3 rvest_1.0.3.tar.gz 2022-08-20  92K
# 3    rvest   1.0.2 rvest_1.0.2.tar.gz 2021-10-17  92K
# 4    rvest   1.0.1 rvest_1.0.1.tar.gz 2021-07-26  92K
# 5    rvest   1.0.0 rvest_1.0.0.tar.gz 2021-03-09  92K
# 6    rvest   0.3.6 rvest_0.3.6.tar.gz 2020-07-25 2.7M
# 7    rvest   0.3.5 rvest_0.3.5.tar.gz 2019-11-09 1.1M
# 8    rvest   0.3.4 rvest_0.3.4.tar.gz 2019-05-15 1.6M
# 9    rvest   0.3.3 rvest_0.3.3.tar.gz 2019-04-11 1.6M
# 10   rvest   0.3.2 rvest_0.3.2.tar.gz 2016-06-17 1.5M
# 11   rvest   0.3.1 rvest_0.3.1.tar.gz 2015-11-11 1.5M
# 12   rvest   0.3.0 rvest_0.3.0.tar.gz 2015-09-23 1.5M
# 13   rvest   0.2.0 rvest_0.2.0.tar.gz 2015-01-01 5.2M
# 14   rvest   0.1.0 rvest_0.1.0.tar.gz 2014-11-22 5.1M

# Find differences in latest version
res <- pkg_diff("rvest", "1.0.3", "1.0.4")

# View differences
print(res)
# # A difference object: rvest package
# - Comparing: v1.0.3/v1.0.4
# - Breaking Changes: FALSE
# - Added Functions:
#   - html_element.LiveHTML()
#   - html_elements.LiveHTML()
#   - html_table.LiveHTML()
#   - LiveHTML()
#   - read_html_live()
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reports.R
\name{repo_stability}
\alias{repo_stability}
\title{Generate Stability Scores for a Repository}
\usage{
repo_stability(pkgs, releases = NULL, months = NULL)
}
\arguments{
\item{pkgs}{A vector of package names.}

\item{releases}{An integer number of releases to assess.  The
function will then limit the scope of the assessment to the specified
number of releases.}

\item{months}{An integer number of months from the current date.  The
function will then limit the scope of the assessment to the specified
number of months.}
}
\value{
A data frame of information regarding the stability of
each package in the input vector.
}
\description{
The function generates stability scores
for a vector of packages. If passing a large number of packages,
be prepared for the
function to run for a considerable amount of time.
}
\details{
To assess stability for a package, \strong{pkgdiff} has to compare all
releases of a package, and identify breaking changes between each release.
This comparison takes time, especially for packages that have been
active for many years.

Therefore, the first step of the function is to compare all releases
for each package in the input vector.  The function will send messages
to the console to let you know which package it is currently comparing.
These messages will let you know how far the function is through the processing.

Once all the comparisons are completed, the function will return a data
frame summary of stability results.  The data frame will have one
row for each package in the input vector.  The data frame will also
show some information about the comparison, and give a stability assessment
for each package.  The data frame columns are as follows:
\itemize{
\item {\strong{Package}: The package name.}
\item {\strong{FV}: The first version of the package.}
\item {\strong{LV}: The last version of the package.}
\item {\strong{FR}: The first release date of the package.}
\item {\strong{LR}: The last release date of the package.}
\item {\strong{TR}: The total number of releases.}
\item {\strong{BR}: The number of breaking releases.}
\item {\strong{Score}: The stability score.}
\item {\strong{Assessment}: The stability assessment.}
}
To learn how the package stability score and assessment are calculated,
see \code{vignette("pkgdiff-stability")}.  Additional information is
included in the documentation of the \code{\link{pkg_stability}}
function.

If a package is not found in the package cache, the function
will download and compare each version of the package on CRAN.  This
process may increase the processing time.  See \code{vignette("pkg-cache")}
and \code{\link{pkg_cache}} for additional information.

If a package does not exist on CRAN, a row for that package will still
be returned.  However, all data values will be NA.  This situation may
occur if a package exists on Github, but has not yet been submitted
to CRAN.  The \strong{pkgdiff} stability functions only work with packages
that have been published to CRAN.
}
\examples{
# Create vector of packages
vct <- c("curl", "dplyr", "rvest", "tidymodels")

# Get stablity scores
res <- repo_stability(vct)
# Getting stability score for package 'curl'...
# Getting stability score for package 'dplyr'...
# Getting stability score for package 'rvest'...
# Getting stability score for package 'tidymodels'...

# View stability results
res
# A repo stability object
# - Run Datetime: 2025-03-01 17:53 UTC
# - Summary:
#   Package    FV    LV         FR         LR TR BR Score        Assessment
# 1       curl   0.2 6.2.1 2014-11-20 2025-02-19 51  1  98.0       Very Stable
# 2      dplyr   0.1 1.1.4 2014-01-16 2023-11-17 45 20  87.5 Somewhat Unstable
# 3      rvest 0.1.0 1.0.4 2014-11-22 2024-02-12 14  4  93.2            Stable
# 4 tidymodels 0.0.1 1.3.0 2018-07-27 2025-02-21 14  0 100.0           Perfect

}
\seealso{
Other prepo: 
\code{\link{pkg_repo}()},
\code{\link{print.prepo}()},
\code{\link{print.rbreak}()},
\code{\link{print.rstability}()},
\code{\link{repo_breakages}()}
}
\concept{prepo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{pkg_diff}
\alias{pkg_diff}
\title{Compare Package Versions}
\usage{
pkg_diff(pkg, v1 = "current", v2 = "latest")
}
\arguments{
\item{pkg}{The package name.}

\item{v1}{The earlier package version. Pass the version value as a quoted string.
Default is "current", which means the function will look up the currently
installed version.}

\item{v2}{The later package version. Pass the version value as a quoted string.
The default is "latest", which is the
latest version of the package available on CRAN.}
}
\value{
An object of class "pdiff", which contains information regarding
the differences between two versions of a package.  The object includes
a list of functions that were added, a list of function parameters that were added,
a list of functions removed, and a list of function parameters that
were removed.  The object also contains some general information like
the package versions examined, and a TRUE or FALSE flag indicating whether
there were any breaking changes.
}
\description{
The \code{pkg_diff} function is used to compare two different versions of
a package.  The function returns an object that contains the differences.
Differences include any functions added or removed, and any function parameters
added or removed. The object may
be printed directly, or stored in a variable and examined programmatically.
}
\details{
The difference operation only compares functions that are exported
from the package. It does not compare internal functions or function documentation.
Exported functions are identified in the package namespace.

The \code{pkg_diff} function can compare any two versions of a package.
They do not need to be consecutive. They must, however, exist in the CRAN
archive.

If the package versions are included in the \strong{pkgdiff} cache, the
function will pull them from there.  Otherwise, they will be pulled from the
CRAN mirror.  The function does not access any packages on the local system.
}
\examples{
# View package stability
pkg_stability("rsample")
# # A stability score: rsample package
# - Age: 7.58 years
# - Score: 80.3
# - Assessment: Somewhat Unstable
# - Version Range: 0.0.1/1.2.1
# - Release Range: 2017-07-08/2024-03-25
# - Release Count: 16
# - Breaking Releases: 6
# - Data:
#   Package Version             FileName    Release   Size  AF AP RF RP BC  TF
# 1  rsample   1.2.1 rsample_1.2.1.tar.gz 2024-03-25 320.9K   1  0  0  0  0 401
# 2  rsample   1.2.0 rsample_1.2.0.tar.gz 2023-08-23   321K  43  6  1  3  1 400
# 3  rsample   1.1.1 rsample_1.1.1.tar.gz 2022-12-07   318K  37  5  0  0  0 358
# 4  rsample   1.1.0 rsample_1.1.0.tar.gz 2022-08-08   306K  72  1  0  0  0 321
# 5  rsample   1.0.0 rsample_1.0.0.tar.gz 2022-06-24   268K   1  0  1  0  1 249
# 6  rsample   0.1.1 rsample_0.1.1.tar.gz 2021-11-08   274K   4  1  0  1  1 249
# 7  rsample   0.1.0 rsample_0.1.0.tar.gz 2021-05-08   274K   0  5  0  0  0 245
# 8  rsample   0.0.9 rsample_0.0.9.tar.gz 2021-02-17   269K  19  0  6  0  1 245
# 9  rsample   0.0.8 rsample_0.0.8.tar.gz 2020-09-23   261K  59  0  0  0  0 232
# 10 rsample   0.0.7 rsample_0.0.7.tar.gz 2020-06-04   248K 101  1  0  0  0 173
# 11 rsample   0.0.6 rsample_0.0.6.tar.gz 2020-03-31   299K   7  2  0  0  0  72
# 12 rsample   0.0.5 rsample_0.0.5.tar.gz 2019-07-13   297K   4  5  0  0  0  65
# 13 rsample   0.0.4 rsample_0.0.4.tar.gz 2019-01-07   254K   0  0  4  0  1  61
# 14 rsample   0.0.3 rsample_0.0.3.tar.gz 2018-11-20   254K   7  0  1  0  1  65
# 15 rsample   0.0.2 rsample_0.0.2.tar.gz 2017-11-12   339K  24  0  0  0  0  59
# 16 rsample   0.0.1 rsample_0.0.1.tar.gz 2017-07-08   180K  35 35  0  0  0  35

# Examine differences between versions
pkg_diff("rsample", "1.1.1", "1.2.0")
# # A difference object: rsample package
# - Comparing: v1.1.1/v1.2.0
# - Breaking Changes: TRUE
# - Added Functions:
#   - analysis.default()
#   - analysis.initial_validation_split()
#   - analysis.rsplit()
#   - analysis.val_split()
#   - assessment.default()
#   - assessment.initial_validation_split()
#   - assessment.rsplit()
#   - assessment.val_split()
#   - dim.initial_validation_split()
#   - group_initial_validation_split()
#   - initial_validation_split()
#   - initial_validation_time_split()
#   - int_bca.bootstraps()
#   - int_pctl.bootstraps()
#   - int_t.bootstraps()
#   - print.initial_validation_split()
#   - testing.default()
#   - testing.initial_validation_split()
#   - testing.rsplit()
#   - testing.val_split()
#   - training.default()
#   - training.initial_validation_split()
#   - training.rsplit()
#   - training.val_split()
#   - validation()
#   - validation.default()
#   - validation.initial_validation_split()
#   - validation.val_split()
#   - validation_set()
#   - vec_cast.data.frame.validation_set()
#   - vec_cast.tbl_df.validation_set()
#   - vec_cast.validation_set.data.frame()
#   - vec_cast.validation_set.tbl_df()
#   - vec_cast.validation_set.validation_set()
#   - vec_ptype_abbr.group_initial_validation_split()
#   - vec_ptype_abbr.initial_validation_split()
#   - vec_ptype_abbr.validation_set()
#   - vec_ptype2.data.frame.validation_set()
#   - vec_ptype2.tbl_df.validation_set()
#   - vec_ptype2.validation_set.data.frame()
#   - vec_ptype2.validation_set.tbl_df()
#   - vec_ptype2.validation_set.validation_set()
#   - vec_restore.validation_set()
# - Added Parameters:
#   - int_pctl(): ...
#   - int_t(): ...
#   - testing(): ...
#   - tidy.nested_cv(): unique_ind
#   - tidy.rset(): unique_ind
#   - training(): ...
# - Removed Functions:
#   - gather()
# - Removed Parameters:
#   - int_bca(): statistics, alpha, .fn
#   - int_pctl(): statistics, alpha
#   - int_t(): statistics, alpha
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgInfo.R
\name{print.pinfo}
\alias{print.pinfo}
\title{Print a Package Info Object}
\usage{
\method{print}{pinfo}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{The package info to print.}

\item{...}{Follow-on parameters to the print function.}

\item{verbose}{Whether to print all function names or not.
Default is FALSE.  When FALSE, only a count of the functions
will be printed.}
}
\value{
The package info object, invisibly.
}
\description{
Print routine for a package info object of class "pinfo".
}
\examples{
# Capture package info
res <- pkg_info("patchwork")

# View package info
print(res)
# # A package info object: patchwork package
# - Version: v1.3.0
# - Release Date: 2024-09-16
# - Title: The Composer of Plots
# - Maintainer: Thomas Lin Pedersen <thomasp85@gmail.com>
#   - License: MIT + file LICENSE
# - Description: The 'ggplot2' package provides a strong API for sequentially
# building up a plot, but does not concern itself with composition of multiple
# plots. 'patchwork' is a package that expands the API to allow for
# arbitrarily complex composition of plots by, among others, providing
# mathematical operators for combining multiple plots. Other packages that try
# to address this need (but with a different approach) are 'gridExtra' and
# 'cowplot'.
# - Imports: ggplot2 (>= 3.0.0), gtable, grid, stats, grDevices, utils,
# graphics, rlang (>= 1.0.0), cli, farver
# - Suggests: knitr, rmarkdown, gridGraphics, gridExtra, ragg,
# testthat (>= 2.1.0), vdiffr, covr, png, gt (>= 0.11.0)
# - Downloads/Month: 239555
# - Repository: CRAN
# - Cached: TRUE
# - Functions: 90
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_stability}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reports.R
\name{print.rbreak}
\alias{print.rbreak}
\title{Print a Repo Breakages Object}
\usage{
\method{print}{rbreak}(x, ..., verbose = FALSE)
}
\arguments{
\item{x}{The repo breakages object to print.}

\item{...}{Follow-on parameters to the print function.}

\item{verbose}{If FALSE, prints only the difference removals.
If TRUE, prints both additions and removals. Default is FALSE.}
}
\value{
The repo breakage object, invisibly.
}
\description{
Print routine for a repo breakages object of class "rbreak".
}
\seealso{
Other prepo: 
\code{\link{pkg_repo}()},
\code{\link{print.prepo}()},
\code{\link{print.rstability}()},
\code{\link{repo_breakages}()},
\code{\link{repo_stability}()}
}
\concept{prepo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stability.R
\name{pkg_stability}
\alias{pkg_stability}
\title{Calculate Stability Score for a Package}
\usage{
pkg_stability(pkg, releases = NULL, months = NULL)
}
\arguments{
\item{pkg}{The name of the package.}

\item{releases}{An integer indicating the number of releases to collect
stability data for. For example, \code{releases = 10} will return stability
data for the last 10 releases of the package. Default is NULL, which means
the function will return data for all releases.}

\item{months}{An integer indicating the number of months back to collect
stability data for.  For example, \code{months = 24} will collect
stability data for the previous 2 years.  Default is NULL, meaning there is
no limitation on the number of release months, and the function will collect
data from all releases.}
}
\value{
An stability score object of class "pdiff_score".  The object
contains the stability score, plus other useful information such as
the release and version ranges, the number of releases, and number of
breaking releases. All of these items can be accessed using dollar sign ($)
syntax.
}
\description{
The \code{pkg_stablity} function calculates a stability score for
a specified package.  The score is calculated as the percentage of
releases with no breaking changes.  A breaking change is defined as
either a removed function or removed function parameter. The object
also returns the raw data from which the stability score is calculated.
This data can be useful for review or even custom analytics.
}
\details{
The \code{pkg_stability} function is used to get an overall feel for the
stability of a package.  The stability score and other information returned
by the function are based on data retrieved from CRAN.  The function
compares each version of a package and determines if any functions or
function parameters have been removed from the previous release.
If a release removes functions or parameters contained in the previous release,
it is flagged as a "breaking release" or "breaking change".

The stability score is calculated as the percentage of non-breaking releases.
For example, if a package has 10 releases, and one breaking release,
the stability score will be .9, or 90\%.

The stability assessment is a categorization of the stability score. The
aim of the assessment is to provide a general evaluation of the package,
whether it is stable or unstable.  The
assessment has 5 categories:
\itemize{
\item {\strong{Perfect}: Stability score of 100.  Package has never
had a breaking release in its entire history.
}
\item {\strong{Very Stable}: Stability score between 95 and 100. Package
has had a breaking release, but very rarely: less than 1 in 20 releases.
}
\item {\strong{Stable}: Stability score between 90 and 95.  The package
has had some breaking releases, but it is still rather rare: less than
1 in 10 releases.
}
\item {\strong{Somewhat unstable}: Stability score between 80 and 90. The
package sometimes has a breaking release, at a rate of about 1 in 5.
}
\item {\strong{Unstable}: Stability score below 80.  The package breaks
more frequently than every 5th release.  This frequency of breaking releases
is considered unstable.
}
}
As can be seen above, the stability assessment is weighted heavily toward
the high end.  This categorization is intentionally designed to encourage
package stability, and discourage the number of breaking changes.
}
\section{Stability Data}{
 Stability calculations are based on a
data frame of stability information gathered from each package release.
The data frame has one row per release.  Each row
contains comparison information against the prior release.  The
columns are organized as follows:
\itemize{
\item {\strong{Package}: The package name.
}
\item {\strong{Version}: The version number of the release.
}
\item {\strong{FileName}: The file name of the package.
}
\item {\strong{Release}: The release date of the package.
}
\item {\strong{Size}: The size of the package file on disk.
}
\item {\strong{AF}: The number of functions added from the previous release.
}
\item {\strong{AP}: The number of functions that had parameters added from the
previous release.
}
\item {\strong{RF}: The number of functions removed from the previous release.
}
\item {\strong{RP}: The number of functions that had parameters removed from
the previous release.
}
\item {\strong{BC}: Whether the release had any breaking changes.
Breaking changes are removed functions or removed parameters.  Values
are zero or one.  The value one (1) means the release had breaking changes.
The value zero (0) means the release had no breaking changes.
}
\item {\strong{TF}: The total number of functions in the package.
}
}
The above data will be included in the function print out.  It can also be
accessed on the returned object from the "data" list item, i.e.
\code{obj$data}.
}

\section{Disclaimers}{
 Note that a "breaking release" does not factor
in the number of
functions deprecated. One deprecated function counts the same as 10 deprecated
functions within a release.  The methodology also does not differentiate
between removed
functions and removed parameters. They are both considered "breaking", and
counted equally.

The methodology also does not factor in
changed functions.  The reason is simply that it is much more difficult to
identify whether or not a changed function is breaking.  That is to say,
you may still have breaking changes that are not identified by the
\strong{pkgdiff} package.
}

\section{Performance}{
  Performance of the \code{pkg_stability} function
can vary greatly depending on the package selected.  The information
for some packages has been cached on Github, and can be retrieved very
quickly.  For packages that have not been cached, the function must
download each version of the package from the CRAN archive, and compare
consecutive versions.  This process can take considerable time.

The
most popular packages on CRAN have been cached.  If there is a package
you query frequently that has not been cached, please submit
an issue to the \strong{pkgdiff} issue log on Github. In the issue description
simply request that the package be added to the cache.

For more information on the package cache, see \code{vignette("pkgdiff-cache")}.

For more information about how the stability score and assessment are
calculated, see \code{vignette("pkgdiff-stability")}.
}

\examples{
# View package stability
pkg_stability("curl")
# # A stability score: curl package
# - Age: 10.33 years
# - Score: 98.0
# - Assessment: Very Stable
# - Version Range: 0.2/6.2.1
# - Release Range: 2014-11-20/2025-02-19
# - Release Count: 51
# - Breaking Releases: 1
# - Data:
#   Package Version          FileName    Release   Size AF AP RF RP BC TF
# 1     curl   6.2.1 curl_6.2.1.tar.gz 2025-02-19 911.7K  0  0  0  0  0 46
# 2     curl   6.2.0 curl_6.2.0.tar.gz 2025-01-23 911.3K  0  1  0  0  0 46
# 3     curl   6.1.0 curl_6.1.0.tar.gz 2025-01-06   911K  0  1  0  0  0 46
# 4     curl   6.0.1 curl_6.0.1.tar.gz 2024-11-14   911K  0  0  0  0  0 46
# 5     curl   6.0.0 curl_6.0.0.tar.gz 2024-11-05   911K  1  1  0  1  1 46
# 6     curl   5.2.3 curl_5.2.3.tar.gz 2024-09-20   700K  0  0  0  0  0 45
# 7     curl   5.2.2 curl_5.2.2.tar.gz 2024-08-26   701K  0  0  0  0  0 45
# 8     curl   5.2.1 curl_5.2.1.tar.gz 2024-03-02   700K  0  0  0  0  0 45
# 9     curl   5.2.0 curl_5.2.0.tar.gz 2023-12-08   699K  1  0  0  0  0 45
# 10    curl   5.1.0 curl_5.1.0.tar.gz 2023-10-02   697K  0  1  0  0  0 44
# 11    curl   5.0.2 curl_5.0.2.tar.gz 2023-08-14   697K  0  0  0  0  0 44
# 12    curl   5.0.1 curl_5.0.1.tar.gz 2023-06-08   697K  0  1  0  0  0 44
# 13    curl   5.0.0 curl_5.0.0.tar.gz 2023-01-12   666K  1  1  0  0  0 44
# 14    curl   4.3.3 curl_4.3.3.tar.gz 2022-10-06   655K  0  0  0  0  0 43
# 15    curl   4.3.2 curl_4.3.2.tar.gz 2021-06-23   775K  0  0  0  0  0 43
# 16    curl   4.3.1 curl_4.3.1.tar.gz 2021-04-30   775K  0  0  0  0  0 43
# 17    curl     4.3   curl_4.3.tar.gz 2019-12-02   658K  0  0  0  0  0 43
# 18    curl     4.2   curl_4.2.tar.gz 2019-09-24   652K  0  0  0  0  0 43
# 19    curl     4.1   curl_4.1.tar.gz 2019-09-16   652K  0  1  0  0  0 43
# 20    curl     4.0   curl_4.0.tar.gz 2019-07-22   370K  3  0  0  0  0 43
# 21    curl     3.3   curl_3.3.tar.gz 2019-01-10   363K  2  0  0  0  0 40
# 22    curl     3.2   curl_3.2.tar.gz 2018-03-28   358K  0  0  0  0  0 38
# 23    curl     3.1   curl_3.1.tar.gz 2017-12-12   358K  0  0  0  0  0 38
# 24    curl     3.0   curl_3.0.tar.gz 2017-10-06   403K  1  2  0  0  0 38
# 25    curl   2.8.1 curl_2.8.1.tar.gz 2017-07-22   400K  0  1  0  0  0 37
# 26    curl     2.7   curl_2.7.tar.gz 2017-06-26   397K  1  1  0  0  0 37
# 27    curl     2.6   curl_2.6.tar.gz 2017-04-27   396K  0  1  0  0  0 36
# 28    curl     2.5   curl_2.5.tar.gz 2017-04-14   396K  1  0  0  0  0 36
# 29    curl     2.4   curl_2.4.tar.gz 2017-03-24   394K  2  0  0  0  0 35
# 30    curl     2.3   curl_2.3.tar.gz 2016-11-24   391K  0  0  0  0  0 33
# 31    curl     2.2   curl_2.2.tar.gz 2016-10-21   389K  0  0  0  0  0 33
# 32    curl     2.1   curl_2.1.tar.gz 2016-09-22   388K  0  0  0  0  0 33
# 33    curl     2.0   curl_2.0.tar.gz 2016-09-17   389K 10  0  0  0  0 33
# 34    curl     1.2   curl_1.2.tar.gz 2016-08-13   287K  0  0  0  0  0 23
# 35    curl     1.1   curl_1.1.tar.gz 2016-07-26   277K  0  0  0  0  0 23
# 36    curl     1.0   curl_1.0.tar.gz 2016-07-24   277K  0  0  0  0  0 23
# 37    curl   0.9.7 curl_0.9.7.tar.gz 2016-04-10   258K  0  0  0  0  0 23
# 38    curl   0.9.6 curl_0.9.6.tar.gz 2016-02-17   256K  1  0  0  0  0 23
# 39    curl   0.9.5 curl_0.9.5.tar.gz 2016-01-23   255K  1  0  0  0  0 22
# 40    curl   0.9.4 curl_0.9.4.tar.gz 2015-11-20   240K  0  1  0  0  0 21
# 41    curl   0.9.3 curl_0.9.3.tar.gz 2015-08-25   240K  0  0  0  0  0 21
# 42    curl   0.9.2 curl_0.9.2.tar.gz 2015-08-08   239K  1  0  0  0  0 21
# 43    curl   0.9.1 curl_0.9.1.tar.gz 2015-07-04   237K  0  0  0  0  0 20
# 44    curl     0.9   curl_0.9.tar.gz 2015-06-19   236K  0  0  0  0  0 20
# 45    curl     0.8   curl_0.8.tar.gz 2015-06-06   235K  0  0  0  0  0 20
# 46    curl     0.7   curl_0.7.tar.gz 2015-05-22   236K  0  0  0  0  0 20
# 47    curl     0.6   curl_0.6.tar.gz 2015-05-19   235K 14  2  0  0  0 20
# 48    curl     0.5   curl_0.5.tar.gz 2015-02-01    16K  0  0  0  0  0  6
# 49    curl     0.4   curl_0.4.tar.gz 2015-01-08    16K  4  0  0  0  0  6
# 50    curl     0.3   curl_0.3.tar.gz 2014-12-01   7.5K  1  0  0  0  0  2
# 51    curl     0.2   curl_0.2.tar.gz 2014-11-20   5.5K  1  1  0  0  0  1

# View package stability
pkg_stability("ggplot2")
# # A stability score: ggplot2 package
# - Age: 17.75 years
# - Score: 82.5
# - Assessment: Somewhat Unstable
# - Version Range: 0.5/3.5.1
# - Release Range: 2007-06-01/2024-04-23
# - Release Count: 50
# - Breaking Releases: 29
# - Data:
#   Package Version               FileName    Release Size  AF  AP  RF RP BC  TF
# 1  ggplot2   3.5.1   ggplot2_3.5.1.tar.gz 2024-04-23 3.4M   0   4   0  0  0 711
# 2  ggplot2   3.5.0   ggplot2_3.5.0.tar.gz 2024-02-23 3.4M  37  92  26  6  1 711
# 3  ggplot2   3.4.4   ggplot2_3.4.4.tar.gz 2023-10-12 3.0M   0   0   0  0  0 700
# 4  ggplot2   3.4.3   ggplot2_3.4.3.tar.gz 2023-08-14 2.1M   0   0   0  0  0 700
# 5  ggplot2   3.4.2   ggplot2_3.4.2.tar.gz 2023-04-03 3.0M   3   0   0  1  1 700
# 6  ggplot2   3.4.1   ggplot2_3.4.1.tar.gz 2023-02-10 3.0M   2   1   0  0  0 697
# 7  ggplot2   3.4.0   ggplot2_3.4.0.tar.gz 2022-11-04 3.0M  28  15   0  0  0 695
# 8  ggplot2   3.3.6   ggplot2_3.3.6.tar.gz 2022-05-03 2.9M   0   0   0  1  1 667
# 9  ggplot2   3.3.5   ggplot2_3.3.5.tar.gz 2021-06-25 2.9M   0   0   0  0  0 667
# 10 ggplot2   3.3.4   ggplot2_3.3.4.tar.gz 2021-06-16 2.9M   8  16   0  1  1 667
# 11 ggplot2   3.3.3   ggplot2_3.3.3.tar.gz 2020-12-30 2.9M   0   0   0  0  0 659
# 12 ggplot2   3.3.2   ggplot2_3.3.2.tar.gz 2020-06-19 2.9M  17   5   0  1  1 659
# 13 ggplot2   3.3.1   ggplot2_3.3.1.tar.gz 2020-05-29 2.9M   0   0   0  0  0 642
# 14 ggplot2   3.3.0   ggplot2_3.3.0.tar.gz 2020-03-05 2.9M  68  47   0  1  1 642
# 15 ggplot2   3.2.1   ggplot2_3.2.1.tar.gz 2019-08-11 3.1M   0   0   0  0  0 574
# 16 ggplot2   3.2.0   ggplot2_3.2.0.tar.gz 2019-06-16 3.0M   3   7   3  0  1 574
# 17 ggplot2   3.1.1   ggplot2_3.1.1.tar.gz 2019-04-07 2.7M   0   2   0  0  0 574
# 18 ggplot2   3.1.0   ggplot2_3.1.0.tar.gz 2018-10-25 2.7M   5   8   0  0  0 574
# 19 ggplot2   3.0.0   ggplot2_3.0.0.tar.gz 2018-07-03 2.7M  89  53   6  4  1 569
# 20 ggplot2   2.2.1   ggplot2_2.2.1.tar.gz 2016-12-30 2.1M   0   0   0  0  0 486
# 21 ggplot2   2.2.0   ggplot2_2.2.0.tar.gz 2016-11-11 2.1M  22  27  21  2  1 486
# 22 ggplot2   2.1.0   ggplot2_2.1.0.tar.gz 2016-03-01 1.5M  11   6   0  2  1 485
# 23 ggplot2   2.0.0   ggplot2_2.0.0.tar.gz 2015-12-18 1.5M 159  94 103 33  1 474
# 24 ggplot2   1.0.1   ggplot2_1.0.1.tar.gz 2015-03-17 2.2M   0   0   0  0  0 418
# 25 ggplot2   1.0.0   ggplot2_1.0.0.tar.gz 2014-05-21 2.2M  12   5   0  0  0 418
# 26 ggplot2 0.9.3.1 ggplot2_0.9.3.1.tar.gz 2013-03-02 2.2M   0   1   0  0  0 406
# 27 ggplot2   0.9.3   ggplot2_0.9.3.tar.gz 2012-12-05 2.2M   4   4   0  0  0 406
# 28 ggplot2 0.9.2.1 ggplot2_0.9.2.1.tar.gz 2012-09-11 2.2M   0   0   0  0  0 402
# 29 ggplot2   0.9.2   ggplot2_0.9.2.tar.gz 2012-09-04 2.2M  26  16   8  6  1 402
# 30 ggplot2   0.9.1   ggplot2_0.9.1.tar.gz 2012-05-08 9.1M  13  11   0  3  1 384
# 31 ggplot2   0.9.0   ggplot2_0.9.0.tar.gz 2012-03-01 1.6M 282   4 149  5  1 371
# 32 ggplot2   0.8.9   ggplot2_0.8.9.tar.gz 2010-12-23 2.0M   4   1   2  0  1 238
# 33 ggplot2   0.8.8   ggplot2_0.8.8.tar.gz 2010-07-05 2.0M   6   0   2  0  1 236
# 34 ggplot2   0.8.7   ggplot2_0.8.7.tar.gz 2010-03-02 2.0M   2   0   0  0  0 232
# 35 ggplot2   0.8.6   ggplot2_0.8.6.tar.gz 2010-02-18 2.0M   0   0   0  0  0 230
# 36 ggplot2   0.8.5   ggplot2_0.8.5.tar.gz 2009-12-16 2.0M   9   2   4  0  1 230
# 37 ggplot2   0.8.4   ggplot2_0.8.4.tar.gz 2009-12-09 2.0M   3   0   2  0  1 225
# 38 ggplot2   0.8.3   ggplot2_0.8.3.tar.gz 2009-04-20 2.0M  51   2   6  2  1 224
# 39 ggplot2   0.8.2   ggplot2_0.8.2.tar.gz 2009-02-25 2.0M  11   1  10  0  1 179
# 40 ggplot2   0.8.1   ggplot2_0.8.1.tar.gz 2008-12-14 1.9M  19   3   3  3  1 178
# 41 ggplot2     0.8     ggplot2_0.8.tar.gz 2008-11-21 1.9M  25   6   2  2  1 162
# 42 ggplot2     0.7     ggplot2_0.7.tar.gz 2008-10-05 1.9M  45   7  26  3  1 139
# 43 ggplot2     0.6     ggplot2_0.6.tar.gz 2008-04-03 1.9M  30   5 125  2  1 120
# 44 ggplot2   0.5.7   ggplot2_0.5.7.tar.gz 2008-01-11 1.8M  16   3   3  0  1 215
# 45 ggplot2   0.5.6   ggplot2_0.5.6.tar.gz 2007-10-20 1.8M  36   0   5  0  1 202
# 46 ggplot2   0.5.5   ggplot2_0.5.5.tar.gz 2007-09-01 1.8M   1   0   0  2  1 171
# 47 ggplot2   0.5.4   ggplot2_0.5.4.tar.gz 2007-07-08 1.8M   3   0   1  0  1 170
# 48 ggplot2   0.5.2   ggplot2_0.5.2.tar.gz 2007-06-18 1.8M   2   0   0  0  0 168
# 49 ggplot2   0.5.1   ggplot2_0.5.1.tar.gz 2007-06-10 1.8M  11   1   5  0  1 166
# 50 ggplot2     0.5     ggplot2_0.5.tar.gz 2007-06-01 7.1M 160 160   0  0  0 160
}
\seealso{
Other pdiff: 
\code{\link{pkg_cache}()},
\code{\link{pkg_diff}()},
\code{\link{pkg_info}()},
\code{\link{pkg_versions}()},
\code{\link{print.pcache}()},
\code{\link{print.pdiff}()},
\code{\link{print.pdiff_score}()},
\code{\link{print.pinfo}()}
}
\concept{pdiff}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkgdiff.R
\docType{package}
\name{pkgdiff}
\alias{pkgdiff}
\alias{pkgdiff-package}
\title{Identify Differences Between Package Versions}
\description{
The \strong{pkgdiff} package contains functions to
identify differences between versions of R packages. The \strong{pkgdiff}
package was written specifically to detect breaking changes and
help with package upgrades.
}
\section{Functions}{

The main functions included in the \strong{pkgdiff} package are
as follows:
\itemize{
\item {\code{\link{pkg_info}}: Retrieves general information about a package,
mostly from the package description file.}
\item {\code{\link{pkg_versions}}: Gets a data frame containing all
versions of a package and the release date for each.}
\item {\code{\link{pkg_diff}}: Compares two package versions, and
returns an object that contains the differences in terms of exported
function signatures.}
\item {\code{\link{pkg_stability}}: Compiles information on package
stability over time, and calculates a score based on this information.
This function is useful when making package selection choices.}
\item {\code{\link{pkg_repo}}: Fetches package names and versions for
packages in a specified repository.}
\item {\code{\link{repo_breakages}}: Identifies breakages when
upgrading from one version of a repository to another.}
\item {\code{\link{repo_stability}}: Performs a stability assessment
on multiple packages, even an entire repository.}
\item {\code{\link{pkg_summary}}: Generates a summary report in HTML
that can be viewed interactively.}
}
For additional information on \strong{pkgdiff}, see the package vignette
\code{vignette('pkgdiff')}, or visit the package web site
\href{https://pkgdiff.r-sassy.org/index.html}{here}.
}

\seealso{
Useful links:
\itemize{
  \item \url{https://pkgdiff.r-sassy.org}
  \item \url{https://github.com/dbosak01/pkgdiff}
  \item Report bugs at \url{https://github.com/dbosak01/pkgdiff/issues}
}

}
\author{
\strong{Maintainer}: David Bosak \email{dbosak01@gmail.com}

Other contributors:
\itemize{
  \item Brian Varney \email{brian.varney@experis.com} [contributor]
  \item Kevin Putschko \email{kevin.putschko@experis.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reports.R
\name{print.rstability}
\alias{print.rstability}
\title{Print a Repo Stability Object}
\usage{
\method{print}{rstability}(x, ...)
}
\arguments{
\item{x}{The repo stability object to print.}

\item{...}{Follow-on parameters to the print function.}
}
\value{
The repo stability object, invisibly.
}
\description{
Print routine for a repot stability object of class "rstability".
}
\examples{
# Create vector of packages
vct <- c("curl", "dplyr", "rvest", "tidymodels")

# Get stablity scores
res <- repo_stability(vct)

# View stability scores
print(res)
# # A repo stability object
# - Run Datetime: 2025-03-01 19:19 UTC
# - Summary:
#      Package    FV    LV         FR         LR TR BR Score        Assessment
# 1       curl   0.2 6.2.1 2014-11-20 2025-02-19 51  1  98.0       Very Stable
# 2      dplyr   0.1 1.1.4 2014-01-16 2023-11-17 45 20  87.5 Somewhat Unstable
# 3      rvest 0.1.0 1.0.4 2014-11-22 2024-02-12 14  4  93.2            Stable
# 4 tidymodels 0.0.1 1.3.0 2018-07-27 2025-02-21 14  0 100.0           Perfect
}
\seealso{
Other prepo: 
\code{\link{pkg_repo}()},
\code{\link{print.prepo}()},
\code{\link{print.rbreak}()},
\code{\link{repo_breakages}()},
\code{\link{repo_stability}()}
}
\concept{prepo}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reports.R
\name{repo_breakages}
\alias{repo_breakages}
\title{Identify Breakages for a Repository}
\usage{
repo_breakages(r1 = "current", r2 = "latest")
}
\arguments{
\item{r1}{A data frame that identifies the source repository packages
and versions. The default value is "current", which means
the function will use the current versions of all packages in
the current R repository.}

\item{r2}{A data frame that identifies the target repository packages
and versions. The default value is "latest", which means
the function will use the latest versions of all packages on CRAN.}
}
\value{
A list containing a summary table and a list of difference
objects. The summary table will contain one row for each package. The
columns show the package name, source and target versions, and whether
there are any breaking changes between the specified versions.  If
breaking changes are found, the difference object for that package
will be included in the details list.
}
\description{
The \code{repo_breakages} function generates a data frame
of breakage information for multiple packages.  This function can be used
to identify breakages for a small set of packages, or even an entire repository.
}
\details{
The \code{repo_breakages} function aims to help with repository upgrades.
The purpose is to identify packages that will break if you upgrade
your repository.

The function accepts two data frames as input.
Each data frame should have two columns: "Package" and "Version".
The "r1" data frame is for your current repo.  The "r2" data frame
is for the upgrade repo.

When executed, \code{repo_breakages} will then match up the
source and target versions, run a difference operation for each package,
and look for breaking changes.

The result will be a table of package names, the source and target version,
and whether or not there are any breakages. If there are
breakages, the difference object for that package will be
added to a "details" list.  This difference object can show you
exactly which functions or function parameters were removed.

The \code{repo_breakages} function is one of primary motivations for
the \strong{pkgdiff} package.  This mass-comparison ability did not
exist in any other R package prior to \strong{pkgdiff}.
}
\examples{
# Create package vector
pkgs <- c("curl", "dplyr", "crayon", "stringr")

# Backdated versions
r1 <- data.frame(Package = pkgs,
                 Version = c("5.2.1", "1.1.4", "1.5.2", "1.5.0"))

# More recent versions
r2 <- data.frame(Package = pkgs,
                 Version = c("6.2.1", "1.1.4", "1.5.3", "1.5.1"))

# Find any breaking changes
# res <- repo_breakages(r1, r2) # - Commented out to pass CRAN checks
# Comparing curl v5.2.1/v6.2.1
# Comparing crayon v1.5.2/v1.5.3
# Comparing stringr v1.5.0/v1.5.1

# View results
# res # - Commented out to pass CRAN checks
# # A repo breakages object
# - Run Datetime: 2025-03-01 17:47 UTC
# - Summary:
#   Package Version1 Version2 Breakages
# 1    curl    5.2.1    6.2.1      TRUE
# 2   dplyr    1.1.4    1.1.4     FALSE
# 3  crayon    1.5.2    1.5.3     FALSE
# 4 stringr    1.5.0    1.5.1     FALSE
# - Details:
#   # A difference object: curl package
#   - Comparing: v5.2.1/v6.2.1
# - Breaking Changes: TRUE
# - Added Functions: 1
# - Added Parameters: 3
# - Removed Parameters:
#   - multi_download(): timeout
}
\seealso{
Use the \code{\link{pkg_repo}} function to help gather
version information for each package set.  Also see \code{\link{repo_stability}}
to collect stability information on multiple packages.

Other prepo: 
\code{\link{pkg_repo}()},
\code{\link{print.prepo}()},
\code{\link{print.rbreak}()},
\code{\link{print.rstability}()},
\code{\link{repo_stability}()}
}
\concept{prepo}
